/* pipe-sim(+vector) ver.4.0 [92/05/30-06/15] */
/* 		     ver.4.0a[92/06/15-06/] */
/* 		     ver.4.3 [94/12/26-95/01/20] */
/* 		     ver.4.3b[96/05/30] */
/* 		     ver.4.3f[97/01/23,97/02/07] */
/* pipeline ALU stage */

#include "defines.h"
#include "latches.h"
#include "vectorls.h"
#include "cmnvar.h"
#include "opcode.h"
#include "extfunc.h"
#include "regdef.h"
#include "register_set.h"
#include "aalu.h"
#include "vpipectl.h"

static decoded_inst i_latch[NIP];

static reglatch alu_internal_outlatch[NIP];

static reglatch internal_fpin1[NIP];
static reglatch internal_fpin2[NIP];

static char alu_0_status[NIP][SSLEN];
static char alu_1_status[NIP][SSLEN];

char *calc_alu(ip,op) /* alu calculation */
int	ip;	/* ipipe number */
int	op;	/* alu operation type */
{
	static char status[SSLEN];

	switch (op) {
	    case OP_NOP:
		/* do nothing */
		strcpy(status,"no operation");
		break;
	    case OP_ADD:
		alu_internal_outlatch[ip].val.i = alu_in2[ip].val.i + alu_in1[ip].val.i;
		sprintf(status,"calculate %d + %d ",alu_in2[ip].val.i,alu_in1[ip].val.i);
		break;
	    case OP_SUB:
		alu_internal_outlatch[ip].val.i = alu_in2[ip].val.i - alu_in1[ip].val.i;
		sprintf(status,"calculate %d - %d ",alu_in2[ip].val.i,alu_in1[ip].val.i);
		break;
	    case OP_MUL:
		alu_internal_outlatch[ip].val.i = alu_in2[ip].val.i * alu_in1[ip].val.i;
		sprintf(status,"calculate %d * %d ",alu_in2[ip].val.i,alu_in1[ip].val.i);
		break;
	    case OP_DIV:
		if (alu_in1[ip].val.i == 0 ) {
		    error("divide by zero",
			  "ALU phase 0",ip);
		    sprintf(status,"calculate %d / %d (divide by 0)",alu_in2[ip].val.i,alu_in1[ip].val.i);
		} else {
		    alu_internal_outlatch[ip].val.i = alu_in2[ip].val.i / alu_in1[ip].val.i;
		    sprintf(status,"calculate %d / %d ",alu_in2[ip].val.i,alu_in1[ip].val.i);
		}
		break;
	    case OP_MOD:
		if (alu_in1[ip].val.i == 0 ) {
		    error("modulo by zero",
			  "ALU phase 0",ip);
		    sprintf(status,"calculate %d / %d (modulo by 0)",alu_in2[ip].val.i,alu_in1[ip].val.i);
		} else {
		    alu_internal_outlatch[ip].val.i = alu_in2[ip].val.i % alu_in1[ip].val.i;
		    sprintf(status,"calculate %d % %d ",alu_in2[ip].val.i,alu_in1[ip].val.i);
		}
		break;
	    case OP_RSUB:
		alu_internal_outlatch[ip].val.i = alu_in1[ip].val.i - alu_in2[ip].val.i;
		sprintf(status,"calculate %d - %d ",alu_in1[ip].val.i,alu_in2[ip].val.i);
		break;
	    case OP_UMUL:
		alu_internal_outlatch[ip].val.u = alu_in2[ip].val.u * alu_in1[ip].val.u;
		sprintf(status,"calculate %u * %u ",alu_in2[ip].val.u,alu_in1[ip].val.u);
		break;
	    case OP_UDIV:
		if (alu_in1[ip].val.u == 0 ) {
		    error("divide by zero",
			  "ALU phase 0",ip);
		    sprintf(status,"calculate %u / %u(divide by 0)",alu_in2[ip].val.u,alu_in1[ip].val.u);
		} else {
		    alu_internal_outlatch[ip].val.u = alu_in2[ip].val.u / alu_in1[ip].val.u;
		    sprintf(status,"calculate %u / %u ",alu_in2[ip].val.u,alu_in1[ip].val.u);
		}
		break;
	    case OP_AND:
		alu_internal_outlatch[ip].val.i = alu_in2[ip].val.i & alu_in1[ip].val.i;
		sprintf(status,"calculate %d & %d ",alu_in2[ip].val.i,alu_in1[ip].val.i);
		break;
	    case OP_OR:
		alu_internal_outlatch[ip].val.i = alu_in2[ip].val.i | alu_in1[ip].val.i;
		sprintf(status,"calculate %d | %d ",alu_in2[ip].val.i,alu_in1[ip].val.i);
		break;
	    case OP_XOR:
		alu_internal_outlatch[ip].val.i = alu_in2[ip].val.i ^ alu_in1[ip].val.i;
		sprintf(status,"calculate %d ^ %d ",alu_in2[ip].val.i,alu_in1[ip].val.i);
		break;
	    case OP_NOT:
		alu_internal_outlatch[ip].val.i = ~alu_in1[ip].val.i;
		sprintf(status,"calculate ~%d ",alu_in1[ip].val.i);
		break;
	    case OP_SRL:
		alu_internal_outlatch[ip].val.i = alu_in2[ip].val.i >> alu_in1[ip].val.u;
		sprintf(status,"calculate %d >> %u(Logical) ",alu_in2[ip].val.i,alu_in1[ip].val.u);
		break;
	    case OP_SLL:
		alu_internal_outlatch[ip].val.i = alu_in2[ip].val.i << alu_in1[ip].val.u;
		sprintf(status,"calculate %d << %u(Logical) ",alu_in2[ip].val.i,alu_in1[ip].val.u);
		break;
	    case OP_SRA:
		alu_internal_outlatch[ip].val.i = alu_in2[ip].val.i >> alu_in1[ip].val.u;
		sprintf(status,"calculate %d >> %u(Alithmetic) ",alu_in2[ip].val.i,alu_in1[ip].val.u);
		break;
	    case OP_SLA:
		alu_internal_outlatch[ip].val.i = alu_in2[ip].val.i << alu_in1[ip].val.u;
		sprintf(status,"calculate %d << %u(Alithmetic) ",alu_in2[ip].val.i,alu_in1[ip].val.u);
		break;
	    default:
		error("undefined alu operation",
		      "ALU phase 0",ip);
		break;
	}

	return status;
}
	    

char *alu_0(ip)
int	ip;	/* ip = instruction pipe number */
{
#ifndef NOVECTOR
	void vectorcalc_0(),vectorls_0(),vectormisc_0();
#endif

	i_latch[ip] = ALUlatch[ip];

	internal_fpin1[ip].val.f = 0.0;
	internal_fpin1[ip].regsrc = FALSE;
	internal_fpin1[ip].regdst = FALSE;
	internal_fpin2[ip] = internal_fpin1[ip];

	switch (i_latch[ip].optype) {
	    case LOAD:
	    case STORE:
		switch (i_latch[ip].subtype1) {
		    case AD_IMMED:
		    case AD_DIRECT:
		    case AD_BASED:
		    case AD_BASEINDX:
		    case AD_BASESFT:
		    case AD_PCBASED:
		    case AD_PCBINDX:
			strcpy(alu_0_status[ip],
			       calc_alu(ip,i_latch[ip].alu_operation));
			break;
		    default:
			error("undefined instruction(irregal addr. mode)",
			      "ALU phase 0",ip);
			break;
		}
		break;
	    case INTCALC:
		strcpy(alu_0_status[ip],
		       calc_alu(ip,i_latch[ip].alu_operation));
		break;
	    case FLOATCALC:
		if (i_latch[ip].subtype1 == REGOPR) {
		    if ((ip & 1) == 0) { /* ip = even number */ 
			switch (i_latch[ip].subtype2) {
			    case FOP_INT:
			    case FOP_FLT:
			    case FOP_ADD:
			    case FOP_SUB:
				if (fas_op[ip>>1] == FOP_NOP) {
                                    fas_op[ip>>1] = i_latch[ip].subtype2;
                                    faspin1[ip>>1] = alu_in1[ip];
                                    faspin2[ip>>1] = alu_in2[ip];
                                } else { /* float add/sub pipe busy */
                                    error("float add/sub pipe is busy(scalar float inst.)",
                                          "alu phase 0",ip);
                                }
				strcpy(alu_0_status[ip],"send float add/sub pipe.");
				break;
			    case FOP_MLT:
				if (fm_op[ip>>1] == FOP_NOP) {
                                    fm_op[ip>>1] = i_latch[ip].subtype2;
                                    fmpin1[ip>>1] = alu_in1[ip];
                                    fmpin2[ip>>1] = alu_in2[ip];
                                } else { /* float mul pipe busy */
                                    error("float mul pipe is busy(scalar float inst.)",
                                          "alu phase 0",ip);
                                }
				strcpy(alu_0_status[ip],"send float mult pipe.");
				break;
			    case FOP_DIV:
				if (fd_op[ip>>1] == FOP_NOP) {
                                    fd_op[ip>>1] = i_latch[ip].subtype2;
                                    fdpin1[ip>>1] = alu_in1[ip];
                                    fdpin2[ip>>1] = alu_in2[ip];
                                } else { /* float div pipe busy */
                                    error("float div pipe is busy(scalar float inst.)",
                                          "alu phase 0",ip);
                                }
				strcpy(alu_0_status[ip],"send float div pipe.");
				break;
			    default:
				error("undefined floatcalc instruction",
				      "alu phase 0",ip);
				break;
			}
		    } else { /* ip is odd number */
			/* copy alu input latch to internal latch */
			internal_fpin1[ip] = alu_in1[ip];
			internal_fpin2[ip] = alu_in2[ip];
			strcpy(alu_0_status[ip],"no operation");
		    }
		} else { /* undefined instruction */
		    error("undefined instruction(float immed operand)",
                          "alu phase 0",ip);
		}
		break; /* end of case FLOATCALC */
	    case JUMP:
		switch (i_latch[ip].subtype1) {
		    case JAD_REL:
			/* do nothing */
			strcpy(alu_0_status[ip],"no operation");
			break;
		    case JAD_REL_LK:
			if (i_latch[ip].jump_taken == TRUE) {
			    alu_in1[ip].val.i = oldPC;
			    alu_in2[ip].val.i = 4;
			    strcpy(alu_0_status[ip],calc_alu(ip,OP_ADD));
			} else { /* jump not taken */
			    strcpy(alu_0_status[ip],"no operation");
			}
			break;
		    case JAD_BASED:
			/* do nothing */
			strcpy(alu_0_status[ip],"no operation");
			break;
		    case JAD_BASED_LK:
			if (i_latch[ip].jump_taken == TRUE) {
			    alu_in1[ip].val.i = oldPC;
			    alu_in2[ip].val.i = 4;
			    strcpy(alu_0_status[ip],calc_alu(ip,OP_ADD));
			} else { /* jump not taken */
			    strcpy(alu_0_status[ip],"no operation");
			}
			break;
		}
		break;
	    case COMPJUMP:
		strcpy(alu_0_status[ip],
		       calc_alu(ip,i_latch[ip].alu_operation));
		break;
#ifndef NOVECTOR
	    case VECTOR:
		switch(i_latch[ip].subtype1){
		    case V_CALC: /* vectore calculation */
			vectorcalc_0(ip,0);
			break;   /* end of case V_CALC */
		    case V_LS:   /* vector load/store */
			vectorls_0(ip);
			break;   /* end of case V_LS */
		    case V_MISC:
			vectormisc_0(ip);
			break;  /* end of case V_MISC */
		}
		break; /* end of case VECTOR */
#endif
	    case NO_OPR:   /* no operation */
		strcpy(alu_0_status[ip],"no operation(by NO_OPR)");
		break;
	} /* end of switch (i_latch[ip].optype) */

	/* return status */
	/* strcpy(alu_0_status[ip],printinst(i_latch[ip],ip)); */
	return(alu_0_status[ip]);
}

char *alu_1(ip)
int	ip;
{
#ifndef NOVECTOR
	void vectormisc_1();
#endif

	memlatch[ip] = i_latch[ip];

	/* alu output latch destination register flag initialize */
	alu_out[ip].regdst = FALSE;

	jumploadpc_alu[ip] = FALSE;

	switch (i_latch[ip].optype) {
	    case LOAD:
		if (i_latch[ip].subtype1 == AD_IMMED) {
		    alu_out[ip] = alu_internal_outlatch[ip];
		    alu_out[ip].regdst = TRUE; /* alu output is immediate value */
		    alu_out[ip].dst = i_latch[ip].dstn;
		    /* return status */
		    sprintf(alu_1_status[ip],
			    "set %d to alu out latch ", alu_out[ip].val.i);
		} else {
		    alu_out[ip] = alu_internal_outlatch[ip];
		    alu_out[ip].regdst = FALSE; /* alu output is memory address */
		    alu_out[ip].dst = i_latch[ip].dstn;
		    /* return status */
		    sprintf(alu_1_status[ip],
			    "set %d to alu out latch ", alu_out[ip].val.i);
		}
		break;
	    case STORE:
		alu_out[ip] = alu_internal_outlatch[ip];
		alu_out[ip].regdst = FALSE; /* alu output is memory address */
		alu_out[ip].dst = i_latch[ip].dstn;
		/* return status */
		sprintf(alu_1_status[ip],
			"set %d to alu out latch ", alu_out[ip].val.i);
		break;
	    case INTCALC:
		alu_out[ip] = alu_internal_outlatch[ip];
		alu_out[ip].regdst = TRUE;
		alu_out[ip].dst = i_latch[ip].dstn;
		switch (i_latch[ip].subtype2) {
		    case OP_UMUL: /* unsigned */
		    case OP_UDIV:
			if (alu_out[ip].val.u == 0) {
			    setCC(ip,ZERO);
			    /* return status */
			    sprintf(alu_1_status[ip],
				    "set %u to alu out latch(Flag = Z)", 
				    alu_out[ip].val.u);
			} else {
			    setCC(ip,PLUS);
			    /* return status */
			    sprintf(alu_1_status[ip],
				    "set %u to alu out latch(Flag = P)", 
				    alu_out[ip].val.u);
			}
			break;
		    default:
			if (alu_out[ip].val.i > 0) {
			    setCC(ip,PLUS);
			    /* return status */
			    sprintf(alu_1_status[ip],
				    "set %d to alu out latch(Flag = P)", 
				    alu_out[ip].val.i);
			} else if (alu_out[ip].val.i == 0) {
			    setCC(ip,ZERO);
			    /* return status */
			    sprintf(alu_1_status[ip],
				    "set %d to alu out latch(Flag = Z)", 
				    alu_out[ip].val.i);
			} else if (alu_out[ip].val.i < 0) {
			    setCC(ip,MINUS);
			    /* return status */
			    sprintf(alu_1_status[ip],
				    "set %d to alu out latch(Flag = M)", 
				    alu_out[ip].val.i);
			}
			break;
		}
		break;
	    case FLOATCALC:
		if (i_latch[ip].subtype1 == REGOPR) {
		    if ((ip & 1) == 1) { /* ip = odd number */ 
			switch (i_latch[ip].subtype2) {
			    case FOP_INT:
			    case FOP_FLT:
			    case FOP_ADD:
			    case FOP_SUB:
				if (fas_op[ip>>1] == FOP_NOP) {
                                    fas_op[ip>>1] = i_latch[ip].subtype2;
                                    faspin1[ip>>1] = alu_in1[ip];
                                    faspin2[ip>>1] = alu_in2[ip];
                                } else { /* float add/sub pipe busy */
                                    error("float add/sub pipe is busy(scalar float inst.)",
                                          "alu phase 1",ip);
                                }
				strcpy(alu_1_status[ip],"send float add/sub pipe.");
				break;
			    case FOP_MLT:
				if (fm_op[ip>>1] == FOP_NOP) {
                                    fm_op[ip>>1] = i_latch[ip].subtype2;
                                    fmpin1[ip>>1] = alu_in1[ip];
                                    fmpin2[ip>>1] = alu_in2[ip];
                                } else { /* float mul pipe busy */
                                    error("float mul pipe is busy(scalar float inst.)",
                                          "alu phase 1",ip);
                                }
				strcpy(alu_1_status[ip],"send float mult pipe.");
				break;
			    case FOP_DIV:
				if (fd_op[ip>>1] == FOP_NOP) {
                                    fd_op[ip>>1] = i_latch[ip].subtype2;
                                    fdpin1[ip>>1] = alu_in1[ip];
                                    fdpin2[ip>>1] = alu_in2[ip];
                                } else { /* float div pipe busy */
                                    error("float div pipe is busy(scalar float inst.)",
                                          "alu phase 1",ip);
                                }
				strcpy(alu_1_status[ip],"send float div pipe.");
				break;
			    default:
				error("undefined floatcalc instruction",
				      "alu phase 1",ip);
				break;
			}
		    } else { /* ip is even number */
			strcpy(alu_1_status[ip],"no operation");
		    }
		} else { /* undefined instruction */
		    error("undefined instruction(float immed operand)",
                          "alu phase 1",ip);
		}
		break; /* end of case FLOATCALC */
	    case JUMP:
		switch (i_latch[ip].subtype1) {
		    case JAD_REL: /* no operation */
			strcpy(alu_1_status[ip],"no operation");
			break;
		    case JAD_REL_LK:
			if (i_latch[ip].jump_taken == TRUE) {
			    alu_out[ip] = alu_internal_outlatch[ip];
			    alu_out[ip].regdst = TRUE;
			    alu_out[ip].dst = i_latch[ip].dstn;
			    /* return status */
			    sprintf(alu_1_status[ip],
				    "set %d to alu out latch ", 
				    alu_out[ip].val.i);
			} else { /* jump not taken */
			    /* return status */
			    strcpy(alu_1_status[ip],"no operation");
			}
			break;
		    case JAD_BASED: /* no operation */
			strcpy(alu_1_status[ip],"no operation");
			break;
		    case JAD_BASED_LK:
			if (i_latch[ip].jump_taken == TRUE) {
			    alu_out[ip] = alu_internal_outlatch[ip];
			    alu_out[ip].regdst = TRUE;
			    alu_out[ip].dst = i_latch[ip].dstn;
			    /* return status */
			    sprintf(alu_1_status[ip],
				    "set %d to alu out latch ",
				    alu_out[ip].val.i);
			} else { /* jump not taken */
			    /* return status */
			    strcpy(alu_1_status[ip],"no operation");
			}
			break;
		}
		break;
	    case COMPJUMP:
		alu_out[ip] = alu_internal_outlatch[ip];
		switch (i_latch[ip].subtype1) {
                    case CJC_PLUS:
                        if (alu_out[ip].val.i > 0) {
			    AALUin1_sel[ip] = AA1_SELECT_PC;
			    AALUin2_sel[ip] = AA2_SELECT_DISP;
		  	    jumploadpc_alu[ip] = TRUE;
			    /* return status */
			    strcpy(alu_1_status[ip], "Comp&jump taken. ");
                        } else { 
			    AALUin1_sel[ip] = AA1_SELECT_PC;
			    AALUin2_sel[ip] = AA2_SELECT_NIP;
			    jumploadpc_alu[ip] = FALSE;
			    /* return status */
			    strcpy(alu_1_status[ip], "Comp&jump not taken. ");
                        }
			break;
                    case CJC_MINUS:
                        if (alu_out[ip].val.i < 0) {
			    AALUin1_sel[ip] = AA1_SELECT_PC;
			    AALUin2_sel[ip] = AA2_SELECT_DISP;
			    jumploadpc_alu[ip] = TRUE;
			    /* return status */
			    strcpy(alu_1_status[ip], "Comp&jump taken. ");
                        } else { 
			    AALUin1_sel[ip] = AA1_SELECT_PC;
			    AALUin2_sel[ip] = AA2_SELECT_NIP;
			    /* return status */
			    strcpy(alu_1_status[ip], "Comp&jump not taken. ");
			    jumploadpc_alu[ip] = FALSE;
                        }
			break;
                    case CJC_ZERO:
                        if (alu_out[ip].val.i == 0) {
			    AALUin1_sel[ip] = AA1_SELECT_PC;
			    AALUin2_sel[ip] = AA2_SELECT_DISP;
			    jumploadpc_alu[ip] = TRUE;
			    /* return status */
			    strcpy(alu_1_status[ip], "Comp&jump taken. ");
                        } else { 
			    AALUin1_sel[ip] = AA1_SELECT_PC;
			    AALUin2_sel[ip] = AA2_SELECT_NIP;
			    /* return status */
			    strcpy(alu_1_status[ip], "Comp&jump not taken. ");
			    jumploadpc_alu[ip] = FALSE;
                        }
			break;
                    case CJC_NONZERO:
                        if (alu_out[ip].val.i != 0) {
			    AALUin1_sel[ip] = AA1_SELECT_PC;
			    AALUin2_sel[ip] = AA2_SELECT_DISP;
			    jumploadpc_alu[ip] = TRUE;
			    /* return status */
			    strcpy(alu_1_status[ip], "Comp&jump taken. ");
                        } else { 
			    AALUin1_sel[ip] = AA1_SELECT_PC;
			    AALUin2_sel[ip] = AA2_SELECT_NIP;
			    /* return status */
			    strcpy(alu_1_status[ip], "Comp&jump not taken. ");
			    jumploadpc_alu[ip] = FALSE;
                        }
			break;
                }
                break;
#ifndef NOVECTOR
	    case VECTOR:
		switch(i_latch[ip].subtype1){
		    case V_MISC:
			vectormisc_1(ip);
			break;  /* end of case V_MISC */
		    default:
			strcpy(alu_1_status[ip],"no operation");
			break;
		}
		break; /* end of case VECTOR */
#endif
	    case NO_OPR:   /* no operation */
		strcpy(alu_1_status[ip],"no operation(by NO_OPR)");
		break;
	}

	/* return status */
	/* strcpy(alu_1_status[ip],printinst(i_latch[ip],ip)); */
	return(alu_1_status[ip]);
}

#ifndef NOVECTOR
/* vector calculation instruction phase 0 */
void vectorcalc_0(ip,phase)
int ip;
int phase;
{
	int	pipeop;

	switch (i_latch[ip].subtype2) {
	    case VOP_FADD:
		pipeop = FOP_ADD;
		startfasp(ip,pipeop,i_latch[ip].src1,i_latch[ip].src2,i_latch[ip].dstn,i_latch[ip].subtype3,i_latch[ip].vmr);
		strcpy(alu_0_status[ip],"start FASPC(VFADD)");
		break;
	    case VOP_FSUB:
		pipeop = FOP_SUB;
		startfasp(ip,pipeop,i_latch[ip].src1,i_latch[ip].src2,i_latch[ip].dstn,i_latch[ip].subtype3,i_latch[ip].vmr);
		strcpy(alu_0_status[ip],"start FASPC(VFSUB)");
		break;
	    case VOP_FMULT:
		pipeop = FOP_MLT;
		startfmp(ip,pipeop,i_latch[ip].src1,i_latch[ip].src2,i_latch[ip].dstn,i_latch[ip].subtype3,i_latch[ip].vmr);
		strcpy(alu_0_status[ip],"start FMPC(VFMUL)");
		break;
	    case VOP_FDIV:
		pipeop = FOP_DIV;
		startfdp(ip,pipeop,i_latch[ip].src1,i_latch[ip].src2,i_latch[ip].dstn,i_latch[ip].subtype3,i_latch[ip].vmr);
		strcpy(alu_0_status[ip],"start FDPC(VFDIV)");
		break;
	    case VOP_IADD:
		pipeop = VILPOP_ADD;
		startvilp(ip,pipeop,i_latch[ip].src1,i_latch[ip].src2,alu_in1[ip].val,i_latch[ip].dstn,i_latch[ip].subtype3,i_latch[ip].vmr);
		strcpy(alu_0_status[ip],"start VILPC(VADD)");
		break;
	    case VOP_ISUB:
		pipeop = VILPOP_SUB;
		startvilp(ip,pipeop,i_latch[ip].src1,i_latch[ip].src2,alu_in1[ip].val,i_latch[ip].dstn,i_latch[ip].subtype3,i_latch[ip].vmr);
		strcpy(alu_0_status[ip],"start VILPC(VSUB)");
		break;
	    case VOP_IMULT:
		pipeop = VILPOP_MUL;
		startvilp(ip,pipeop,i_latch[ip].src1,i_latch[ip].src2,alu_in1[ip].val,i_latch[ip].dstn,i_latch[ip].subtype3,i_latch[ip].vmr);
		strcpy(alu_0_status[ip],"start VILPC(VMUL)");
		break;
	    case VOP_IDIV:
		pipeop = VILPOP_DIV;
		startvilp(ip,pipeop,i_latch[ip].src1,i_latch[ip].src2,alu_in1[ip].val,i_latch[ip].dstn,i_latch[ip].subtype3,i_latch[ip].vmr);
		strcpy(alu_0_status[ip],"start VILPC(VDIV)");
		break;
	    case VOP_LAND:
		pipeop = VILPOP_AND;
		startvilp(ip,pipeop,i_latch[ip].src1,i_latch[ip].src2,alu_in1[ip].val,i_latch[ip].dstn,i_latch[ip].subtype3,i_latch[ip].vmr);
		strcpy(alu_0_status[ip],"start VILPC(VAND)");
		break;
	    case VOP_LOR:
		pipeop = VILPOP_OR;
		startvilp(ip,pipeop,i_latch[ip].src1,i_latch[ip].src2,alu_in1[ip].val,i_latch[ip].dstn,i_latch[ip].subtype3,i_latch[ip].vmr);
		strcpy(alu_0_status[ip],"start VILPC(VOR)");
		break;
	    case VOP_LXOR:
		pipeop = VILPOP_XOR;
		startvilp(ip,pipeop,i_latch[ip].src1,i_latch[ip].src2,alu_in1[ip].val,i_latch[ip].dstn,i_latch[ip].subtype3,i_latch[ip].vmr);
		strcpy(alu_0_status[ip],"start VILPC(VXOR)");
		break;
	    case VOP_LNOT:
		pipeop = VILPOP_NOT;
		startvilp(ip,pipeop,i_latch[ip].src1,i_latch[ip].src2,alu_in1[ip].val,i_latch[ip].dstn,i_latch[ip].subtype3,i_latch[ip].vmr);
		strcpy(alu_0_status[ip],"start VILPC(VNOT)");
		break;
	    case VOP_BCAST:
		pipeop = VILPOP_BC;
		startvilp(ip,pipeop,i_latch[ip].src1,i_latch[ip].src2,alu_in1[ip].val,i_latch[ip].dstn,i_latch[ip].subtype3,i_latch[ip].vmr);
		strcpy(alu_0_status[ip],"start VILPC(VBCAST)");
		break;
	    case VOP_VSRL:
		pipeop = VILPOP_SRL;
		startvilp(ip,pipeop,i_latch[ip].src1,i_latch[ip].src2,alu_in1[ip].val,i_latch[ip].dstn,i_latch[ip].subtype3,i_latch[ip].vmr);
		strcpy(alu_0_status[ip],"start VILPC(VSRL)");
		break;
	    case VOP_VSLL:
		pipeop = VILPOP_SLL;
		startvilp(ip,pipeop,i_latch[ip].src1,i_latch[ip].src2,alu_in1[ip].val,i_latch[ip].dstn,i_latch[ip].subtype3,i_latch[ip].vmr);
		strcpy(alu_0_status[ip],"start VILPC(VSLL)");
		break;
	}
}

void vectorls_0(ip)	/* vector load/save instruction phase 0 */
int ip;
{
	switch (i_latch[ip].subtype2) {
	    case VOP_LDIR:
		/* vector load, increment is s-reg. */
		if (vluflag[ip]==VLSU_NOP) {
		    vloadinst[ip]=i_latch[ip].subtype2;
		    vloadbase[ip]=alu_in1[ip].val.i;
		    vloadincr[ip]=alu_in2[ip].val.i;
		    vloaddstn[ip]=i_latch[ip].dstn;
		    vloadvmr[ip]=i_latch[ip].vmr;
		    /* Vector Load Counter Reg. initialize */
		    if (SPECIALREG(N_VLCR+ip).i == 0) {
			SPECIALREG(N_VLCR+ip).i = VREGLEN;
		    } else if (SPECIALREG(N_VLCR+ip).i > VREGLEN) {
			error("vector length(VLCR) is out of range",
			      "alu phase 0",ip);
		    }
		    vluflag[ip]=VLSU_ADDRSET;
		    strcpy(alu_0_status[ip],"VLU Address set(LDIR).");
		} else { /* vlu is busy. */
		    error("vector load unit is busy",
			  "alu phase 0",ip);
		    break;
		}
		break;
	    case VOP_LDII:
		/* vector load, increment is immed. */
		if (vluflag[ip]==VLSU_NOP) {
		    vloadinst[ip]=i_latch[ip].subtype2;
		    vloadbase[ip]=alu_in1[ip].val.i;
		    vloadincr[ip]=alu_in2[ip].val.i;
		    vloaddstn[ip]=i_latch[ip].dstn;
		    vloadvmr[ip]=i_latch[ip].vmr;
		    /* Vector Load Counter Reg. initialize */
		    if (SPECIALREG(N_VLCR+ip).i == 0) {
			SPECIALREG(N_VLCR+ip).i = VREGLEN;
		    } else if (SPECIALREG(N_VLCR+ip).i > VREGLEN) {
			error("vector length(VLCR) is out of range",
			      "alu phase 0",ip);
		    }
		    vluflag[ip]=VLSU_ADDRSET;
		    strcpy(alu_0_status[ip],"VLU Address set(LDII).");
		} else { /* vlu is busy. */
		    error("vector load unit is busy",
			  "alu phase 0",ip);
		    break;
		}
		break;
	    case VOP_STIR:
		/* vector store, increment is s-reg. */
		if (vsuflag[ip]==VLSU_NOP) {
		    vstoreinst[ip]=i_latch[ip].subtype2;
		    vstorebase[ip]=alu_in1[ip].val.i;
		    vstoreincr[ip]=alu_in2[ip].val.i;
		    vstoresrc[ip]=i_latch[ip].dstn;
		    vstorevmr[ip]=i_latch[ip].vmr;
		    /* Vector Store Counter Reg. initialize */
		    if (SPECIALREG(N_VSCR+ip).i == 0) {
			SPECIALREG(N_VSCR+ip).i = VREGLEN;
		    } else if (SPECIALREG(N_VSCR+ip).i > VREGLEN) {
			error("vector length(VSCR) is out of range",
			      "alu phase 0",ip);
		    }
		    vsuflag[ip]=VLSU_ADDRSET;
		    strcpy(alu_0_status[ip],"VSU Address set(STIR).");
		} else { /* vsu is busy. */
		    error("vector store unit is busy",
			  "alu phase 0",ip);
		    break;
		}
		break;
	    case VOP_STII:
		/* vector store, increment is immed. */
		if (vsuflag[ip]==VLSU_NOP) {
		    vstoreinst[ip]=i_latch[ip].subtype2;
		    vstorebase[ip]=alu_in1[ip].val.i;
		    vstoreincr[ip]=alu_in2[ip].val.i;
		    vstoresrc[ip]=i_latch[ip].dstn;
		    vstorevmr[ip]=i_latch[ip].vmr;
		    /* Vector Store Counter Reg. initialize */
		    if (SPECIALREG(N_VSCR+ip).i == 0) {
			SPECIALREG(N_VSCR+ip).i = VREGLEN;
		    } else if (SPECIALREG(N_VSCR+ip).i > VREGLEN) {
			error("vector length(VSCR) is out of range",
			      "alu phase 0",ip);
		    }
		    vsuflag[ip]=VLSU_ADDRSET;
		    strcpy(alu_0_status[ip],"VSU Address set(STII).");
		} else { /* vsu is busy. */
		    error("vector store unit is busy",
			  "alu phase 0",ip);
		    break;
		}
		break;
	}
}

void vectormisc_0(ip)	/* vector misc. instructions phase 0 */
int ip;
{
	int vilp;
	int vrp;

	vilp = ip >> 1;
	switch (i_latch[ip].subtype2) {
	    case VOP_VMROP: /* vector mask reg. operation */
		error("VMR operation is not implemented.",
		      "alu phase 0",ip);
#define NOTDONE /* unimplemented. */
#ifndef NOTDONE
		if (vil_op[vilp] == VILPOP_NOP) {
		    switch (i_latch[ip].subtype3) {
			case MOP_AND:
			    vil_op[vilp] = VILPOP_AND;
			    vilpin1[vilp].i = vmaskreg[i_latch[ip].src1][0];
			    vilpin2[vilp].i = vmaskreg[i_latch[ip].src2][0];
			    break;
			case MOP_OR:
			    vil_op[vilp] = VILPOP_OR;
			    vilpin1[vilp].i = vmaskreg[i_latch[ip].src1][0];
			    vilpin2[vilp].i = vmaskreg[i_latch[ip].src2][0];
			    break;
			case MOP_XOR:
			    vil_op[vilp] = VILPOP_XOR;
			    vilpin1[vilp].i = vmaskreg[i_latch[ip].src1][0];
			    vilpin2[vilp].i = vmaskreg[i_latch[ip].src2][0];
			    break;
			case MOP_NOT:
			    vil_op[vilp] = VILPOP_NOT;
			    vilpin1[vilp].i = vmaskreg[i_latch[ip].src1][0];
			    break;
		    }
		} else {
		    error("vector ilp is busy",
			  "alu phase 0",ip);
		}
#endif
#undef NOTDONE
		break;
	    case VOP_VMRSET: /* vector mask reg. set */
                if (vilpcflag[vilp] == VILPC_NOP) {
		    switch (i_latch[ip].subtype3) {
			case MC_F_ZERO:
			case MC_F_NOTZERO:
			case MC_F_PLUS:
			case MC_F_NOTPLUS:
			case MC_F_MINUS:
			case MC_F_NOTMINUS:
			    vilpop[vilp] = VILPOP_FCZ;
			    vilpchain[vilp] = CHAIN_OFF;
			    vilpsrc1[vilp] = i_latch[ip].src1;
			    vilpsrc2[vilp] = i_latch[ip].src2;
			    vilpvmr[vilp] = i_latch[ip].dstn;
			    vilpvmrcond[vilp] = i_latch[ip].subtype3;
			    break;
			case MC_I_ZERO:
			case MC_I_NOTZERO:
			case MC_I_PLUS:
			case MC_I_NOTPLUS:
			case MC_I_MINUS:
			case MC_I_NOTMINUS:
			    vilpop[vilp] = VILPOP_ICZ;
			    vilpchain[vilp] = CHAIN_OFF;
			    vilpsrc1[vilp] = i_latch[ip].src1;
			    vilpsrc2[vilp] = i_latch[ip].src2;
			    vilpvmr[vilp] = i_latch[ip].dstn;
			    vilpvmrcond[vilp] = i_latch[ip].subtype3;
			    break;
		    }
                    vilpcflag[vilp] = VILPC_INIT;
		    /* Vector Int/logic Counter Reg. initialize */
		    if (SPECIALREG(N_VICR+vilp).i == 0) {
			SPECIALREG(N_VICR+vilp).i = VREGLEN;
		    } else if (SPECIALREG(N_VICR+vilp).i > VREGLEN) {
			error("vector length(VICR) is out of range",
			      "alu phase 0",ip);
		    }
                } else {
                    error("vector ilp controller is busy",
                          "alu phase 0",ip);
                }
		break;
	    case VOP_LVLS:
	    case VOP_LVBLS:
		switch (i_latch[ip].subtype3) {
		    case VOP_LVLD: /*list vector load*/
		    case VOP_LVBLD:/*based list vector load*/
			if (vluflag[ip]==VLSU_NOP) {
			    vloadinst[ip]=i_latch[ip].subtype3;
			    vloadlstv[ip]=i_latch[ip].src1;
			    vloaddstn[ip]=i_latch[ip].dstn;
			    vloadvmr[ip]=i_latch[ip].vmr;
			    if(i_latch[ip].subtype3 == VOP_LVLD) {
				vloadvmr[ip]=i_latch[ip].vmr;
				vloadbase[ip]=0;
			    } else {
				vloadvmr[ip]=0;
				vloadbase[ip]=alu_in1[ip].val.i;
			    }
			    /* Vector Load Counter Reg. initialize */
			    if (SPECIALREG(N_VLCR+ip).i == 0) {
				SPECIALREG(N_VLCR+ip).i = VREGLEN;
			    } else if (SPECIALREG(N_VLCR+ip).i > VREGLEN) {
				error("vector length(VLCR) is out of range",
				      "alu phase 0",ip);
			    }
			    vluflag[ip]=VLSU_ADDRSET;
			    strcpy(alu_0_status[ip],"VLU Address set(List Vector).");
			} else {
			    error("vector load unit is busy",
				  "alu phase 0",ip);
			    break;
			}
			break;
		    case VOP_LVST: /*list vector load*/
		    case VOP_LVBST:/*based list vector load*/
			if (vsuflag[ip]==VLSU_NOP) {
			    vstoreinst[ip]=i_latch[ip].subtype3;
			    vstorelstv[ip]=i_latch[ip].src1;
			    vstoresrc[ip]=i_latch[ip].dstn;
			    vstorevmr[ip]=i_latch[ip].vmr;
			    if(i_latch[ip].subtype3 == VOP_LVST) {
				vstorevmr[ip]=i_latch[ip].vmr;
				vstorebase[ip]=0;
			    } else {
				vstorevmr[ip]=0;
				vstorebase[ip]=alu_in1[ip].val.i;
			    }
			    /* Vector Store Counter Reg. initialize */
			    if (SPECIALREG(N_VSCR+ip).i == 0) {
				SPECIALREG(N_VSCR+ip).i = VREGLEN;
			    } else if (SPECIALREG(N_VSCR+ip).i > VREGLEN) {
				error("vector length(VSCR) is out of range",
				      "alu phase 0",ip);
			    }
			    vsuflag[ip]=VLSU_ADDRSET;
			    strcpy(alu_0_status[ip],"VSU Address set(List Vector).");
			} else {
			    error("vector store unit is busy",
				  "alu phase 0",ip);
			    break;
			}
			break;
		}
		break; /*end of case VOP_LVLS,VOP_LVBLS */
	    case VOP_CONV: /* vector float<->int conv */
                if (vilpcflag[vilp] == VILPC_NOP) {
		    switch (i_latch[ip].subtype3) {
			case VOP_CONVI2F:
			    vilpop[vilp] = VILPOP_I2F;
			    vilpchain[vilp] = CHAIN_OFF;
			    vilpsrc1[vilp] = i_latch[ip].src1;
			    vilpdstn[vilp] = i_latch[ip].dstn;
			    vilpvmr[vilp] = i_latch[ip].vmr;
			    break;
			case VOP_CONVF2I:
			    vilpop[vilp] = VILPOP_F2I;
			    vilpchain[vilp] = CHAIN_OFF;
			    vilpsrc1[vilp] = i_latch[ip].src1;
			    vilpdstn[vilp] = i_latch[ip].dstn;
			    vilpvmr[vilp] = i_latch[ip].vmr;
			    break;
		    }
                    vilpcflag[vilp] = VILPC_INIT;
		    /* Vector Int/logic Counter Reg. initialize */
		    if (SPECIALREG(N_VICR+vilp).i == 0) {
			SPECIALREG(N_VICR+vilp).i = VREGLEN;
		    } else if (SPECIALREG(N_VICR+vilp).i > VREGLEN) {
			error("vector length(VICR) is out of range",
			      "alu phase 0",ip);
		    }
                } else {
                    error("vector ilp controller is busy",
                          "alu phase 0",ip);
                }
		break;
	    case VOP_MOVES2V: /* move s-reg to v-reg */
		vrp = alu_in2[ip].val.i;
		if ((vrp >= 0)&&(vrp < VREGLEN)) {
		    vectorreg[i_latch[ip].dstn][vrp] = alu_in1[ip].val;
		    sprintf(alu_0_status[ip],"SR[%d](%d) -> VR[%d][%d]",
		    	    i_latch[ip].src1,alu_in1[ip].val.i,
		    	    i_latch[ip].dstn,vrp);
		} else {
		    error("movsv VR pointer out of range",
			  "alu phase 0",ip);
		}
		break;
	    case VOP_MOVEV2S: /* move v-reg to s-reg */
		vrp = alu_in2[ip].val.i;
		if ((vrp >= 0)&&(vrp < VREGLEN)) {
		    alu_internal_outlatch[ip].val=vectorreg[i_latch[ip].dstn][vrp];
		    sprintf(alu_0_status[ip],"VR[%d][%d](%d) -> alu_internal_outlatch",
		    	    i_latch[ip].dstn,vrp,alu_internal_outlatch[ip].val.i);
		} else {
		    error("movvs VR pointer out of range",
			  "alu phase 0",ip);
		}
		break;
	}
}

void vectormisc_1(ip)	/* vector misc. instructions phase 1 */
int ip;
{
	int vilp;

	vilp = ip >> 1;
	switch (i_latch[ip].subtype2) {
	    case VOP_VMROP: /* vector mask reg. operation */
		error("VMR operation is not implemented.",
		      "alu phase 0",ip);
#define NOTDONE /* unimplemented. */
#ifndef NOTDONE
		if (vil_op[vilp] == VILPOP_NOP) {
		    switch (i_latch[ip].subtype3) {
			case MOP_AND:
			    vil_op[vilp] = VILPOP_AND;
			    vilpin1[vilp].i = vmaskreg[i_latch[ip].src1][1];
			    vilpin2[vilp].i = vmaskreg[i_latch[ip].src2][1];
			    break;
			case MOP_OR:
			    vil_op[vilp] = VILPOP_OR;
			    vilpin1[vilp].i = vmaskreg[i_latch[ip].src1][1];
			    vilpin2[vilp].i = vmaskreg[i_latch[ip].src2][1];
			    break;
			case MOP_XOR:
			    vil_op[vilp] = VILPOP_XOR;
			    vilpin1[vilp].i = vmaskreg[i_latch[ip].src1][1];
			    vilpin2[vilp].i = vmaskreg[i_latch[ip].src2][1];
			    break;
			case MOP_NOT:
			    vil_op[vilp] = VILPOP_NOT;
			    vilpin1[vilp].i = vmaskreg[i_latch[ip].src1][1];
			    break;
		    }
		} else {
		    error("vector ilp is busy",
			  "alu phase 1",ip);
		}
#endif
#undef NOTDONE
		break;
	    case VOP_VMRSET: /* vector mask reg. set */
	    case VOP_CONV: /* vector float<->int conv */
		strcpy(alu_1_status[ip],"no operation");
		break;
	    case VOP_MOVEV2S: /* move v-reg to s-reg */
		alu_out[ip] = alu_internal_outlatch[ip];
		alu_out[ip].regdst = TRUE; 
		alu_out[ip].dst = i_latch[ip].src1; /* src1 = destination scalar reg. */
		/* return status */
		sprintf(alu_1_status[ip],
			"set %d to alu out latch ",alu_out[ip].val.i);
		break;
	    case VOP_MOVES2V: /* move s-reg to v-reg */
		strcpy(alu_1_status[ip],"no operation");
		break;
	}
}
#endif  /* end of #ifndef NOVECTOR*/
