/* pipe-sim(+vector) ver.3.0 [92/05/28-7/29] */
/* 		     ver.3.1 [92/08/26] */
/* 		     ver.3.2 [92/08/27-8/29] */
/* 		     ver.3.3 [92/09/3-9/10] */
/* 		     ver.3.4 [92/09/14] */
/* 		     ver.3.5 [92/12/14] */
/*                   ver.3.5a-R1.0 [93/10/25] */
/* 					     */
/* pipe-sim(+vector) ver.4.0 [94/05/30-6/15] */
/* 		     ver.4.3 [94/12/26-95/01/20] */
/* 		     ver.4.3b[96/05/30] */
/* 		     ver.4.3d[96/08/13] */
/* 		     ver.4.3f[97/01/23,97/02/07] */
/* debugger procedure */

#include <curses.h>
#include <string.h>
#include <ctype.h>
#include <signal.h>

#include "defines.h"
#include "latches.h"
#include "vectorls.h"
#include "cmnvar.h"
#include "vpipectl.h"
#include "opcode.h"
#include "display.h"
#include "debugger.h"
#include "extfunc.h"
#include "regdef.h"
#include "register_set.h"
#include "aalu.h"

static	char	dumpfile[256]; /* dump file name */

debugger_main()
{
	char	cmdbuff[128];
	char	command[128];
	char	parameter[128];
	int	endstate;
	char	ch;
	int	address;
	int	i,tmp;
	int	dumpstart,dumpend;
	extern char wait_inkey();
	extern int progload();dumpmem();

	initial();		/* initialize simulator */
	clearbreakpoint();	/* clear breakpoint */
	clearbreakcycle();	/* clear breakcycle */
	endstate = HALT_END;	/* initialize end status */

	display_cyclephase(cycle,0.0);
	display_status(cycle,0.0);
	dispflush();
	mvprintw(COMMANDLINE+1,0,"load object filename : %s",objfile);
	progload(objfile); 
	for(;;) {
	    mvinput(COMMANDLINE,0,"COMMAND: ",cmdbuff);
	    clearline(COMMANDLINE+1);
	    clearline(COMMANDLINE+2);
	    clearline(COMMANDLINE+3);

	    cmdlineanalyze(cmdbuff,command,parameter);
	    switch (checkcommand(command)) {
		case CMD_RUN:	/* initialize and run program */
		    initial();
		    if (parameter[0] != NULL) {
			PC = (int)strtol(parameter,(char **)NULL,0);
			addrALU_out[0] = PC;
			mvprintw(COMMANDLINE+1,0,"initial PC = 0x%X(%d)",PC,PC);
			dispflush();
		    }
		    endstate = simulate();
		    disp_endstate(endstate);
		    dispflush();
		    break;
		case CMD_INIT:	/* initialize simulator */
		    initial();
		    clearbreakpoint();	/* clear breakpoint */
		    clearbreakcycle();	/* clear breakcycle */
		    mvprintw(COMMANDLINE+1,0,"initialize finished.");
		    dispflush();
		    break;
		case CMD_TRACE: /* set trace on/off */
		    if (parameter[0] == (char)NULL) {
			mvinput(COMMANDLINE+1,0,"input on/off: ",parameter);
		    } else {
			mvprintw(COMMANDLINE+1,0,"input on/off: %s",parameter);
			clrtoeol();
		    }
		    if (strcmp(parameter,"on")==0) {
			statusdisp = TRUE;
		    } else if (strcmp(parameter,"off")==0) {
			statusdisp = FALSE;
			stepexec = FALSE;
		    }
		    mvprintw(COMMANDLINE+2,0,"trace mode is set to %s.",
			     (statusdisp == TRUE) ? "TRACE ON" : "TRACE OFF");
		    dispflush();
		    break;
		case CMD_CONT:	/* continue stopped execution */
		    if ((endstate == BREAK_STOP)||(endstate == INTERRUPT)) {
		        endstate = simulate();
		        disp_endstate(endstate);
		    } else { /* can't continue */
			mvprintw(COMMANDLINE+1,0,"can't continue.");
		    }
		    dispflush();
		    break;
		case CMD_DISPSR:	/* display scalar registers */
		    dispsr(SRdisptop);
		    dispflush();
		    regmemdisp = DISPSR;
		    while ((ch = wait_inkey()) != 'q') {
			switch (ch) {
			    case 'j':	/* display next 2 registers */
				SRdisptop += 2;
				break;
			    case 'n':	/* display next 20 registers */
				SRdisptop += 20;
				break;
			    case 'k':	/* display previous 2 registers */
				SRdisptop -= 2;
				break;
			    case 'p':	/* display previous 20 registers */
				SRdisptop -= 20;
				break;
			    default:
				break;
			}
			if (SRdisptop < 0) { /* VRdisptop range check */
			    SRdisptop = 0;
			} else if (SRdisptop > REGSIZE - 20) {
			    SRdisptop = REGSIZE - 20;
			}
			dispsr(SRdisptop);
			dispflush();
		    }
		    break;
		case CMD_DISPVR:	/* display vector registers */
		    dispvr(VRdispnum,VRdisptop);
		    dispflush();
		    regmemdisp = DISPVR;
		    while ((ch = wait_inkey()) != 'q') {
			switch (ch) {
			    case 'j':	/* display next 2 elements */
				VRdisptop += 2;
				break;
			    case 'n':	/* display next 20 elements */
				VRdisptop += 20;
				break;
			    case 'k':	/* display previous 2 elements */
				VRdisptop -= 2;
				break;
			    case 'p':	/* display previous 20 elements */
				VRdisptop -= 20;
				break;
			    case 'h':	/* display previous VR */
				VRdispnum--;
				break;
			    case 'l':	/* display next VR */
				VRdispnum++;
				break;
			    default:
				break;
			}
			if (VRdisptop < 0) { /* VRdisptop range check */
			    VRdisptop = 0;
			} else if (VRdisptop > VREGLEN - 20) {
			    VRdisptop = VREGLEN - 20;
			}
			if (VRdispnum < 0) { /* VRdispnum range check */
			    VRdispnum = 0;
			} else if (VRdispnum >= VREGSIZE) {
			    VRdispnum = VREGSIZE - 1;
			}
			dispvr(VRdispnum,VRdisptop);
			dispflush();
		    }
		    break;
		case CMD_DISPMEM:	/* display memories */
		    if (parameter[0] != NULL) { /* parameter is disptop addr. */
			memdisptop = (int)strtol(parameter,(char **)NULL,0);
			if (memdisptop < 0) { /* memdisptop range check */
			    memdisptop = 0;
			} else if (memdisptop > MEMSIZE - 20) {
			    memdisptop = MEMSIZE - 20;
			}
			mvprintw(COMMANDLINE+1,0,"display top addr = 0x%X(%d)",
				 memdisptop,memdisptop);
			dispflush();
		    }
		    dispmem(memdisptop);
		    dispflush();
		    regmemdisp = DISPMEM;
		    while ((ch = wait_inkey()) != 'q') {
			switch (ch) {
			    case 'j':	/* display next 2 memories */
				memdisptop += 2;
				break;
			    case 'n':	/* display next 20 memories */
				memdisptop += 20;
				break;
			    case 'N':	/* display next 1K memories */
				memdisptop += 1024;
				break;
			    case '>':	/* display tail 20 memories */
				memdisptop = MEMSIZE - 20;
				break;
			    case 'k':	/* display previous 2 memories */
				memdisptop -= 2;
				break;
			    case 'p':	/* display previous 20 memories */
				memdisptop -= 20;
				break;
			    case 'P':	/* display previous 1K memories */
				memdisptop -= 1024;
				break;
			    case '<':	/* display top 20 memories */
				memdisptop = 0;
				break;
			    default:
				break;
			}
			if (memdisptop < 0) { /* memdisptop range check */
			    memdisptop = 0;
			} else if (memdisptop > MEMSIZE - 20) {
			    memdisptop = MEMSIZE - 20;
			}
			dispmem(memdisptop);
			dispflush();
		    }
		    break;
		case CMD_DISPVMR:	/* display vector mask registers */
		    while ((ch = wait_inkey()) != 'q') {
			switch (ch) {
			    case 'j':
			    case 'n':	/* display next VMR */
				VMRdispnum++;
				break;
			    case 'k':
			    case 'p':	/* display previous VMR */
				VMRdispnum--;
				break;
			    default:
				break;
			}
			if (VMRdispnum < 0) { /* VMRdispnum range check */
			    VMRdispnum = 0;
			} else if (VMRdispnum >= VMRSIZE) {
			    VMRdispnum = VMRSIZE - 1;
			}
			dispvmr(VMRdispnum);
			dispflush();
		    }
		    break;
		case CMD_BREAKPOINT:	/* set breakpoint */
		    if (parameter[0] == (char)NULL) {
			mvinput(COMMANDLINE+1,0,"input breakpoint address : ",parameter);
		    } else {
			mvprintw(COMMANDLINE+1,0,"input breakpoint address : %s",parameter);
			clrtoeol();
		    }
		    address = (int) strtol(parameter,(char **)NULL,0);
		    setbreakpoint(address);
		    mvprintw(COMMANDLINE+2,0,"breakpoint is set at 0x%08X(%d).",
			     address,address);
		    dispflush();
		    break;
		case CMD_CLEARBP:	/* clear breakpoint*/
		    clearbreakpoint();
		    mvprintw(COMMANDLINE+1,0,"breakpoint is cleared.");
		    dispflush();
		    break;
		case CMD_BREAKCYCLE:	/* set breakcycle */
		    if (parameter[0] == (char)NULL) {
			mvinput(COMMANDLINE+1,0,"input breakcycle : ",parameter);
		    } else {
			mvprintw(COMMANDLINE+1,0,"input breakcycle : %s",parameter);
			clrtoeol();
		    }
		    address = (int) strtol(parameter,(char **)NULL,0);
		    setbreakcycle(address);
		    mvprintw(COMMANDLINE+2,0,"breakcycle is set at %d(0x%08X).",
			     address,address);
		    dispflush();
		    break;
		case CMD_CLEARBC:	/* clear breakcycle*/
		    clearbreakcycle();
		    mvprintw(COMMANDLINE+1,0,"breakcycle is cleared.");
		    dispflush();
		    break;
		case CMD_WAIT:	/* set wait time */
		    if (parameter[0] == (char)NULL) {
			mvinput(COMMANDLINE+1,0,"input wait time (msec) : ",parameter);
		    } else {
			mvprintw(COMMANDLINE+1,0,"input wait time (msec) : %s",parameter);
			clrtoeol();
		    }
		    tmp = waittime;
		    waittime = (int) strtol(parameter,(char **)NULL,0) * 1000;
		    if (waittime <= 0) waittime = tmp;
		    mvprintw(COMMANDLINE+2,0,"wait time is set to %dmsec.",
			     (int)(waittime/1000));
		    dispflush();
		    break;
		case CMD_STEP:	/* set step execution mode */
		    if (parameter[0] == (char)NULL) {
			mvinput(COMMANDLINE+1,0,"step execution on/off : ",parameter);
		    } else {
			mvprintw(COMMANDLINE+1,0,"step execution on/off : %s",parameter);
			clrtoeol();
		    }
		    if (strcmp(parameter,"on") == 0) {
			stepexec = TRUE;
			statusdisp = TRUE;
			mvprintw(COMMANDLINE+2,0,"step mode ON, trace mode ON.");
		    } else if (strcmp(parameter,"off") == 0) {
			stepexec = FALSE;
			mvprintw(COMMANDLINE+2,0,"step mode OFF.");
		    } else {  /* irregal input */
			mvprintw(COMMANDLINE+2,0,"step mode %s.",
				 (stepexec == TRUE)?"ON":"OFF");
		    }
		    dispflush();
		    break;
		case CMD_LOAD:	/* load object file */
		    if (parameter[0] == (char)NULL) {
			mvinput(COMMANDLINE+1,0,"input object filename : ",parameter);
		    } else {
			mvprintw(COMMANDLINE+1,0,"load object filename : %s",parameter);
			clrtoeol();
		    }
		    if (progload(parameter) == TRUE) { /* if load success */
			initial();		/* initialize simulator */
			clearbreakpoint();	/* clear breakpoint */
			clearbreakcycle();	/* clear breakcycle */
			endstate = HALT_END;/* initialize end status */
			display_cyclephase(cycle,0.0);
			display_status();
		    }
		    dispflush();
		    break;
		case CMD_DUMPMEM:	/* dump memory to file */
		    if (parameter[0] == (char)NULL) {
			mvinput(COMMANDLINE+1,0,"input memory dump filename : ",parameter);
		    } else {
			mvprintw(COMMANDLINE+1,0,"memory dump filename : %s",parameter);
			clrtoeol();
		    }
		    strcpy(dumpfile,parameter);
		    mvinput(COMMANDLINE+2,0,"input dump start address : ",parameter);
		    dumpstart = (int) strtol(parameter,(char **)NULL,0);
		    mvinput(COMMANDLINE+2,0,"input dump end address : ",parameter);
		    dumpend = (int) strtol(parameter,(char **)NULL,0);
		    if (dumpmem(dumpfile,dumpstart,dumpend)==TRUE) { /* dump success */
			mvprintw(COMMANDLINE+3,0,"memory[0x%X-0x%X] successfully dumped to the file '%s'.",
				 dumpstart,dumpend,dumpfile);
		    } else {	/* dump failed */
			mvprintw(COMMANDLINE+3,0,"memory dump failed.");
		    }
		    dispflush();
		    break;
		case CMD_STATUS:/* simulator status display */
		    clear();	/* clear screen */
		    mvprintw(COMMANDLINE+1,0,"TRACE %s, STEP %s, wait time %d msec.",
			     (statusdisp == TRUE)?"ON":"OFF" ,
			     (stepexec == TRUE)?"ON":"OFF" ,
			     (int)(waittime/1000)) ;
		    if (breakpoint.enable == TRUE) {
			mvprintw(COMMANDLINE+2,0,"breakpoint = 0x%08X(%d).",
				 breakpoint.address,breakpoint.address);
		    }
		    if (breakcycle.enable == TRUE) {
			mvprintw(COMMANDLINE+3,0,"breakcycle = %d.",
						  breakcycle.cycle);
		    }
		    display_cyclephase(cycle,0.0);
		    display_status();
		    dispflush();
		    break;
		case CMD_QUIT:	/* quit debugger */
		    mvinput(COMMANDLINE+1,0,"hit return to quit simulator.",parameter);
		    dispflush();
		    return;
		    break;
		case CMD_ERROR:
		default:
		    mvprintw(COMMANDLINE+1,0,"Command error.");
		    mvprintw(COMMANDLINE+2,0,"Commands : ");
		    i = 0;
		    while (cmdtable[i].type != CMD_ERROR) {
			printw("%s ",cmdtable[i].str);
			    i++;
		    }
		    dispflush();
		    break;
	    }
	}
}

int checkcommand(cmd)	/* check command and return command type */
char	*cmd;	/* command string */
{
	int	i,lastmatch,matchcount;

	/* perfect match test */
	i = 0;
	while (cmdtable[i].type != CMD_ERROR) {
		if (strcmp(cmd,cmdtable[i].str) == 0) {
			return (cmdtable[i].type);
		}
		i++;
	}

	/* substring match test */
	i=0;
        matchcount = 0;
        while (cmdtable[i].type != CMD_ERROR) {
                if (strncmp(cmd,cmdtable[i].str,strlen(cmd)) == 0) {
                        lastmatch = i;
                        matchcount ++;
                }
                i++;
        }
        if (matchcount == 1) {
                return (cmdtable[lastmatch].type);
        }

	return (CMD_ERROR);	/* command string no match */
}

cmdlineanalyze(buff,cmd,param)	/* command line analyzer */
char	*buff;	/* input buffer */
char	*cmd;	/* command */
char	*param;	/* parameter */
{
	char	*cp;
	char	c;
	int	i;

	if (strchr(buff,(int)' ') == NULL) { /* if space is not found in buff */
	    strcpy(cmd,buff);
	    strcpy(param,"");
	} else {
	    cp = strchr(buff,(int)' ');
	    i = cp - buff;
	    strncpy(cmd,buff,i);
	    cmd[i] = (char) NULL;
	    strcpy(param,cp+1);
	}
	return;
}

mvinput(y,x,msg,str)	/* print message and get string */
int	y,x;	/* display message coordinates */
char	*msg;	/* message string */
char	*str;	/* input string */
{
	mvprintw(y,x,msg);
	clrtoeol();
	dispflush();
	getstrwithbs(str);
}

getstrwithbs(str)	/* get string which can use back-spacing */
char	*str;
{
	char	ch;
	char	buff[128];	/* input buffer */
	int	x,y;
	int	cp;
	char	wait_inkey();

	cp = 0;
	getyx(stdscr,y,x);
	while((ch = wait_inkey())!='\n') {
	    if((ch >= ' ')&&(ch<='~')) { /* ch = space(0x20) - '~' (0x7e) */
		if (cp < 127) {		 /* append one char */
		    buff[cp] = ch;
		    addch(ch);
		    dispflush();
		    cp++;
		    buff[cp] = (char)0;
		}
	    } else if ((ch=='\b')||((int)ch==0x7f)) {	/* ch = backspace,DEL */
		if (cp > 0) {				/* delete one char */
		    ch = '\b';
		    addch(ch);
		    addch(' ');
		    addch(ch);
		    dispflush();
		    cp--;
		    buff[cp] = (char)0;
		}
	    } else if ((int)ch==0x1b) { /* ch = escape(0x1b) */
		move(y,x);		/* cancel all input characters */
		clrtoeol();
		dispflush();
		cp=0;
		buff[cp] = (char)0;
	    }
	}

	strcpy(str,buff);
}

char wait_inkey()	/* get one char with wait (no echo to screen) */
{
	char	ch;

	cbreak();	/* set cbreak mode */
	ch = getch();
	nocbreak();	/* set nocbreak mode */
	return ch;
}

void interrupt()	/* keyboard interrupt handler */
{
	interruptflag = TRUE; /* interrupt flag enable */

#ifdef SYSV	/* if SYSV, signal handler must be resetted. */
	signal (SIGINT,interrupt);
#endif
	return;
}

setbreakpoint(addr)	/* set breakpoint */
int	addr;	/* breakpoint address */
{
	breakpoint.enable = TRUE;
	breakpoint.address = addr;
	breakpoint.count = 0;
	breakpoint.limit = 0;
}

clearbreakpoint()	/* clear breakpoint */
{
	breakpoint.enable = FALSE;
	breakpoint.address = 0;
	breakpoint.count = 0;
	breakpoint.limit = 0;
}

int checkbreakpoint(pc) /* check breakpoint */
int	pc;	/* program counter value */
{
	if ((breakpoint.enable == TRUE) && (breakpoint.address == pc)) {
		return TRUE;
	} else {
		return FALSE;
	}
}

setbreakcycle(cycle)	/* set breakcycle */
int	cycle;	/* breakpoint address */
{
	breakcycle.enable = TRUE;
	breakcycle.cycle = cycle;
}

clearbreakcycle()	/* clear breakcycle */
{
	breakcycle.enable = FALSE;
	breakcycle.cycle = 0;
}

int checkbreakcycle(cycle) /* check breakcycle */
int	cycle;	/* execution cycle value */
{
	if ((breakcycle.enable == TRUE) && (breakcycle.cycle == cycle)) {
		return TRUE;
	} else {
		return FALSE;
	}
}

disp_endstate(endstate)	/* display program stop status */
int	endstate;	/* end state */
{
	switch (endstate) {
	    case HALT_END:
		mvprintw(COMMANDLINE+1,0,"program is terminated by HALT instruction.");
		clrtoeol();
		break;
	    case ERROR_END:
		mvprintw(COMMANDLINE+1,0,"program is terminated by Run-time ERROR.");
		clrtoeol();
		break;
	    case BREAK_STOP:
		mvprintw(COMMANDLINE+1,0,"program is stopped at BREAKPOINT.");
		clrtoeol();
		break;
	    case INTERRUPT:
		mvprintw(COMMANDLINE+1,0,"program is stopped by INTERRUPT.");
		clrtoeol();
		break;
	    default:
		mvprintw(COMMANDLINE+1,0,"program is stopped with irregal endstate.");
		clrtoeol();
		break;
	}
}
