/* pipe-sim(+vector) ver.3.0 [92/05/28-7/29] */
/* 		     ver.3.2 [92/08/27-8/29] */
/* 		     ver.3.3 [92/09/03-9/10] */
/* 		     ver.3.4 [92/09/14] */
/* 		     ver.3.5a-R1.0 [93/10/25] */
/* 					      */
/* 		     ver.4.0 [94/5/30-6/15]   */
/* 		     ver.4.0c[94/11/18]   */
/* 		     ver.4.3 [94/12/26-95/01/20]   */
/* 		     ver.4.3c[96/8/9]   */
/* display procedures */

#include <stdio.h>
#include <curses.h>

#include "defines.h"
#include "latches.h"
#include "cmnvar.h"
#include "opcode.h"
#include "extfunc.h"
#include "regdef.h"
#include "register_set.h"

#include "display.h"
#include "debugger.h"
#include "memwait.h"

static char *optype_s[]={"LOAD","STORE","INTCALC","FLOATCALC","JUMP","COMPJUMP","VECTOR"};
static char *TFstr[]={"FALSE","TRUE"};

display(cycle,phase)	/* display status and wait */
int cycle;
double phase;
{
	char	ch;
	extern	char	wait_inkey();
	
	/* if batchmode then nothing to do */
	if (batchmode == TRUE) return;

	display_cyclephase(cycle,phase);

	if (statusdisp == TRUE) {
	    display_status(cycle,phase);

            if (stepexec == FALSE) { /* continuous execution mode */
		dispflush();	     /* flush display */
                waitusec(waittime); /* waiting... */
            } else { /* step execution mode */
		dispflush();	     /* flush display */
		while ((ch = wait_inkey()) != '\n') {   /* wait for key input */
		    if ((ch == 'q')||(ch == 'Q')) {
			interruptflag = TRUE;	/* type 'q' then interrupt */
		    }
		}
	    }
	} else {
	    dispflush();     /* flush display */
	}
}

display_cyclephase(cycle,phase) /* display cycle & phase */
int	cycle;
double	phase;
{
	mvprintw (CYCLELINE,0,"cycle %8d(+memwait %6d): phase %1.0f : PC = 0x%08X(%8d): write_collision %d",
		  cycle,totalmemwait,phase,PC,PC,write_collision);
	clrtoeol();
}

display_status()  /* display simulator status */
{
	dispflags();	/* display flags */

	switch(regmemdisp) {
	    case DISPSR:
		dispsr(SRdisptop);	/* display SR */
		break;
	    case DISPVR:
		dispvr(VRdispnum,VRdisptop); /* display VR */
		break;
	    case DISPMEM:
		dispmem(memdisptop);/* display memory */
		break;
	    default:
		break;
	}

	dispvmr(VMRdispnum);/* display VMR */

	disppipeandunitstatus();   /* display pipes & units status */
}

error(errortype,errorplace,ip)
char	*errortype,*errorplace;
int	ip;
{
	if (batchmode == FALSE) {
	    mvprintw(ERRORLINE,0,"ERROR! -> %s at %s(ip=%1d)",errortype,errorplace,ip);
	    dispflush();
	} else {
	    printf("ERROR! -> %s at %s(ip=%1d)\n",errortype,errorplace,ip);
	}

	FLAGREG |= ERROR;
}

dispinit()	/* initialize display */
{
    initscr();
    noecho();	/* set echo off */
}

dispflush()	/* flush display */
{
    refresh();
}

waitusec(usec)	/* wait for usec passing. */
int	usec;	/* sleep micro-sec. */
{
    usleep(usec);
}

dispfinish()	/* close display */
{
	endwin();
}

void clearline(line)	/* clear from top to end of specified line */
int	line;
{
	move(line,0);
	clrtoeol();
}

void dispsr(top)/* display scalar registers */
int	top;	/* display start SR number */
{
	int	i;
	int	line;

	line = REGMEMLINE + 1;

	mvprintw (REGMEMLINE,0,"Scalar Registers:"); clrtoeol();
	for(i=top ; i<top+20 ; i+=2) {
	    mvprintw (line,0,"SR%03d:%15d(%15.7f)",
			i,linear_reg_read_nochk(i).i,
			  linear_reg_read_nochk(i).f);
	    clrtoeol();
	    mvprintw (line,40,"SR%03d:%15d(%15.7f)",
			i+1,linear_reg_read_nochk(i+1).i,
			  linear_reg_read_nochk(i+1).f);
	    clrtoeol();
	    line++;
	}
	clearline(line);
}

void dispvr(number,top)	/* display vector registers */
int	number;		/* display VR number */
int	top;		/* display start VR element number */
{
	int	i;
	int	line;

	line = REGMEMLINE + 1;

	mvprintw (REGMEMLINE,0,"Vector Registers:"); clrtoeol();
	for(i=top ; i<top+20 ; i+=2) {
	    mvprintw (line,0,"VR%02d[%03d]:%12d(%12.5f)",
			number,i,
			vectorreg[number][i].i,
			vectorreg[number][i].f);
	    clrtoeol();
	    mvprintw (line,40,"VR%02d[%03d]:%12d(%12.5f)",
			number,i+1,
			vectorreg[number][i+1].i,
			vectorreg[number][i+1].f);
	    clrtoeol();
	    line++;
	}
	clearline(line);
}

void dispmem(top)	/* display memory contents */
int	top;		/* display start address */
{
	int	i;
	int	line;

	line = REGMEMLINE + 1;

	mvprintw (REGMEMLINE,0,"Memories :"); clrtoeol();
	for(i=top ; i<top+20 ; i+=2) {
	    mvprintw (line,0,"M[%08X]:%12d(%12.5f)",
			i,memory[i].i,memory[i].f);
	    clrtoeol();
	    mvprintw (line,40,"M[%08X]:%12d(%12.5f)",
			i+1,memory[i+1].i,memory[i+1].f);
	    clrtoeol();
	    line++;
	}
	clearline(line);
}

dispflags()	/* display flags */
{
	int	i;

	mvprintw (FLAGLINE,0,"Flags: ERR HLT");
	for (i=0 ; i < NIP ; i++) printw(" SF%1d ZF%1d",i,i);
	mvprintw (FLAGLINE+1,0,"        %1d   %1d ",
		   ((FLAGREG & ERROR)!=0) ? 1 : 0,
		   ((FLAGREG & HALT)!=0) ? 1 : 0);
	for (i=0 ; i < NIP ; i++) printw("  %1d   %1d ",
		   ((CC(i) & 2)!=0) ? 1 : 0,
		   ((CC(i) & 1)!=0) ? 1 : 0);

	clearline(FLAGLINE+2);
}

void dispvmr(num)	/* display vmr */
int	num;		/* display vmr number */
{
	int	i,c;
	char	head64[128],tail64[128];

	for(i = c = 0 ; i < 64 ; i++) {
	    if ((i % 8) == 0) {
		head64[c] = tail64[c] = ' ';
		c++;
	    }
	    head64[c] = (VMR(num,i)!=0) ? '1' : '0' ;
	    tail64[c] = (VMR(num,i+64)!=0) ? '1' : '0' ;
	    c++;
	}
	head64[c] = tail64[c] = (char)0;

	mvprintw(VMRLINE,0,"VMR%1d:%s",num,head64);
	clrtoeol();
	mvprintw(VMRLINE+1,0,"     %s",tail64);
	clrtoeol();
	clearline(VMRLINE+2);
}

disppipeandunitstatus()	/* display pipeline/unit internal status */
{
	int	i;
	int	line;

	line = STATUSLINE;

	for (i = 0 ; i  < NIP ; i++) {
		mvprintw (line++,0,"ip%1d: IF = %s",i,if_ss[i]);
		clrtoeol();
		mvprintw (line++,0,"ip%1d: ID = %s",i,id_ss[i]);
		clrtoeol();
		mvprintw (line++,0,"ip%1d: ALU = %s",i,alu_ss[i]);
		clrtoeol();
		mvprintw (line++,0,"ip%1d: MEM = %s",i,mem_ss[i]);
		clrtoeol();
		mvprintw (line++,0,"ip%1d: WB = %s",i,wb_ss[i]);
		clrtoeol();
		mvprintw (line++,0,"ip%1d: FWB = %s",i,fwb_ss[i]);
		clrtoeol();
		line++;
	}
	for (i = 0 ; i  < NFASP ; i++) {
		mvprintw (line++,0,"fasp%1d: %s",i,fasp_in_ss[i]);
		clrtoeol();
	}
	for (i = 0 ; i  < NFMP ; i++) {
		mvprintw (line++,0,"fmp%1d: %s",i,fmp_in_ss[i]);
		clrtoeol();
	}
	for (i = 0 ; i  < NFDP ; i++) {
		mvprintw (line++,0,"fdp%1d: %s",i,fdp_in_ss[i]);
		clrtoeol();
	}
#ifndef NOVECTOR
	for (i = 0 ; i  < NVILP ; i++) {
		mvprintw (line++,0,"vilp%1d: %s",i,vilp_in_ss[i]);
		clrtoeol();
	}
#endif
	clearline(line++);
#ifndef NOVECTOR
	for (i = 0 ; i  < NFASP ; i++) {
		mvprintw (line++,0,"fasp-ctl%1d: %s",i,faspctl_ss[i]);
		clrtoeol();
	}
	for (i = 0 ; i  < NFMP ; i++) {
		mvprintw (line++,0,"fmp-ctl%1d: %s",i,fmpctl_ss[i]);
		clrtoeol();
	}
	for (i = 0 ; i  < NFDP ; i++) {
		mvprintw (line++,0,"fdp-ctl%1d: %s",i,fdpctl_ss[i]);
		clrtoeol();
	}
	for (i = 0 ; i  < NVILP ; i++) {
		mvprintw (line++,0,"vilp-ctl%1d: %s",i,vilpctl_ss[i]);
		clrtoeol();
	}
#endif
	clearline(line++);
	for (i = 0 ; i  < NIP ; i++) {
		mvprintw (line++,0,"loadu%1d : %s",i,lu_ss[i]);
		clrtoeol();
	}
	for (i = 0 ; i  < NIP ; i++) {
		mvprintw (line++,0,"storeu%1d: %s",i,su_ss[i]);
		clrtoeol();
	}
#ifndef NOVECTOR
	for (i = 0 ; i  < NIP ; i++) {
		mvprintw (line++,0,"vlu%1d: %s",i,vlu_ss[i]);
		clrtoeol();
	}
	for (i = 0 ; i  < NIP ; i++) {
		mvprintw (line++,0,"vsu%1d: %s",i,vsu_ss[i]);
		clrtoeol();
	}
#endif
	clearline(line++);
}
