/* pipe-sim(+vector) ver.3.0 [92/05/28-7/29] */
/* 		     ver.4.3 [94/12/26-95/01/20] */
/* vector floating add/sub pipeline control unit */

#include "defines.h"
#include "latches.h"
#include "vectorls.h"
#include "vpipectl.h"
#include "cmnvar.h"
#include "opcode.h"

static int faspc_op[NFASP],faspc_chain[NFASP];
static int faspc_src1[NFASP],faspc_src2[NFASP],faspc_dstn[NFASP];
static int faspc_vmr[NFASP];
static int fasploop[NFASP];

static char faspc0_status[NFASP][SSLEN];
static char faspc1_status[NFASP][SSLEN];

char *faspctl0(fp)
int	fp;
{
	switch(faspcflag[fp]) {
	    case FASPC_NOP: /* no operation */
		strcpy(faspc0_status[fp],"no operation");
		break;
	    case FASPC_INIT: /* vector fasp controler initialize */
		faspc_op[fp] = faspop[fp];
		faspc_chain[fp] = faspchain[fp];
		faspc_src1[fp] = faspsrc1[fp];
		faspc_src2[fp] = faspsrc2[fp];
		faspc_dstn[fp] = faspdstn[fp];
		faspc_vmr[fp] = faspvmr[fp];
		fasploop[fp] = 0;
		faspc_vectorlength[fp] = SPECIALREG(N_VACR+fp).i;
		if (fas_op[fp] != FOP_NOP) {
		    error("float add/sub pipe busy.",
			  "vector fasp ctl. unit phase 0",fp);
		    break;
		} else {
		    fas_op[fp] = faspc_op[fp];
		}
		/* src1,src2 -> floating add/sub pipe */
		write_fasp(fp);

		/* floating add/sub pipe -> dstn */
		/* read_fasp(fp); */ /* resut is not available yet. */

		/* make return status */
		strcpy(faspc0_status[fp],"initialize: ");
		strcat(faspc0_status[fp],
		       faspc_state(fp,
				   faspc_op[fp],
				   faspc_src1[fp],
				   faspc_src2[fp],
				   faspc_dstn[fp],
				   faspc_chain[fp],
				   faspc_vmr[fp],
				   fasploop[fp])
		      );

		/* decrement Vector Add/Sub Count Reg. */
		SPECIALREG(N_VACR+fp).i--;

		/* increment loop Count */
		fasploop[fp]++;

		faspcflag[fp] = FASPC_RUN;
		break;	/* end of case FASPC_INIT */

	    case FASPC_RUN:  /* vector fasp controler is running */
		if (SPECIALREG(N_VACR+fp).i > 0) {
		    if (fas_op[fp] != FOP_NOP) {
			error("float add/sub pipe busy.",
			      "vector fasp ctl. unit phase 0",fp);
			break;
		    } else {
			fas_op[fp] = faspc_op[fp];
		    }
		    /* src1,src2 -> floating add/sub pipe */
		    write_fasp(fp);
		}

		/* floating add/sub pipe -> dstn */
		if (fasploop[fp] >= FASPLEN) { 
		    read_fasp(fp);
		}

		/* make return status */
		strcpy(faspc0_status[fp],"running: ");
		strcat(faspc0_status[fp],
		       faspc_state(fp,
				   faspc_op[fp],
				   faspc_src1[fp],
				   faspc_src2[fp],
				   faspc_dstn[fp],
				   faspc_chain[fp],
				   faspc_vmr[fp],
				   fasploop[fp])
		      );

		/* decrement Vector Add/Sub Count Reg. */
		if (SPECIALREG(N_VACR+fp).i>0) SPECIALREG(N_VACR+fp).i--;

		if (++fasploop[fp] == faspc_vectorlength[fp] + FASPLEN ) {
		    faspcflag[fp] = FASPC_NOP;
		}

		break;	/* end of case FASPC_RUN */
	} /* end of switch(faspcflag[fp]) */

        /* return status */
	return(faspc0_status[fp]);
}

char *faspctl1(fp)
int	fp;
{
	switch (faspcflag[fp]) {
	    case FASPC_NOP: /* no operation */
		strcpy(faspc1_status[fp],"no operation");
		break;
	    case FASPC_RUN:  /* vector fasp controler is running */
		if (SPECIALREG(N_VACR+fp).i > 0) {
		    if (fas_op[fp] != FOP_NOP) {
			error("float add/sub pipe busy.",
			      "vector fasp ctl. unit phase 1",fp);
			break;
		    } else {
			fas_op[fp] = faspc_op[fp];
		    }
		    /* src1,src2 -> floating add/sub pipe */
		    write_fasp(fp);
		}

		/* floating add/sub pipe -> dstn */
		if (fasploop[fp] >= FASPLEN) { 
		    read_fasp(fp);
		}

		/* make return status */
		strcpy(faspc1_status[fp],"running: ");
		strcat(faspc1_status[fp],
		       faspc_state(fp,
				   faspc_op[fp],
				   faspc_src1[fp],
				   faspc_src2[fp],
				   faspc_dstn[fp],
				   faspc_chain[fp],
				   faspc_vmr[fp],
				   fasploop[fp])
		      );

		/* decrement Vector Add/Sub Count Reg. */
		if (SPECIALREG(N_VACR+fp).i>0) SPECIALREG(N_VACR+fp).i--;

		if (++fasploop[fp] == faspc_vectorlength[fp] + FASPLEN ) {
		    faspcflag[fp] = FASPC_NOP;
		}

		break;	/* end of case FASPC_RUN */
	} /* end of switch(faspcflag[fp]) */

        /* return status */
	return(faspc1_status[fp]);
}

write_fasp(fp)    /* src1,src2 -> floating add/sub pipe */
int fp;
{
	switch (faspc_chain[fp]) {
	    case CHAIN_OFF:
		faspin1[fp].val = vectorreg[faspc_src1[fp]][fasploop[fp]];
		faspin1[fp].regsrc = VREG;
		faspin1[fp].src = VR_ENCODE(faspc_src1[fp],fasploop[fp]);
		faspin1[fp].regdst = VREG;
		faspin1[fp].dst = VR_ENCODE(faspc_dstn[fp],fasploop[fp]);
		faspin2[fp].val = vectorreg[faspc_src2[fp]][fasploop[fp]];
		faspin2[fp].regsrc = VREG;
		faspin2[fp].src = VR_ENCODE(faspc_src2[fp],fasploop[fp]);
		break;
	    case CHAIN_RRC:
		faspin1[fp].val = vectorreg[faspc_src1[fp]][fasploop[fp]];
		faspin1[fp].regsrc = VREG;
		faspin1[fp].src = VR_ENCODE(faspc_src1[fp],fasploop[fp]);
		faspin1[fp].regdst = CNET;
		faspin1[fp].dst = faspc_dstn[fp];
		faspin2[fp].val = vectorreg[faspc_src2[fp]][fasploop[fp]];
		faspin2[fp].regsrc = VREG;
		faspin2[fp].src = VR_ENCODE(faspc_src2[fp],fasploop[fp]);
		break;
	    case CHAIN_RCR:
		faspin1[fp].val = vectorreg[faspc_src1[fp]][fasploop[fp]];
		faspin1[fp].regsrc = VREG;
		faspin1[fp].src = VR_ENCODE(faspc_src1[fp],fasploop[fp]);
		faspin1[fp].regdst = VREG;
		faspin1[fp].dst = VR_ENCODE(faspc_dstn[fp],fasploop[fp]);
		faspin2[fp].val = chainnet_out[faspc_src2[fp]];
		faspin2[fp].regsrc = CNET;
		faspin2[fp].src = faspc_src2[fp];
		break;
	    case CHAIN_RCC:
		faspin1[fp].val = vectorreg[faspc_src1[fp]][fasploop[fp]];
		faspin1[fp].regsrc = VREG;
		faspin1[fp].src = VR_ENCODE(faspc_src1[fp],fasploop[fp]);
		faspin1[fp].regdst = CNET;
		faspin1[fp].dst = faspc_dstn[fp];
		faspin2[fp].val = chainnet_out[faspc_src2[fp]];
		faspin2[fp].regsrc = CNET;
		faspin2[fp].src = faspc_src2[fp];
		break;
	    case CHAIN_CRR:
		faspin1[fp].val = chainnet_out[faspc_src1[fp]];
		faspin1[fp].regsrc = CNET;
		faspin1[fp].src = faspc_src1[fp];
		faspin1[fp].regdst = VREG;
		faspin1[fp].dst = VR_ENCODE(faspc_dstn[fp],fasploop[fp]);
		faspin2[fp].val = vectorreg[faspc_src2[fp]][fasploop[fp]];
		faspin2[fp].regsrc = VREG;
		faspin2[fp].src = VR_ENCODE(faspc_src2[fp],fasploop[fp]);
	    case CHAIN_CRC:
		faspin1[fp].val = chainnet_out[faspc_src1[fp]];
		faspin1[fp].regsrc = CNET;
		faspin1[fp].src = faspc_src1[fp];
		faspin1[fp].regdst = CNET;
		faspin1[fp].dst = faspc_dstn[fp];
		faspin2[fp].val = vectorreg[faspc_src2[fp]][fasploop[fp]];
		faspin2[fp].regsrc = VREG;
		faspin2[fp].src = VR_ENCODE(faspc_src2[fp],fasploop[fp]);
		break;
	    case CHAIN_CCR:
		faspin1[fp].val = chainnet_out[faspc_src1[fp]];
		faspin1[fp].regsrc = CNET;
		faspin1[fp].src = faspc_src1[fp];
		faspin1[fp].regdst = VREG;
		faspin1[fp].dst = VR_ENCODE(faspc_dstn[fp],fasploop[fp]);
		faspin2[fp].val = chainnet_out[faspc_src2[fp]];
		faspin2[fp].regsrc = CNET;
		faspin2[fp].src = faspc_src2[fp];
		break;
	    case CHAIN_CCC:
		faspin1[fp].val = chainnet_out[faspc_src1[fp]];
		faspin1[fp].regsrc = CNET;
		faspin1[fp].src = faspc_src1[fp];
		faspin1[fp].regdst = CNET;
		faspin1[fp].dst = faspc_dstn[fp];
		faspin2[fp].val = chainnet_out[faspc_src2[fp]];
		faspin2[fp].regsrc = CNET;
		faspin2[fp].src = faspc_src2[fp];
		break;
	}
}

read_fasp(fp)	/* floating add/sub pipe -> dstn */
int fp;
{
	int reg,pos;

	switch (faspc_chain[fp]) {
	    case CHAIN_OFF:
	    case CHAIN_RCR:
	    case CHAIN_CRR:
	    case CHAIN_CCR:
		if (faspout[fp].regdst == VREG) {
		    reg = VR_DECODE_REG(faspout[fp].dst);
		    pos = VR_DECODE_POS(faspout[fp].dst);
		    /* write to register */
		    if (VMR(faspc_vmr[fp],fasploop[fp]-FASPLEN)==TRUE) {
			vectorreg[reg][pos] = faspout[fp].val;
		    }
		} else { 
		    error("irregal destflag on float add/sub pipe out latch.",
			  "vector fasp ctl. unit",fp);
		}
		break;
	    case CHAIN_RRC:
	    case CHAIN_RCC:
	    case CHAIN_CRC:
	    case CHAIN_CCC:
		if (faspout[fp].regdst == CNET) {
		    /* write to chaining network */
		    chainnet_in[faspout[fp].dst] = faspout[fp].val;
		} else { 
		    error("irregal destflag on float add/sub pipe out latch.",
			  "vector fasp ctl. unit",fp);
		}
		break;
	}
}
