/* pipe-sim(+vector) ver.3.0 [92/05/28-7/29] */
/* 		     ver.4.3 [94/12/26-95/01/20] */
/* vector floating div pipeline control unit */

#include "defines.h"
#include "latches.h"
#include "vectorls.h"
#include "vpipectl.h"
#include "cmnvar.h"
#include "opcode.h"

static int fdpc_op[NFDP],fdpc_chain[NFDP];
static int fdpc_src1[NFDP],fdpc_src2[NFDP],fdpc_dstn[NFDP];
static int fdpc_vmr[NFDP];
static int fdploop[NFDP];

static char fdpc0_status[NFDP][SSLEN];
static char fdpc1_status[NFDP][SSLEN];

char *fdpctl0(fp)
int	fp;
{
	switch(fdpcflag[fp]) {
	    case FDPC_NOP: /* no operation */
		strcpy(fdpc0_status[fp],"no operation");
		break;
	    case FDPC_INIT: /* vector fdp controler initialize */
		fdpc_op[fp] = fdpop[fp];
		fdpc_chain[fp] = fdpchain[fp];
		fdpc_src1[fp] = fdpsrc1[fp];
		fdpc_src2[fp] = fdpsrc2[fp];
		fdpc_dstn[fp] = fdpdstn[fp];
		fdpc_vmr[fp] = fdpvmr[fp];
		fdploop[fp] = 0;
		fdpc_vectorlength[fp] = SPECIALREG(N_VDCR+fp).i;
		if (fd_op[fp] != FOP_NOP) {
		    error("float div pipe busy.",
			  "vector fdp ctl. unit phase 0",fp);
		    break;
		} else {
		    fd_op[fp] = fdpc_op[fp];
		}
		/* src1,src2 -> floating div pipe */
		write_fdp(fp);

		/* floating div pipe -> dstn */
		/* read_fdp(fp); */ /* resut is not available yet. */

		/* make return status */
		strcpy(fdpc0_status[fp],"initialize: ");
		strcat(fdpc0_status[fp],
		       fdpc_state(fp,
				   fdpc_op[fp],
				   fdpc_src1[fp],
				   fdpc_src2[fp],
				   fdpc_dstn[fp],
				   fdpc_chain[fp],
				   fdpc_vmr[fp],
				   fdploop[fp])
		      );

		/* decrement Vector Div Count Reg. */
		SPECIALREG(N_VDCR+fp).i--;

		/* increment loop Count */
		fdploop[fp]++;

		fdpcflag[fp] = FDPC_RUN;
		break;	/* end of case FDPC_INIT */

	    case FDPC_RUN:  /* vector fdp controler is running */
		if (SPECIALREG(N_VDCR+fp).i > 0) {
		    if (fd_op[fp] != FOP_NOP) {
			error("float div pipe busy.",
			      "vector fdp ctl. unit phase 0",fp);
			break;
		    } else {
			fd_op[fp] = fdpc_op[fp];
		    }
		    /* src1,src2 -> floating div pipe */
		    write_fdp(fp);
		}

		/* floating div pipe -> dstn */
		if (fdploop[fp] >= FDPLEN) { 
		    read_fdp(fp);
		}

		/* make return status */
		strcpy(fdpc0_status[fp],"running: ");
		strcat(fdpc0_status[fp],
		       fdpc_state(fp,
				   fdpc_op[fp],
				   fdpc_src1[fp],
				   fdpc_src2[fp],
				   fdpc_dstn[fp],
				   fdpc_chain[fp],
				   fdpc_vmr[fp],
				   fdploop[fp])
		      );

		/* decrement Vector Div Count Reg. */
		if (SPECIALREG(N_VDCR+fp).i > 0) SPECIALREG(N_VDCR+fp).i--;

		if (++fdploop[fp] == fdpc_vectorlength[fp] + FDPLEN ) {
		    fdpcflag[fp] = FDPC_NOP;
		}

		break;	/* end of case FDPC_RUN */
	} /* end of switch(fdpcflag[fp]) */

        /* return status */
	return(fdpc0_status[fp]);
}

char *fdpctl1(fp)
int	fp;
{
	switch (fdpcflag[fp]) {
	    case FDPC_NOP: /* no operation */
		strcpy(fdpc1_status[fp],"no operation");
		break;
	    case FDPC_RUN:  /* vector fdp controler is running */
		if (SPECIALREG(N_VDCR+fp).i > 0) {
		    if (fd_op[fp] != FOP_NOP) {
			error("float div pipe busy.",
			      "vector fdp ctl. unit phase 1",fp);
			break;
		    } else {
			fd_op[fp] = fdpc_op[fp];
		    }
		    /* src1,src2 -> floating div pipe */
		    write_fdp(fp);
		}

		/* floating div pipe -> dstn */
		if (fdploop[fp] >= FDPLEN) { 
		    read_fdp(fp);
		}

		/* make return status */
		strcpy(fdpc1_status[fp],"running: ");
		strcat(fdpc1_status[fp],
		       fdpc_state(fp,
				   fdpc_op[fp],
				   fdpc_src1[fp],
				   fdpc_src2[fp],
				   fdpc_dstn[fp],
				   fdpc_chain[fp],
				   fdpc_vmr[fp],
				   fdploop[fp])
		      );

		/* decrement Vector Div Count Reg. */
		if (SPECIALREG(N_VDCR+fp).i > 0) SPECIALREG(N_VDCR+fp).i--;

		if (++fdploop[fp] == fdpc_vectorlength[fp] + FDPLEN ) {
		    fdpcflag[fp] = FDPC_NOP;
		}

		break;	/* end of case FDPC_RUN */
	} /* end of switch(fdpcflag[fp]) */

        /* return status */
	return(fdpc1_status[fp]);
}

write_fdp(fp)    /* src1,src2 -> floating div pipe */
int fp;
{
	switch (fdpc_chain[fp]) {
	    case CHAIN_OFF:
		fdpin1[fp].val = vectorreg[fdpc_src1[fp]][fdploop[fp]];
		fdpin1[fp].regsrc = VREG;
		fdpin1[fp].src = VR_ENCODE(fdpc_src1[fp],fdploop[fp]);
		fdpin1[fp].regdst = VREG;
		fdpin1[fp].dst = VR_ENCODE(fdpc_dstn[fp],fdploop[fp]);
		fdpin2[fp].val = vectorreg[fdpc_src2[fp]][fdploop[fp]];
		fdpin2[fp].regsrc = VREG;
		fdpin2[fp].src = VR_ENCODE(fdpc_src2[fp],fdploop[fp]);
		break;
	    case CHAIN_RRC:
		fdpin1[fp].val = vectorreg[fdpc_src1[fp]][fdploop[fp]];
		fdpin1[fp].regsrc = VREG;
		fdpin1[fp].src = VR_ENCODE(fdpc_src1[fp],fdploop[fp]);
		fdpin1[fp].regdst = CNET;
		fdpin1[fp].dst = fdpc_dstn[fp];
		fdpin2[fp].val = vectorreg[fdpc_src2[fp]][fdploop[fp]];
		fdpin2[fp].regsrc = VREG;
		fdpin2[fp].src = VR_ENCODE(fdpc_src2[fp],fdploop[fp]);
		break;
	    case CHAIN_RCR:
		fdpin1[fp].val = vectorreg[fdpc_src1[fp]][fdploop[fp]];
		fdpin1[fp].regsrc = VREG;
		fdpin1[fp].src = VR_ENCODE(fdpc_src1[fp],fdploop[fp]);
		fdpin1[fp].regdst = VREG;
		fdpin1[fp].dst = VR_ENCODE(fdpc_dstn[fp],fdploop[fp]);
		fdpin2[fp].val = chainnet_out[fdpc_src2[fp]];
		fdpin2[fp].regsrc = CNET;
		fdpin2[fp].src = fdpc_src2[fp];
		break;
	    case CHAIN_RCC:
		fdpin1[fp].val = vectorreg[fdpc_src1[fp]][fdploop[fp]];
		fdpin1[fp].regsrc = VREG;
		fdpin1[fp].src = VR_ENCODE(fdpc_src1[fp],fdploop[fp]);
		fdpin1[fp].regdst = CNET;
		fdpin1[fp].dst = fdpc_dstn[fp];
		fdpin2[fp].val = chainnet_out[fdpc_src2[fp]];
		fdpin2[fp].regsrc = CNET;
		fdpin2[fp].src = fdpc_src2[fp];
		break;
	    case CHAIN_CRR:
		fdpin1[fp].val = chainnet_out[fdpc_src1[fp]];
		fdpin1[fp].regsrc = CNET;
		fdpin1[fp].src = fdpc_src1[fp];
		fdpin1[fp].regdst = VREG;
		fdpin1[fp].dst = VR_ENCODE(fdpc_dstn[fp],fdploop[fp]);
		fdpin2[fp].val = vectorreg[fdpc_src2[fp]][fdploop[fp]];
		fdpin2[fp].regsrc = VREG;
		fdpin2[fp].src = VR_ENCODE(fdpc_src2[fp],fdploop[fp]);
	    case CHAIN_CRC:
		fdpin1[fp].val = chainnet_out[fdpc_src1[fp]];
		fdpin1[fp].regsrc = CNET;
		fdpin1[fp].src = fdpc_src1[fp];
		fdpin1[fp].regdst = CNET;
		fdpin1[fp].dst = fdpc_dstn[fp];
		fdpin2[fp].val = vectorreg[fdpc_src2[fp]][fdploop[fp]];
		fdpin2[fp].regsrc = VREG;
		fdpin2[fp].src = VR_ENCODE(fdpc_src2[fp],fdploop[fp]);
		break;
	    case CHAIN_CCR:
		fdpin1[fp].val = chainnet_out[fdpc_src1[fp]];
		fdpin1[fp].regsrc = CNET;
		fdpin1[fp].src = fdpc_src1[fp];
		fdpin1[fp].regdst = VREG;
		fdpin1[fp].dst = VR_ENCODE(fdpc_dstn[fp],fdploop[fp]);
		fdpin2[fp].val = chainnet_out[fdpc_src2[fp]];
		fdpin2[fp].regsrc = CNET;
		fdpin2[fp].src = fdpc_src2[fp];
		break;
	    case CHAIN_CCC:
		fdpin1[fp].val = chainnet_out[fdpc_src1[fp]];
		fdpin1[fp].regsrc = CNET;
		fdpin1[fp].src = fdpc_src1[fp];
		fdpin1[fp].regdst = CNET;
		fdpin1[fp].dst = fdpc_dstn[fp];
		fdpin2[fp].val = chainnet_out[fdpc_src2[fp]];
		fdpin2[fp].regsrc = CNET;
		fdpin2[fp].src = fdpc_src2[fp];
		break;
	}
}

read_fdp(fp)	/* floating div pipe -> dstn */
int fp;
{
	int reg,pos;

	switch (fdpc_chain[fp]) {
	    case CHAIN_OFF:
	    case CHAIN_RCR:
	    case CHAIN_CRR:
	    case CHAIN_CCR:
		if (fdpout[fp].regdst == VREG) {
		    reg = VR_DECODE_REG(fdpout[fp].dst);
		    pos = VR_DECODE_POS(fdpout[fp].dst);
		    /* write to register */
		    if (VMR(fdpc_vmr[fp],fdploop[fp]-FDPLEN)==TRUE) {
			vectorreg[reg][pos] = fdpout[fp].val;
		    }
		} else { 
		    error("irregal destflag on float div pipe out latch.",
			  "vector fdp ctl. unit",fp);
		}
		break;
	    case CHAIN_RRC:
	    case CHAIN_RCC:
	    case CHAIN_CRC:
	    case CHAIN_CCC:
		if (fdpout[fp].regdst == CNET) {
		    /* write to chaining network */
		    chainnet_in[fdpout[fp].dst] = fdpout[fp].val;
		} else { 
		    error("irregal destflag on float div pipe out latch.",
			  "vector fdp ctl. unit",fp);
		}
		break;
	}
}
