/* pipe-sim(+vector) ver.3.0 [92/05/28-7/29] */
/* vector floating mult/div pipeline control unit */

#include "defines.h"
#include "latches.h"
#include "vectorls.h"
#include "vpipectl.h"
#include "cmnvar.h"
#include "opcode.h"

static int fmdpc_op[NFMDP],fmdpc_chain[NFMDP];
static int fmdpc_src1[NFMDP],fmdpc_src2[NFMDP],fmdpc_dstn[NFMDP];
static int fmdpc_vmr[NFMDP];
static int fmdploop[NFMDP];

static char fmdpc0_status[NFMDP][SSLEN];
static char fmdpc1_status[NFMDP][SSLEN];
static char fmdpc2_status[NFMDP][SSLEN];
static char fmdpc3_status[NFMDP][SSLEN];

char *fmdpctl0(fp)
int	fp;
{
	switch (fmdpcflag[fp]) {
	    case FMDPC_NOP: /* no operation */
		strcpy(fmdpc0_status[fp],"no operation");
		break;
	    case FMDPC_RUN:  /* vector fmdp controler is running */
		if (fmdploop[fp] < VREGLEN) {
		    if (fmd_op[fp] != FOP_NOP) {
			error("float mult/div pipe busy.",
			      "vector fmdp ctl. unit phase 0",fp);
			break;
		    } else {
			fmd_op[fp] = fmdpc_op[fp];
		    }
		    /* src1,src2 -> floating mult/div pipe */
		    write_fmdp(fp);
		}

		/* floating mult/div pipe -> dstn */
		read_fmdp(fp);

                /* make return status */
                strcpy(fmdpc0_status[fp],"running: ");
                strcat(fmdpc0_status[fp],
                       fmdpc_state(fp,
                                   fmdpc_op[fp],
                                   fmdpc_src1[fp],
                                   fmdpc_src2[fp],
                                   fmdpc_dstn[fp],
                                   fmdpc_chain[fp],
                                   fmdpc_vmr[fp],
                                   fmdploop[fp])
                      );

		if (++fmdploop[fp] == VREGLEN + FMDPLEN) {
		    fmdpcflag[fp] = FMDPC_NOP;
		}

		break;	/* end of case FMDPC_RUN */
	} /* end of switch(fmdpcflag[fp]) */

	/* return status */
	return(fmdpc0_status[fp]);
}

char *fmdpctl1(fp)
int	fp;
{
	switch (fmdpcflag[fp]) {
	    case FMDPC_NOP: /* no operation */
		strcpy(fmdpc1_status[fp],"no operation");
		break;
	    case FMDPC_RUN:  /* vector fmdp controler is running */
		if (fmdploop[fp] < VREGLEN) {
		    if (fmd_op[fp] != FOP_NOP) {
			error("float mult/div pipe busy.",
			      "vector fmdp ctl. unit phase 1",fp);
			break;
		    } else {
			fmd_op[fp] = fmdpc_op[fp];
		    }
		    /* src1,src2 -> floating mult/div pipe */
		    write_fmdp(fp);
		}

		/* floating mult/div pipe -> dstn */
		read_fmdp(fp);

                /* make return status */
                strcpy(fmdpc1_status[fp],"running: ");
                strcat(fmdpc1_status[fp],
                       fmdpc_state(fp,
                                   fmdpc_op[fp],
                                   fmdpc_src1[fp],
                                   fmdpc_src2[fp],
                                   fmdpc_dstn[fp],
                                   fmdpc_chain[fp],
                                   fmdpc_vmr[fp],
                                   fmdploop[fp])
                      );

		if (++fmdploop[fp] == VREGLEN + FMDPLEN) {
		    fmdpcflag[fp] = FMDPC_NOP;
		}

		break;	/* end of case FMDPC_RUN */
	} /* end of switch(fmdpcflag[fp]) */

	/* return status */
	return(fmdpc1_status[fp]);
}

char *fmdpctl2(fp)
int	fp;
{
	switch(fmdpcflag[fp]) {
	    case FMDPC_NOP: /* no operation */
		strcpy(fmdpc2_status[fp],"no operation");
		break;
	    case FMDPC_INIT: /* vector fmdp controler initialize */
		fmdpc_op[fp] = fmdpop[fp];
		fmdpc_chain[fp] = fmdpchain[fp];
		fmdpc_src1[fp] = fmdpsrc1[fp];
		fmdpc_src2[fp] = fmdpsrc2[fp];
		fmdpc_dstn[fp] = fmdpdstn[fp];
		fmdpc_vmr[fp] = fmdpvmr[fp];
		fmdploop[fp] = FMDPLEN;
		if (fmd_op[fp] != FOP_NOP) {
		    error("float mult/div pipe busy.",
			  "vector fmdp ctl. unit phase 2",fp);
		    break;
		} else {
		    fmd_op[fp] = fmdpc_op[fp];
		}
		/* src1,src2 -> floating mult/div pipe */
		write_fmdp(fp);

		/* floating mult/div pipe -> dstn */
		read_fmdp(fp);

                /* make return status */
                strcpy(fmdpc2_status[fp],"initialize: ");
                strcat(fmdpc2_status[fp],
                       fmdpc_state(fp,
                                   fmdpc_op[fp],
                                   fmdpc_src1[fp],
                                   fmdpc_src2[fp],
                                   fmdpc_dstn[fp],
                                   fmdpc_chain[fp],
                                   fmdpc_vmr[fp],
                                   fmdploop[fp])
                      );

		fmdploop[fp]++;
		fmdpcflag[fp] = FMDPC_RUN;
		break;	/* end of case FMDPC_INIT */

	    case FMDPC_RUN:  /* vector fmdp controler is running */
		if (fmdploop[fp] < VREGLEN) {
		    if (fmd_op[fp] != FOP_NOP) {
			error("float mult/div pipe busy.",
			      "vector fmdp ctl. unit phase 2",fp);
			break;
		    } else {
			fmd_op[fp] = fmdpc_op[fp];
		    }
		    /* src1,src2 -> floating mult/div pipe */
		    write_fmdp(fp);
		}

		/* floating mult/div pipe -> dstn */
		read_fmdp(fp);

                /* make return status */
                strcpy(fmdpc2_status[fp],"running: ");
                strcat(fmdpc2_status[fp],
                       fmdpc_state(fp,
                                   fmdpc_op[fp],
                                   fmdpc_src1[fp],
                                   fmdpc_src2[fp],
                                   fmdpc_dstn[fp],
                                   fmdpc_chain[fp],
                                   fmdpc_vmr[fp],
                                   fmdploop[fp])
                      );

		if (++fmdploop[fp] == VREGLEN + FMDPLEN) {
		    fmdpcflag[fp] = FMDPC_NOP;
		}

		break;	/* end of case FMDPC_RUN */
	} /* end of switch(fmdpcflag[fp]) */

	/* return status */
	return(fmdpc2_status[fp]);
}

char *fmdpctl3(fp)
int	fp;
{
	switch (fmdpcflag[fp]) {
	    case FMDPC_NOP: /* no operation */
		strcpy(fmdpc3_status[fp],"no operation");
		break;
	    case FMDPC_RUN:  /* vector fmdp controler is running */
		if (fmdploop[fp] < VREGLEN) {
		    if (fmd_op[fp] != FOP_NOP) {
			error("float mult/div pipe busy.",
			      "vector fmdp ctl. unit phase 3",fp);
			break;
		    } else {
			fmd_op[fp] = fmdpc_op[fp];
		    }
		    /* src1,src2 -> floating mult/div pipe */
		    write_fmdp(fp);
		}

		/* floating mult/div pipe -> dstn */
		read_fmdp(fp);

                /* make return status */
                strcpy(fmdpc3_status[fp],"running: ");
                strcat(fmdpc3_status[fp],
                       fmdpc_state(fp,
                                   fmdpc_op[fp],
                                   fmdpc_src1[fp],
                                   fmdpc_src2[fp],
                                   fmdpc_dstn[fp],
                                   fmdpc_chain[fp],
                                   fmdpc_vmr[fp],
                                   fmdploop[fp])
                      );

		if (++fmdploop[fp] == VREGLEN + FMDPLEN) {
		    fmdpcflag[fp] = FMDPC_NOP;
		}

		break;	/* end of case FMDPC_RUN */
	} /* end of switch(fmdpcflag[fp]) */

	/* return status */
	return(fmdpc3_status[fp]);
}

write_fmdp(fp)    /* src1,src2 -> floating mult/div pipe */
{
	switch (fmdpc_chain[fp]) {
	    case CHAIN_OFF:
	    case CHAIN_RRC:
		fmdpin1[fp] = vectorreg[fmdpc_src1[fp]][fmdploop[fp]].f;
		fmdpin2[fp] = vectorreg[fmdpc_src2[fp]][fmdploop[fp]].f;
		break;
	    case CHAIN_RCR:
	    case CHAIN_RCC:
		fmdpin1[fp] = vectorreg[fmdpc_src1[fp]][fmdploop[fp]].f;
		fmdpin2[fp] = chainnet_out[fmdpc_src2[fp]].f;
		break;
	    case CHAIN_CRR:
	    case CHAIN_CRC:
		fmdpin1[fp] = chainnet_out[fmdpc_src1[fp]].f;
		fmdpin2[fp] = vectorreg[fmdpc_src2[fp]][fmdploop[fp]].f;
		break;
	    case CHAIN_CCR:
	    case CHAIN_CCC:
		fmdpin1[fp] = chainnet_out[fmdpc_src1[fp]].f;
		fmdpin2[fp] = chainnet_out[fmdpc_src2[fp]].f;
		break;
	}
}

read_fmdp(fp)	/* floating mult/div pipe -> dstn */
{
	switch (fmdpc_chain[fp]) {
	    case CHAIN_OFF:
	    case CHAIN_RCR:
	    case CHAIN_CRR:
	    case CHAIN_CCR:
		/* write to register */
		if (VMR(fmdpc_vmr[fp],fmdploop[fp]-FMDPLEN)==TRUE) {
		    vectorreg[fmdpc_dstn[fp]][fmdploop[fp]-FMDPLEN].f = fmdpout[fp];
		}
		break;
	    case CHAIN_RRC:
	    case CHAIN_RCC:
	    case CHAIN_CRC:
	    case CHAIN_CCC:
		/* write to chaining network */
		chainnet_in[fmdpc_dstn[fp]].f = fmdpout[fp];
		break;
	}
}
