/* pipe-sim(+vector) ver.3.0 [92/05/28-7/29] */
/* 		     ver.4.3 [94/12/26-95/01/20] */
/* vector floating mult pipeline control unit */

#include "defines.h"
#include "latches.h"
#include "vectorls.h"
#include "vpipectl.h"
#include "cmnvar.h"
#include "opcode.h"

static int fmpc_op[NFMP],fmpc_chain[NFMP];
static int fmpc_src1[NFMP],fmpc_src2[NFMP],fmpc_dstn[NFMP];
static int fmpc_vmr[NFMP];
static int fmploop[NFMP];

static char fmpc0_status[NFMP][SSLEN];
static char fmpc1_status[NFMP][SSLEN];

char *fmpctl0(fp)
int	fp;
{
	switch(fmpcflag[fp]) {
	    case FMPC_NOP: /* no operation */
		strcpy(fmpc0_status[fp],"no operation");
		break;
	    case FMPC_INIT: /* vector fmp controler initialize */
		fmpc_op[fp] = fmpop[fp];
		fmpc_chain[fp] = fmpchain[fp];
		fmpc_src1[fp] = fmpsrc1[fp];
		fmpc_src2[fp] = fmpsrc2[fp];
		fmpc_dstn[fp] = fmpdstn[fp];
		fmpc_vmr[fp] = fmpvmr[fp];
		fmploop[fp] = 0;
		fmpc_vectorlength[fp] = SPECIALREG(N_VMCR+fp).i;
		if (fm_op[fp] != FOP_NOP) {
		    error("float mult pipe busy.",
			  "vector fmp ctl. unit phase 0",fp);
		    break;
		} else {
		    fm_op[fp] = fmpc_op[fp];
		}
		/* src1,src2 -> floating mult pipe */
		write_fmp(fp);

		/* floating mult pipe -> dstn */
		/* read_fmp(fp); */ /* resut is not available yet. */

		/* make return status */
		strcpy(fmpc0_status[fp],"initialize: ");
		strcat(fmpc0_status[fp],
		       fmpc_state(fp,
				   fmpc_op[fp],
				   fmpc_src1[fp],
				   fmpc_src2[fp],
				   fmpc_dstn[fp],
				   fmpc_chain[fp],
				   fmpc_vmr[fp],
				   fmploop[fp])
		      );

		/* decrement Vector Mult Count Reg. */
		SPECIALREG(N_VMCR+fp).i--;

		/* increment loop Count */
		fmploop[fp]++;

		fmpcflag[fp] = FMPC_RUN;
		break;	/* end of case FMPC_INIT */

	    case FMPC_RUN:  /* vector fmp controler is running */
		if (SPECIALREG(N_VMCR+fp).i > 0) {
		    if (fm_op[fp] != FOP_NOP) {
			error("float mult pipe busy.",
			      "vector fmp ctl. unit phase 0",fp);
			break;
		    } else {
			fm_op[fp] = fmpc_op[fp];
		    }
		    /* src1,src2 -> floating mult pipe */
		    write_fmp(fp);
		}

		/* floating mult pipe -> dstn */
		if (fmploop[fp] >= FMPLEN) { 
		    read_fmp(fp);
		}

		/* make return status */
		strcpy(fmpc0_status[fp],"running: ");
		strcat(fmpc0_status[fp],
		       fmpc_state(fp,
				   fmpc_op[fp],
				   fmpc_src1[fp],
				   fmpc_src2[fp],
				   fmpc_dstn[fp],
				   fmpc_chain[fp],
				   fmpc_vmr[fp],
				   fmploop[fp])
		      );

		/* decrement Vector Mult Count Reg. */
		if (SPECIALREG(N_VMCR+fp).i > 0) SPECIALREG(N_VMCR+fp).i--;

		if (++fmploop[fp] == fmpc_vectorlength[fp] + FMPLEN ) {
		    fmpcflag[fp] = FMPC_NOP;
		}

		break;	/* end of case FMPC_RUN */
	} /* end of switch(fmpcflag[fp]) */

        /* return status */
	return(fmpc0_status[fp]);
}

char *fmpctl1(fp)
int	fp;
{
	switch (fmpcflag[fp]) {
	    case FMPC_NOP: /* no operation */
		strcpy(fmpc1_status[fp],"no operation");
		break;
	    case FMPC_RUN:  /* vector fmp controler is running */
		if (SPECIALREG(N_VMCR+fp).i > 0) {
		    if (fm_op[fp] != FOP_NOP) {
			error("float mult pipe busy.",
			      "vector fmp ctl. unit phase 1",fp);
			break;
		    } else {
			fm_op[fp] = fmpc_op[fp];
		    }
		    /* src1,src2 -> floating mult pipe */
		    write_fmp(fp);
		}

		/* floating mult pipe -> dstn */
		if (fmploop[fp] >= FMPLEN) { 
		    read_fmp(fp);
		}

		/* make return status */
		strcpy(fmpc1_status[fp],"running: ");
		strcat(fmpc1_status[fp],
		       fmpc_state(fp,
				   fmpc_op[fp],
				   fmpc_src1[fp],
				   fmpc_src2[fp],
				   fmpc_dstn[fp],
				   fmpc_chain[fp],
				   fmpc_vmr[fp],
				   fmploop[fp])
		      );

		/* decrement Vector Mult Count Reg. */
		if (SPECIALREG(N_VMCR+fp).i > 0) SPECIALREG(N_VMCR+fp).i--;

		if (++fmploop[fp] == fmpc_vectorlength[fp] + FMPLEN ) {
		    fmpcflag[fp] = FMPC_NOP;
		}

		break;	/* end of case FMPC_RUN */
	} /* end of switch(fmpcflag[fp]) */

        /* return status */
	return(fmpc1_status[fp]);
}

write_fmp(fp)    /* src1,src2 -> floating mult pipe */
int fp;
{
	switch (fmpc_chain[fp]) {
	    case CHAIN_OFF:
		fmpin1[fp].val = vectorreg[fmpc_src1[fp]][fmploop[fp]];
		fmpin1[fp].regsrc = VREG;
		fmpin1[fp].src = VR_ENCODE(fmpc_src1[fp],fmploop[fp]);
		fmpin1[fp].regdst = VREG;
		fmpin1[fp].dst = VR_ENCODE(fmpc_dstn[fp],fmploop[fp]);
		fmpin2[fp].val = vectorreg[fmpc_src2[fp]][fmploop[fp]];
		fmpin2[fp].regsrc = VREG;
		fmpin2[fp].src = VR_ENCODE(fmpc_src2[fp],fmploop[fp]);
		break;
	    case CHAIN_RRC:
		fmpin1[fp].val = vectorreg[fmpc_src1[fp]][fmploop[fp]];
		fmpin1[fp].regsrc = VREG;
		fmpin1[fp].src = VR_ENCODE(fmpc_src1[fp],fmploop[fp]);
		fmpin1[fp].regdst = CNET;
		fmpin1[fp].dst = fmpc_dstn[fp];
		fmpin2[fp].val = vectorreg[fmpc_src2[fp]][fmploop[fp]];
		fmpin2[fp].regsrc = VREG;
		fmpin2[fp].src = VR_ENCODE(fmpc_src2[fp],fmploop[fp]);
		break;
	    case CHAIN_RCR:
		fmpin1[fp].val = vectorreg[fmpc_src1[fp]][fmploop[fp]];
		fmpin1[fp].regsrc = VREG;
		fmpin1[fp].src = VR_ENCODE(fmpc_src1[fp],fmploop[fp]);
		fmpin1[fp].regdst = VREG;
		fmpin1[fp].dst = VR_ENCODE(fmpc_dstn[fp],fmploop[fp]);
		fmpin2[fp].val = chainnet_out[fmpc_src2[fp]];
		fmpin2[fp].regsrc = CNET;
		fmpin2[fp].src = fmpc_src2[fp];
		break;
	    case CHAIN_RCC:
		fmpin1[fp].val = vectorreg[fmpc_src1[fp]][fmploop[fp]];
		fmpin1[fp].regsrc = VREG;
		fmpin1[fp].src = VR_ENCODE(fmpc_src1[fp],fmploop[fp]);
		fmpin1[fp].regdst = CNET;
		fmpin1[fp].dst = fmpc_dstn[fp];
		fmpin2[fp].val = chainnet_out[fmpc_src2[fp]];
		fmpin2[fp].regsrc = CNET;
		fmpin2[fp].src = fmpc_src2[fp];
		break;
	    case CHAIN_CRR:
		fmpin1[fp].val = chainnet_out[fmpc_src1[fp]];
		fmpin1[fp].regsrc = CNET;
		fmpin1[fp].src = fmpc_src1[fp];
		fmpin1[fp].regdst = VREG;
		fmpin1[fp].dst = VR_ENCODE(fmpc_dstn[fp],fmploop[fp]);
		fmpin2[fp].val = vectorreg[fmpc_src2[fp]][fmploop[fp]];
		fmpin2[fp].regsrc = VREG;
		fmpin2[fp].src = VR_ENCODE(fmpc_src2[fp],fmploop[fp]);
	    case CHAIN_CRC:
		fmpin1[fp].val = chainnet_out[fmpc_src1[fp]];
		fmpin1[fp].regsrc = CNET;
		fmpin1[fp].src = fmpc_src1[fp];
		fmpin1[fp].regdst = CNET;
		fmpin1[fp].dst = fmpc_dstn[fp];
		fmpin2[fp].val = vectorreg[fmpc_src2[fp]][fmploop[fp]];
		fmpin2[fp].regsrc = VREG;
		fmpin2[fp].src = VR_ENCODE(fmpc_src2[fp],fmploop[fp]);
		break;
	    case CHAIN_CCR:
		fmpin1[fp].val = chainnet_out[fmpc_src1[fp]];
		fmpin1[fp].regsrc = CNET;
		fmpin1[fp].src = fmpc_src1[fp];
		fmpin1[fp].regdst = VREG;
		fmpin1[fp].dst = VR_ENCODE(fmpc_dstn[fp],fmploop[fp]);
		fmpin2[fp].val = chainnet_out[fmpc_src2[fp]];
		fmpin2[fp].regsrc = CNET;
		fmpin2[fp].src = fmpc_src2[fp];
		break;
	    case CHAIN_CCC:
		fmpin1[fp].val = chainnet_out[fmpc_src1[fp]];
		fmpin1[fp].regsrc = CNET;
		fmpin1[fp].src = fmpc_src1[fp];
		fmpin1[fp].regdst = CNET;
		fmpin1[fp].dst = fmpc_dstn[fp];
		fmpin2[fp].val = chainnet_out[fmpc_src2[fp]];
		fmpin2[fp].regsrc = CNET;
		fmpin2[fp].src = fmpc_src2[fp];
		break;
	}
}

read_fmp(fp)	/* floating mult pipe -> dstn */
int fp;
{
	int reg,pos;

	switch (fmpc_chain[fp]) {
	    case CHAIN_OFF:
	    case CHAIN_RCR:
	    case CHAIN_CRR:
	    case CHAIN_CCR:
		if (fmpout[fp].regdst == VREG) {
		    reg = VR_DECODE_REG(fmpout[fp].dst);
		    pos = VR_DECODE_POS(fmpout[fp].dst);
		    /* write to register */
		    if (VMR(fmpc_vmr[fp],fmploop[fp]-FMPLEN)==TRUE) {
			vectorreg[reg][pos] = fmpout[fp].val;
		    }
		} else { 
		    error("irregal destflag on float mult pipe out latch.",
			  "vector fmp ctl. unit",fp);
		}
		break;
	    case CHAIN_RRC:
	    case CHAIN_RCC:
	    case CHAIN_CRC:
	    case CHAIN_CCC:
		if (fmpout[fp].regdst == CNET) {
		    /* write to chaining network */
		    chainnet_in[fmpout[fp].dst] = fmpout[fp].val;
		} else { 
		    error("irregal destflag on float mult pipe out latch.",
			  "vector fmp ctl. unit",fp);
		}
		break;
	}
}
