/* pipe-sim(+vector) ver.4.0 [94/05/30-06/15] */
/* 		     ver.4.3 [94/12/26-95/01/20] */
/* floating pipeline (add/subtract) */

#include "defines.h"
#include "latches.h"
#include "cmnvar.h"
#include "opcode.h"

static reglatch	fasp[NFASP][FASPLEN],fasp_clear;
static int	fasptop[NFASP];

static char	faspstate[NFASP][SSLEN];
static char	fasp_outstate[NFASP][30];

void faspinit()	/* floating add/sub pipe initialize */
{
	int	fp,i;

	/* initialize for fasp clear structure */
	fasp_clear.val.f = 0.0;
	fasp_clear.regsrc = FALSE;
	fasp_clear.src = 0;
	fasp_clear.regdst = FALSE;
	fasp_clear.dst = 0;

	/* initializing fasp */
	for (fp = 0 ; fp < NFASP ; fp++) {
	    fas_op[fp] = FOP_NOP;
	    faspin1[fp] = faspin2[fp] = faspout[fp] = fasp_clear;
	    fasptop[fp] = 0;
	    for (i=0 ; i < FASPLEN ; i++) fasp[fp][i] = fasp_clear;
	}
}

void fasp_out(fp) /* fpipe -> fpipe_out_latch */
int fp;
{
	faspout[fp] = fasp[fp][fasptop[fp]];
	sprintf(fasp_outstate[fp]," ->%6.2f",faspout[fp].val.f);
	fasp[fp][fasptop[fp]] = fasp_clear;
}

char *fasp_in(fp)/*  fpipe_in_latches -> fpipe */
int	fp;
{
	int	lc,t;
	char	fasp_instate[40],fasp_tmpstr[30];

	switch (fas_op[fp]) {
		case FOP_NOP:	/* do nothing */
			fasp[fp][fasptop[fp]] = fasp_clear;
			sprintf(fasp_instate,"(no operation)=");
			faspin1[fp] = faspin2[fp] = fasp_clear;
			break;
		case FOP_ADD:	/* floating add */
			fasp[fp][fasptop[fp]].val.f = faspin2[fp].val.f+faspin1[fp].val.f ;
			fasp[fp][fasptop[fp]].regdst = faspin1[fp].regdst;
			fasp[fp][fasptop[fp]].dst = faspin1[fp].dst ;
			sprintf(fasp_instate,"%6.2f+%6.2f =",
				faspin2[fp].val.f,faspin1[fp].val.f);
			fas_op[fp] = FOP_NOP;
			faspin1[fp] = faspin2[fp] = fasp_clear;
			break;
		case FOP_SUB:	/* floating subtract */
			fasp[fp][fasptop[fp]].val.f = faspin2[fp].val.f-faspin1[fp].val.f ;
			fasp[fp][fasptop[fp]].regdst = faspin1[fp].regdst;
			fasp[fp][fasptop[fp]].dst = faspin1[fp].dst ;
			sprintf(fasp_instate,"%6.2f-%6.2f =",
				faspin2[fp].val.f,faspin1[fp].val.f);
			fas_op[fp] = FOP_NOP;
			faspin1[fp] = faspin2[fp] = fasp_clear;
			break;
		case FOP_INT:	/* floating to integer conversion */
			fasp[fp][fasptop[fp]].val.i = (int)faspin1[fp].val.f ;
			fasp[fp][fasptop[fp]].regdst = faspin1[fp].regdst;
			fasp[fp][fasptop[fp]].dst = faspin1[fp].dst ;
			sprintf(fasp_instate,"%(int)%6.2f =",faspin1[fp].val.f);
			fas_op[fp] = FOP_NOP;
			faspin1[fp] = faspin2[fp] = fasp_clear;
			break;
		case FOP_FLT:	/* floating to integer conversion */
			fasp[fp][fasptop[fp]].val.f = (float)faspin1[fp].val.i ;
			fasp[fp][fasptop[fp]].regdst = faspin1[fp].regdst;
			fasp[fp][fasptop[fp]].dst = faspin1[fp].dst ;
			sprintf(fasp_instate,"%(float)%6d =",faspin1[fp].val.i);
			fas_op[fp] = FOP_NOP;
			faspin1[fp] = faspin2[fp] = fasp_clear;
			break;
		default:
		        error("irregal float add/sub pipe operation",
			      "float add/sub pipe",fp);
		        break;
	}

	/* make fasp internal status */
	strcpy(faspstate[fp],fasp_instate);
	for (lc = fasptop[fp] ; lc < fasptop[fp]+FASPLEN ; lc++) {
		t = lc % FASPLEN;
		sprintf(fasp_tmpstr," %6.2f",fasp[fp][t].val.f);
		strcat(faspstate[fp],fasp_tmpstr);
	}
	strcat (faspstate[fp],fasp_outstate[fp]);

	/* refresh fasptop pointer */
	if (--fasptop[fp] < 0) fasptop[fp]=FASPLEN-1;

	/* return status */
	return (faspstate[fp]);
}
