/* pipe-sim(+vector) ver.4.0 [92/05/30-06/15] */
/* 		     ver.4.0a[92/06/16-06/] */
/* 		     ver.4.0b[94/10/21] */
/* 		     ver.4.1a[94/12/5] */
/* 		     ver.4.2 [94/12/19] */
/* 		     ver.4.3 [94/12/26-95/01/20] */
/* pipeline instruction decode/register fetch stage */

#include "defines.h"
#include "latches.h"
#include "cmnvar.h"
#include "opcode.h"
#include "extfunc.h"
#include "regdef.h"
#include "register_set.h"
#include "aalu.h"

static int bitmask[]={0,     0x1,       0x3,       0x7,       0xf,
			    0x1f,      0x3f,      0x7f,      0xff,
		           0x1ff,     0x3ff,     0x7ff,     0xfff,
		          0x1fff,    0x3fff,    0x7fff,    0xffff,
		         0x1ffff,   0x3ffff,   0x7ffff,   0xfffff,
		        0x1fffff,  0x3fffff,  0x7fffff,  0xffffff,
		       0x1ffffff, 0x3ffffff, 0x7ffffff, 0xfffffff,
		      0x1fffffff,0x3fffffff,0x7fffffff,0xffffffff};

#define GETBIT(var,pos,width) (((var) >> (pos)) & bitmask[(width)])

static int inst_latch[NIP];
static int addr[NIP];
static int int_ip;

static char dec_0_status[NIP][SSLEN];
static char dec_1_status[NIP][SSLEN];

/* register read with forwarding */
word reg_read_fwd(int ip, int no)
{
	int	chkip;

	if (no == 0) {	/* zero register */
	    return 0;
	} else {
	    for (chkip = 0; chkip < NIP ; chkip++) {
		/* alu output latch forwarding check */
		if (alu_out[chkip].regdst == TRUE) {
		    if (regmatch(ip,no,chkip,alu_out[chkip].dst) == TRUE) {
			/* register matched!! fowarding... */
			return alu_out[chkip].val;
		    }
		} 
	    }
	    for (chkip = 0; chkip < NIP ; chkip++) {
		/* writeback latch forwarding check */
		if (wb_latch[chkip].regdst == TRUE) {
		    if (regmatch(ip,no,chkip,wb_latch[chkip].dst) == TRUE) {
			/* register matched!! fowarding... */
			return wb_latch[chkip].val;
		    }
		} 
	    }
	    for (chkip = 0; chkip < NIP ; chkip++) {
		/* float writeback latch forwarding check */
		if (fwb_latch[chkip].regdst == TRUE) {
		    if (regmatch(ip,no,chkip,fwb_latch[chkip].dst) == TRUE) {
			/* register matched!! fowarding... */
			return fwb_latch[chkip].val;
		    }
		} 
	    }
	    /* forwarding check not matched. read register... */
	    return reg_read(ip,no);
	}
}

char *decode0(ip)
int	ip;
{
	if (idcancel != TRUE) {
	    inst_latch[ip] = pmdr[ip].i;
	    addr[ip] = pmar[ip];

	    /* return status */
	    sprintf(dec_0_status[ip],
		    "fetch inst. 0x%08X from addr. 0x%X",
		    inst_latch[ip],
		    addr[ip]);
	} else {
	    /* instruction decode cancelled */
	    sprintf(dec_0_status[ip],
		    "fetch inst. cancelled.");
	}
	return (dec_0_status[ip]);
}

char *decode1(ip)
int	ip;
{
	void dec_ls(),dec_ic(),dec_fc(),dec_jp(),dec_cj();
#ifndef NOVECTOR
	void dec_v();
#endif

	if (idcancel != TRUE) {
	    int_ip = ip;
	    jumploadpc_id[ip] = FALSE;
	    ALUlatch[ip].address = addr[ip];
	    ALUlatch[ip].optype = GETBIT(inst_latch[ip],29,3);
	    switch (ALUlatch[ip].optype) {
		case LOAD:
		case STORE:
			dec_ls(inst_latch[ip],&ALUlatch[ip]);
			break;
		case INTCALC:
			dec_ic(inst_latch[ip],&ALUlatch[ip]);
			/* return status */
			strcpy(dec_1_status[ip],"decoded inst : ");
			strcat(dec_1_status[ip],printinst(ALUlatch[ip],ip));
			break;
		case FLOATCALC:
			dec_fc(inst_latch[ip],&ALUlatch[ip]);
			/* return status */
			strcpy(dec_1_status[ip],"decoded inst : ");
			strcat(dec_1_status[ip],printinst(ALUlatch[ip],ip));
			break;
		case JUMP:
			dec_jp(inst_latch[ip],&ALUlatch[ip]);
			/* return status */
			strcpy(dec_1_status[ip],"decoded inst : ");
			strcat(dec_1_status[ip],printinst(ALUlatch[ip],ip));
			break;
		case COMPJUMP:
			dec_cj(inst_latch[ip],&ALUlatch[ip]);
			/* return status */
			strcpy(dec_1_status[ip],"decoded inst : ");
			strcat(dec_1_status[ip],printinst(ALUlatch[ip],ip));
			break;
#ifndef NOVECTOR
		case VECTOR:
			dec_v(inst_latch[ip],&ALUlatch[ip]);
			/* return status */
			strcpy(dec_1_status[ip],"decoded inst : ");
			strcat(dec_1_status[ip],printinst(ALUlatch[ip],ip));
			break;
#endif
		case NO_OPR: /* no operation */
			/* return status */
			strcpy(dec_1_status[ip],"decoded inst : ");
			strcat(dec_1_status[ip],"no operation");
			break;
		default:
			error("undefined instruction","decode phase 1",ip);
			break;
	    }

	} else {

	    /* instruction decode cancelled */
	    jumploadpc_id[ip] = FALSE;
	    ALUlatch[ip].address = addr[ip];
	    ALUlatch[ip].optype = NO_OPR;

	    /* return status */
	    strcpy(dec_1_status[ip],"inst decode cancelled.");

	}

	return (dec_1_status[ip]);
}

void dec_ls(inst,decode) /* decode load/store */
int	inst;
decoded_inst	*decode;
{
	decode->subtype1 = GETBIT(inst,26,3);
	decode->subtype2 = 0;

	switch (decode->subtype1) {
	    case AD_IMMED:
		if (decode->optype == STORE) {
		    error("undefined immediate store instruction","decode phase 1",int_ip);
		}
		decode->src1 = GETBIT(inst,8,17);
		decode->src2 = 0;
		decode->dstn = GETBIT(inst,0,8);
		decode->alu_operation = OP_OR;
		/* sign extension */
		if (GETBIT(decode->src1,16,1) == 1) {
		    decode->src1 |= ~0x1FFff;
		}
		/* set data to alu input latch */
		alu_in1[int_ip].val.i = decode->src1;
		alu_in1[int_ip].regsrc = FALSE;
		alu_in2[int_ip].val.i = 0;
		alu_in2[int_ip].regsrc = FALSE;
		/* return status */
		strcpy(dec_1_status[int_ip],"decoded inst : ");
		strcat(dec_1_status[int_ip],printinst(*decode,int_ip));
		break;
	    case AD_DIRECT:
		decode->src1 = GETBIT(inst,8,17);
		decode->src2 = 0;
		decode->dstn = GETBIT(inst,0,8);
		decode->alu_operation = OP_OR;
		/* sign extension */
		/* sign extension is not need for direct addressing */
		/*
		if (GETBIT(decode->src1,16,1) == 1) {
		    decode->src1 |= ~0x1FFff;
		}
		*/
		/* set data to alu input latch */
		alu_in1[int_ip].val.i = decode->src1;
		alu_in1[int_ip].regsrc = FALSE;
		alu_in2[int_ip].val.i = 0;
		alu_in2[int_ip].regsrc = FALSE;
		/* return status */
		strcpy(dec_1_status[int_ip],"decoded inst : ");
		strcat(dec_1_status[int_ip],printinst(*decode,int_ip));
		break;
	    case AD_PCBASED:
		decode->src1 = GETBIT(inst,8,17);
		decode->src2 = 0;
		decode->dstn = GETBIT(inst,0,8);
		decode->alu_operation = OP_ADD;
		/* sign extension */
		if (GETBIT(decode->src1,16,1) == 1) {
		    decode->src1 |= ~0x1FFff;
		}
		/* set data to alu input latch */
		alu_in1[int_ip].val.i = decode->src1;
		alu_in1[int_ip].regsrc = FALSE;
		alu_in2[int_ip].val.i = PC;
		alu_in2[int_ip].regsrc = TRUE;
		alu_in2[int_ip].src = PCREGNO;
		/* return status */
		strcpy(dec_1_status[int_ip],"decoded inst : ");
		strcat(dec_1_status[int_ip],printinst(*decode,int_ip));
		break;
	    case AD_BASED:
		decode->src1 = GETBIT(inst,16,9);
		decode->src2 = GETBIT(inst,8,8);
		decode->dstn = GETBIT(inst,0,8);
		decode->alu_operation = OP_ADD;
		/* sign extension */
		if (GETBIT(decode->src1,8,1) == 1) {
		    decode->src1 |= ~0x1ff;
		}
		/* set data to alu input latch */
		alu_in1[int_ip].val.i = decode->src1;
		alu_in1[int_ip].regsrc = FALSE;
		alu_in2[int_ip].val.i = reg_read_fwd(int_ip,decode->src2).i;
		alu_in2[int_ip].regsrc = TRUE;
		alu_in2[int_ip].src = decode->src2;
		/* return status */
		strcpy(dec_1_status[int_ip],"decoded inst : ");
		strcat(dec_1_status[int_ip],printinst(*decode,int_ip));
		break;
	    case AD_BASESFT:
		decode->src1 = GETBIT(inst,16,9);
		decode->src2 = GETBIT(inst,8,8);
		decode->dstn = GETBIT(inst,0,8);
		decode->alu_operation = OP_SRL;
		/* sign extension */
		if (GETBIT(decode->src1,8,1) == 1) {
		    decode->src1 |= ~0x1ff;
		}
		/* set data to alu input latch */
		alu_in1[int_ip].val.i = decode->src1;
		alu_in1[int_ip].regsrc = FALSE;
		alu_in2[int_ip].val.i = reg_read_fwd(int_ip,decode->src2).i;
		alu_in2[int_ip].regsrc = TRUE;
		alu_in2[int_ip].src = decode->src2;
		/* return status */
		strcpy(dec_1_status[int_ip],"decoded inst : ");
		strcat(dec_1_status[int_ip],printinst(*decode,int_ip));
		break;
	    case AD_BASEINDX:
		decode->src1 = GETBIT(inst,16,8);
		decode->src2 = GETBIT(inst,8,8);
		decode->dstn = GETBIT(inst,0,8);
		decode->alu_operation = OP_ADD;
		/* set data to alu input latch */
		alu_in1[int_ip].val.i = reg_read_fwd(int_ip,decode->src1).i;
		alu_in1[int_ip].regsrc = TRUE;
		alu_in1[int_ip].src = decode->src1;
		alu_in2[int_ip].val.i = reg_read_fwd(int_ip,decode->src2).i;
		alu_in2[int_ip].regsrc = TRUE;
		alu_in2[int_ip].src = decode->src2;
		/* return status */
		strcpy(dec_1_status[int_ip],"decoded inst : ");
		strcat(dec_1_status[int_ip],printinst(*decode,int_ip));
		break;
	    case AD_PCBINDX:
		decode->src1 = GETBIT(inst,16,8);
		decode->src2 = 0;
		decode->dstn = GETBIT(inst,0,8);
		decode->alu_operation = OP_ADD;
		/* set data to alu input latch */
		alu_in1[int_ip].val.i = reg_read_fwd(int_ip,decode->src1).i;
		alu_in1[int_ip].regsrc = TRUE;
		alu_in1[int_ip].src = decode->src1;
		alu_in2[int_ip].val.i = PC;
		alu_in2[int_ip].regsrc = TRUE;
		alu_in2[int_ip].src = PCREGNO;
		/* return status */
		strcpy(dec_1_status[int_ip],"decoded inst : ");
		strcat(dec_1_status[int_ip],printinst(*decode,int_ip));
		break;
	default:
		error("undefined load/store instruction","decode phase 1",int_ip);
		break;
	}
}

void dec_ic(inst,decode) /* decode intcalc */
int	inst;
decoded_inst	*decode;
{
	decode->subtype1 = GETBIT(inst,28,1);
	decode->subtype2 = GETBIT(inst,24,4);
	decode->alu_operation = GETBIT(inst,24,4);
	decode->src2 = GETBIT(inst,8,8);
	decode->dstn = GETBIT(inst,0,8);

	/* set data to alu input latch 2 */
	alu_in2[int_ip].val = reg_read_fwd(int_ip,decode->src2);
	alu_in2[int_ip].regsrc = TRUE;
	alu_in2[int_ip].src = decode->src2;

	switch (decode->subtype1) {
	    case REGOPR:
		decode->src1 = GETBIT(inst,16,8);
		/* set data to alu input latch 1 */
		alu_in1[int_ip].val = reg_read_fwd(int_ip,decode->src1);
		alu_in1[int_ip].regsrc = TRUE;
		alu_in1[int_ip].src = decode->src1;
		break;
	    case IMMOPR:
		switch (decode->subtype2) {
		    case OP_UMUL: /*unsigned*/
		    case OP_UDIV:
		    case OP_SRL:  /* shift operation */
		    case OP_SLL:
		    case OP_SRA:
		    case OP_SLA:
			decode->src1 = GETBIT(inst,16,8);
			/* set data to alu input latch 1 */
			alu_in1[int_ip].val.u = (unsigned)decode->src1;
			alu_in1[int_ip].regsrc = FALSE;
			break;
		    default:
			decode->src1 = GETBIT(inst,16,8);
			/* sign extension */
			if (GETBIT(decode->src1,7,1) == 1) {
			    decode->src1 |= ~0xff;
			}
			/* set data to alu input latch 1 */
			alu_in1[int_ip].val.i = decode->src1;
			alu_in1[int_ip].regsrc = FALSE;
			break;
		}
		break;
	}
}

void dec_fc(inst,decode) /* decode floatcalc */
int	inst;
decoded_inst	*decode;
{
	decode->subtype1 = GETBIT(inst,28,1);
	decode->subtype2 = GETBIT(inst,24,4);
	decode->alu_operation = NO_OPR;
	decode->src1 = GETBIT(inst,16,8);
	decode->src2 = GETBIT(inst,8,8);
	decode->dstn = GETBIT(inst,0,8);

	/* set data to floating point pipeline input latch */
	switch (decode->subtype2) {
	    case FOP_INT:
		alu_in1[int_ip].val.f = reg_read_fwd(int_ip,decode->src1).f;
		alu_in1[int_ip].regsrc = TRUE;
		alu_in1[int_ip].src = decode->src1;
		alu_in1[int_ip].regdst = TRUE;
		alu_in1[int_ip].dst = decode->dstn;
		alu_in2[int_ip].regsrc = FALSE;
		alu_in2[int_ip].regdst = FALSE;
		break;
	    case FOP_FLT:
		alu_in1[int_ip].val.i = reg_read_fwd(int_ip,decode->src1).i;
		alu_in1[int_ip].regsrc = TRUE;
		alu_in1[int_ip].src = decode->src1;
		alu_in1[int_ip].regdst = TRUE;
		alu_in1[int_ip].dst = decode->dstn;
		alu_in2[int_ip].regsrc = FALSE;
		alu_in2[int_ip].regdst = FALSE;
		break;
	    case FOP_ADD:
	    case FOP_SUB:
		alu_in1[int_ip].val.f = reg_read_fwd(int_ip,decode->src1).f;
		alu_in1[int_ip].regsrc = TRUE;
		alu_in1[int_ip].src = decode->src1;
		alu_in1[int_ip].regdst = TRUE;
		alu_in1[int_ip].dst = decode->dstn;
		alu_in2[int_ip].val.f = reg_read_fwd(int_ip,decode->src2).f;
		alu_in2[int_ip].regsrc = TRUE;
		alu_in2[int_ip].src = decode->src2;
		alu_in2[int_ip].regdst = TRUE;
		alu_in2[int_ip].dst = decode->dstn;
		break;
	    case FOP_MLT:
		alu_in1[int_ip].val.f = reg_read_fwd(int_ip,decode->src1).f;
		alu_in1[int_ip].regsrc = TRUE;
		alu_in1[int_ip].src = decode->src1;
		alu_in1[int_ip].regdst = TRUE;
		alu_in1[int_ip].dst = decode->dstn;
		alu_in2[int_ip].val.f = reg_read_fwd(int_ip,decode->src2).f;
		alu_in2[int_ip].regsrc = TRUE;
		alu_in2[int_ip].src = decode->src2;
		alu_in2[int_ip].regdst = TRUE;
		alu_in2[int_ip].dst = decode->dstn;
		break;
	    case FOP_DIV:
		alu_in1[int_ip].val.f = reg_read_fwd(int_ip,decode->src1).f;
		alu_in1[int_ip].regsrc = TRUE;
		alu_in1[int_ip].src = decode->src1;
		alu_in1[int_ip].regdst = TRUE;
		alu_in1[int_ip].dst = decode->dstn;
		alu_in2[int_ip].val.f = reg_read_fwd(int_ip,decode->src2).f;
		alu_in2[int_ip].regsrc = TRUE;
		alu_in2[int_ip].src = decode->src2;
		alu_in2[int_ip].regdst = TRUE;
		alu_in2[int_ip].dst = decode->dstn;
		break;
	}
}

void dec_jp(inst,decode) /* decode jump */
int	inst;
decoded_inst	*decode;
{
	decode->subtype1 = GETBIT(inst,27,2);
	decode->subtype2 = GETBIT(inst,24,3);
	switch (decode->subtype1) {
	    case JAD_REL:
		decode->src1 = GETBIT(inst,0,24);
		if (GETBIT(decode->src1,23,1) == 1) {
		    decode->src1 |= ~0xffffff;
		}   /* sign expansion */
	        decode->src1 <<= PADSFT;
		decode->src2 = 0;
		decode->dstn = 0;
		/* set data to addr.ALU input latch */
		jump_disp[int_ip] = decode->src1;
		AALUin1_sel[int_ip] = AA1_SELECT_PC;
		if (jumpCCchk(int_ip,decode->subtype2) == TRUE) {
		    /* jump taken */
		    AALUin2_sel[int_ip] = AA2_SELECT_DISP;
		    decode->jump_taken = TRUE;
		    jumploadpc_id[int_ip] = TRUE;
		} else {
		    /* jump not taken */
		    AALUin2_sel[int_ip] = AA2_SELECT_NIP;
		    decode->jump_taken = FALSE;
		}
		break;
	    case JAD_BASED:
		decode->src1 = GETBIT(inst,0,16);
		if (GETBIT(decode->src1,15,1) == 1) {
		    decode->src1 |= ~0xffff;
		}   /* sign expansion */
	        decode->src1 <<= PADSFT;
		decode->src2 = GETBIT(inst,16,8);
		decode->dstn = 0;
		/* set data to addr.ALU input latch */
		jump_disp[int_ip] = decode->src1;
		jump_basereg[int_ip] = reg_read_fwd(int_ip,decode->src2).i;
		if (jumpCCchk(int_ip,decode->subtype2) == TRUE) {
		    /* jump taken */
		    AALUin1_sel[int_ip] = AA1_SELECT_JBR;
		    AALUin2_sel[int_ip] = AA2_SELECT_DISP;
		    decode->jump_taken = TRUE;
		    jumploadpc_id[int_ip] = TRUE;
		} else {
		    /* jump not taken */
		    AALUin1_sel[int_ip] = AA1_SELECT_PC;
		    AALUin2_sel[int_ip] = AA2_SELECT_NIP;
		    decode->jump_taken = FALSE;
		}
		break;
	    case JAD_REL_LK:
		decode->src1 = GETBIT(inst,0,24);
		if (GETBIT(decode->src1,23,1) == 1) {
		    decode->src1 |= ~0xffffff;
		}   /* sign expansion */
	        decode->src1 <<= PADSFT;
		decode->src2 = 0;
		decode->dstn = LINKREG;
		/* set data to addr.ALU input latch */
		jump_disp[int_ip] = decode->src1;
		AALUin1_sel[int_ip] = AA1_SELECT_PC;
		if (jumpCCchk(int_ip,decode->subtype2) == TRUE) {
		    /* jump taken */
		    AALUin2_sel[int_ip] = AA2_SELECT_DISP;
		    decode->jump_taken = TRUE;
		    jumploadpc_id[int_ip] = TRUE;
		} else {
		    /* jump not taken */
		    AALUin2_sel[int_ip] = AA2_SELECT_NIP;
		    decode->jump_taken = FALSE;
		}
		break;
	    case JAD_BASED_LK:
		decode->src1 = GETBIT(inst,0,16);
		if (GETBIT(decode->src1,15,1) == 1) {
		    decode->src1 |= ~0xffff;
		}   /* sign expansion */
	        decode->src1 <<= PADSFT;
		decode->src2 = GETBIT(inst,16,8);
		decode->dstn = LINKREG;
		/* set data to addr.ALU input latch */
		jump_disp[int_ip] = decode->src1;
		jump_basereg[int_ip] = reg_read_fwd(int_ip,decode->src2).i;
		if (jumpCCchk(int_ip,decode->subtype2) == TRUE) {
		    /* jump taken */
		    AALUin1_sel[int_ip] = AA1_SELECT_JBR;
		    AALUin2_sel[int_ip] = AA2_SELECT_DISP;
		    decode->jump_taken = TRUE;
		    jumploadpc_id[int_ip] = TRUE;
		} else {
		    /* jump not taken */
		    AALUin1_sel[int_ip] = AA1_SELECT_PC;
		    AALUin2_sel[int_ip] = AA2_SELECT_NIP;
		    decode->jump_taken = FALSE;
		}
		break;
	}
}

/* Condition Code check for Jump inst. */
/* Returns TRUE(jump taken) or FALSE(jump not taken) */
int jumpCCchk(ip,cond)
int	ip;	/* ip number */
int	cond;	/* jump condition */
{
	switch (cond) {
	    case JC_NONCND:
		return TRUE;	/* jump always */
		break;
	    case JC_PLUS:
		if (CC(ip) == PLUS) return TRUE;
		else return FALSE;
		break;
	    case JC_MINUS:
		if (CC(ip) == MINUS) return TRUE;
		else return FALSE;
		break;
	    case JC_ZERO:
		if (CC(ip) == ZERO) return TRUE;
		else return FALSE;
		break;
	    case JC_NOTPLUS:
		if (CC(ip) != PLUS) return TRUE;
		else return FALSE;
		break;
	    case JC_NOTMINUS:
		if (CC(ip) != MINUS) return TRUE;
		else return FALSE;
		break;
	    case JC_NOTZERO:
		if (CC(ip) != ZERO) return TRUE;
		else return FALSE;
		break;
	    default:
		error("undefined jump condition","jumpCCchk",int_ip);
		break;
	}
	return FALSE;
}

void dec_cj(inst,decode) /* decode compare&jump */
int	inst;
decoded_inst	*decode;
{
	ifcancel = TRUE;
	decode->subtype1 = GETBIT(inst,27,2);
	decode->subtype2 = GETBIT(inst,24,3);
	decode->src2 = GETBIT(inst,8,8);
	alu_in2[int_ip].val = reg_read_fwd(int_ip,decode->src2);
	alu_in2[int_ip].regsrc = TRUE;
	alu_in2[int_ip].src = decode->src2;
	switch (decode->subtype2) {
	    case CJO_SUB:
		decode->alu_operation = OP_SUB;
		decode->src1 = GETBIT(inst,16,8);
		alu_in1[int_ip].val = reg_read_fwd(int_ip,decode->src1);
		alu_in1[int_ip].regsrc = TRUE;
		alu_in1[int_ip].src = decode->src1;
		break;
	    case CJO_RSUB:
		decode->alu_operation = OP_RSUB;
		decode->src1 = GETBIT(inst,16,8);
		alu_in1[int_ip].val = reg_read_fwd(int_ip,decode->src1);
		alu_in1[int_ip].regsrc = TRUE;
		alu_in1[int_ip].src = decode->src1;
		break;
	    case CJO_AND:
		decode->alu_operation = OP_AND;
		decode->src1 = GETBIT(inst,16,8);
		alu_in1[int_ip].val = reg_read_fwd(int_ip,decode->src1);
		alu_in1[int_ip].regsrc = TRUE;
		alu_in1[int_ip].src = decode->src1;
		break;
	    case CJO_OR:
		decode->alu_operation = OP_OR;
		decode->src1 = GETBIT(inst,16,8);
		alu_in1[int_ip].val = reg_read_fwd(int_ip,decode->src1);
		alu_in1[int_ip].regsrc = TRUE;
		alu_in1[int_ip].src = decode->src1;
		break;
	    case CJO_IMSUB:
		decode->alu_operation = OP_SUB;
		decode->src1 = GETBIT(inst,16,8);
		if (GETBIT(decode->src1,7,1) == 1) {
		    decode->src1 |= ~0xff;
		}   /* sign expansion */
		alu_in1[int_ip].val.i = decode->src1;
		alu_in1[int_ip].regsrc = FALSE;
		alu_in1[int_ip].src = decode->src1;
		break;
	    case CJO_IMRSUB:
		decode->alu_operation = OP_RSUB;
		decode->src1 = GETBIT(inst,16,8);
		if (GETBIT(decode->src1,7,1) == 1) {
		    decode->src1 |= ~0xff;
		}   /* sign expansion */
		alu_in1[int_ip].val.i = decode->src1;
		alu_in1[int_ip].regsrc = FALSE;
		alu_in1[int_ip].src = decode->src1;
		break;
	    case CJO_IMAND:
		decode->alu_operation = OP_AND;
		decode->src1 = GETBIT(inst,16,8);
		if (GETBIT(decode->src1,7,1) == 1) {
		    decode->src1 |= ~0xff;
		}   /* sign expansion */
		alu_in1[int_ip].val.i = decode->src1;
		alu_in1[int_ip].regsrc = FALSE;
		alu_in1[int_ip].src = decode->src1;
		break;
	    case CJO_IMOR:
		decode->alu_operation = OP_OR;
		decode->src1 = GETBIT(inst,16,8);
		if (GETBIT(decode->src1,7,1) == 1) {
		    decode->src1 |= ~0xff;
		}   /* sign expansion */
		alu_in1[int_ip].val.i = decode->src1;
		alu_in1[int_ip].regsrc = FALSE;
		alu_in1[int_ip].src = decode->src1;
		break;
	}
	decode->dstn = inst & 0xff;
	if (decode->dstn >> 7 == 1) {
	    decode->dstn |= ~0xff;
	}   /* sign expansion */
	decode->dstn <<= PADSFT;
	jump_disp[int_ip] = decode->dstn;
}

#ifndef NOVECTOR
void dec_v(inst,decode) /* decode vector */
int	inst;
decoded_inst	*decode;
{
	decode->subtype1 = GETBIT(inst,27,2);
	decode->alu_operation = NO_OPR;
	switch(decode->subtype1) {
	    case V_CALC:
		decode->vmr = GETBIT(inst,19,3);
		decode->subtype2 = GETBIT(inst,22,4);
		switch(decode->subtype2) {
		    case VOP_FADD:
		    case VOP_FSUB:
		    case VOP_FMULT:
		    case VOP_FDIV:
		    case VOP_IADD:
		    case VOP_ISUB:
		    case VOP_IMULT:
		    case VOP_IDIV:
		    case VOP_LAND:
		    case VOP_LOR:
		    case VOP_LXOR:
		    case VOP_LNOT:
			if (GETBIT(inst,18,1) == 0) {
			    /* chaining OFF */
			    decode->subtype3 = CHAIN_OFF;
			    decode->src1 = GETBIT(inst,12,6);
			    decode->src2 = GETBIT(inst,6,6);
			    decode->dstn = GETBIT(inst,0,6);
			} else { 
			    /* chaining ON */
			    decode->subtype3 = GETBIT(inst,15,3);
			    switch(decode->subtype3) {
				case CHAIN_RRC:
				    decode->src1 = GETBIT(inst,9,6);
				    decode->src2 = GETBIT(inst,3,6);
				    decode->dstn = GETBIT(inst,0,3);
				    break;
				case CHAIN_RCR:
				    decode->src1 = GETBIT(inst,9,6);
				    decode->src2 = GETBIT(inst,6,3);
				    decode->dstn = GETBIT(inst,0,6);
				    break;
				case CHAIN_RCC:
				    decode->src1 = GETBIT(inst,9,6);
				    decode->src2 = GETBIT(inst,6,3);
				    decode->dstn = GETBIT(inst,0,3);
				    break;
				case CHAIN_CRR:
				    decode->src1 = GETBIT(inst,12,3);
				    decode->src2 = GETBIT(inst,6,6);
				    decode->dstn = GETBIT(inst,0,6);
				    break;
				case CHAIN_CRC:
				    decode->src1 = GETBIT(inst,12,3);
				    decode->src2 = GETBIT(inst,6,6);
				    decode->dstn = GETBIT(inst,0,3);
				    break;
				case CHAIN_CCR:
				    decode->src1 = GETBIT(inst,12,3);
				    decode->src2 = GETBIT(inst,6,3);
				    decode->dstn = GETBIT(inst,0,6);
				    break;
				case CHAIN_CCC:
				    decode->src1 = GETBIT(inst,12,3);
				    decode->src2 = GETBIT(inst,6,3);
				    decode->dstn = GETBIT(inst,0,3);
				    break;
				default:
				    error("undefined chaining instruction",
					  "decode phase 1",int_ip);
				    break;
			    }
			}
			break;
		    case VOP_BCAST:
		    case VOP_VSRL:
		    case VOP_VSLL:
			decode->src1 = GETBIT(inst,12,7);
			decode->src2 = GETBIT(inst,6,6);
			decode->dstn = GETBIT(inst,0,6);
			/* alu_in1 = source scalar reg. */
			alu_in1[int_ip].val = reg_read_fwd(int_ip,decode->src1);
			alu_in1[int_ip].regsrc = TRUE;
			alu_in1[int_ip].src = decode->src1;
			alu_in1[int_ip].regdst = FALSE;
			break;
		    default:
			error("undefined vector calc instruction","decode phase 1",int_ip);
			break;
		}
		break; /* end of case V_CALC */
	    case V_LS: /* vector load/store */
		decode->subtype2 = GETBIT(inst,25,2);
		decode->src1 = GETBIT(inst,8,8);
		decode->src2 = GETBIT(inst,0,8);
		decode->vmr = GETBIT(inst,22,3);
		decode->dstn = GETBIT(inst,16,6);
		switch (decode->subtype2) {
		    case VOP_LDIR:	/* vector load, increment is s-reg. */
			/* alu_in1 = vload base reg. */
			alu_in1[int_ip].val.i = reg_read_fwd(int_ip,decode->src1).i;
			alu_in1[int_ip].regsrc = TRUE;
			alu_in1[int_ip].src = decode->src1;
			alu_in1[int_ip].regdst = FALSE;
			/* alu_in2 = vload increment reg. */
			alu_in2[int_ip].val.i = reg_read_fwd(int_ip,decode->src2).i;
			alu_in2[int_ip].regsrc = TRUE;
			alu_in2[int_ip].src = decode->src2;
			alu_in2[int_ip].regdst = FALSE;
			break;
		    case VOP_LDII:	/* vector load, increment is immed. */
			/* alu_in1 = vload base reg. */
			alu_in1[int_ip].val.i = reg_read_fwd(int_ip,decode->src1).i;
			alu_in1[int_ip].regsrc = TRUE;
			alu_in1[int_ip].src = decode->src1;
			alu_in1[int_ip].regdst = FALSE;
			/* alu_in2 = vload increment immed. */
			alu_in2[int_ip].val.i = decode->src2;
			alu_in2[int_ip].regsrc = FALSE;
			alu_in2[int_ip].regdst = FALSE;
			break;
		    case VOP_STIR:	/* vector store, increment is s-reg. */
			/* alu_in1 = vstore base reg. */
			alu_in1[int_ip].val.i = reg_read_fwd(int_ip,decode->src1).i;
			alu_in1[int_ip].regsrc = TRUE;
			alu_in1[int_ip].src = decode->src1;
			alu_in1[int_ip].regdst = FALSE;
			/* alu_in2 = vstore increment reg. */
			alu_in2[int_ip].val.i = reg_read_fwd(int_ip,decode->src2).i;
			alu_in2[int_ip].regsrc = TRUE;
			alu_in2[int_ip].src = decode->src2;
			alu_in2[int_ip].regdst = FALSE;
			break;
		    case VOP_STII:	/* vector store, increment is immed. */
			/* alu_in1 = vstore base reg. */
			alu_in1[int_ip].val.i = reg_read_fwd(int_ip,decode->src1).i;
			alu_in1[int_ip].regsrc = TRUE;
			alu_in1[int_ip].src = decode->src1;
			alu_in1[int_ip].regdst = FALSE;
			/* alu_in2 = vstore increment immed. */
			alu_in2[int_ip].val.i = decode->src2;
			alu_in2[int_ip].regsrc = FALSE;
			alu_in2[int_ip].regdst = FALSE;
			break;
		}	/* end of switch(decode->subtype2) */
		break;	/* end of case V_LS */
	    case V_MISC: /* misc. vector instruction */
		decode->subtype2 = GETBIT(inst,23,3);
		switch (decode->subtype2) {
		    case VOP_VMROP: /* VMR operation */
			decode->subtype3 = GETBIT(inst,9,2);
			decode->src1 = GETBIT(inst,6,3);
			decode->src2 = GETBIT(inst,3,3);
			decode->dstn = GETBIT(inst,0,3);
			break;
		    case VOP_VMRSET: /* VMR set */
			decode->subtype3 = GETBIT(inst,9,4);
			decode->src1 = GETBIT(inst,3,6);
			decode->dstn = GETBIT(inst,0,3);
			break;
		    case VOP_LVLS: /* list vector load/store */
			/* load bit check (bit12=0->load;1->store) */
			if (GETBIT(inst,12,1)==0) {
			    /* list vector load */
			    decode->subtype3 = VOP_LVLD;
			    decode->src1 = GETBIT(inst,6,6);
			    decode->vmr  = GETBIT(inst,13,3);
			    decode->dstn = GETBIT(inst,0,6);
			} else {
			    /* list vector store */
			    decode->subtype3 = VOP_LVST;
			    decode->src1 = GETBIT(inst,6,6);
			    decode->vmr  = GETBIT(inst,13,3);
			    decode->dstn = GETBIT(inst,0,6);
			}
			break;
		    case VOP_LVBLS: /* based list vector load/store */
			/* load bit check (bit12=0->load;1->store) */
			if (GETBIT(inst,12,1)==0) {
			    /* list vector load with base S-reg. */
			    decode->subtype3 = VOP_LVBLD;
			    decode->src1 = GETBIT(inst,6,6);
			    decode->src2 = GETBIT(inst,13,8);
			    decode->dstn = GETBIT(inst,0,6);
			    /* alu_in1 = lv load base reg. */
			    alu_in1[int_ip].val.i = reg_read_fwd(int_ip,decode->src2).i;
			    alu_in1[int_ip].regsrc = TRUE;
			    alu_in1[int_ip].src = decode->src2;
			} else {
			    /* list vector store with base S-reg. */
			    decode->subtype3 = VOP_LVBST;
			    decode->src1 = GETBIT(inst,6,6);
			    decode->src2 = GETBIT(inst,13,8);
			    decode->dstn = GETBIT(inst,0,6);
			    /* alu_in1 = lv store base reg. */
			    alu_in1[int_ip].val.i = reg_read_fwd(int_ip,decode->src2).i;
			    alu_in1[int_ip].regsrc = TRUE;
			    alu_in1[int_ip].src = decode->src2;
			}
			break;
		    case VOP_CONV: /* int/float conversion */
			decode->subtype3 = GETBIT(inst,12,1);
			decode->src1 = GETBIT(inst,6,6);
			decode->vmr  = GETBIT(inst,13,3);
			decode->dstn = GETBIT(inst,0,6);
			break;
		    case VOP_MOVES2V:	/* move scalar to vector */
			decode->src1 = GETBIT(inst,0,8);
			decode->src2 = GETBIT(inst,8,8);
			decode->dstn = GETBIT(inst,16,6);
			/* alu_in1 = source scalar reg. */
			alu_in1[int_ip].val = reg_read_fwd(int_ip,decode->src1);
			alu_in1[int_ip].regsrc = TRUE;
			alu_in1[int_ip].src = decode->src1;
			alu_in1[int_ip].regdst = FALSE;
			/* alu_in2 = VR pointer reg. */
			alu_in2[int_ip].val.i = reg_read_fwd(int_ip,decode->src2).i;
			alu_in2[int_ip].regsrc = TRUE;
			alu_in2[int_ip].src = decode->src2;
			alu_in2[int_ip].regdst = FALSE;
			break;
		    case VOP_MOVEV2S:	/* move vector to scalar */
			decode->src1 = GETBIT(inst,0,8);
			decode->src2 = GETBIT(inst,8,8);
			decode->dstn = GETBIT(inst,16,6);
			/* alu_in1 = destination scalar reg. */
			alu_in1[int_ip].regsrc = FALSE;
			alu_in1[int_ip].regdst = TRUE;
			alu_in1[int_ip].dst = decode->src1;
			/* alu_in2 = VR pointer reg. */
			alu_in2[int_ip].val.i = reg_read_fwd(int_ip,decode->src2).i;
			alu_in2[int_ip].regsrc = TRUE;
			alu_in2[int_ip].src = decode->src2;
			alu_in2[int_ip].regdst = FALSE;
			break;
		    default:
			error("undefined vector misc. instruction",
			      "decode phase 1",int_ip);
			break;
		}
		break; /* end of case V_MISC: */
	    default:
		error("undefined vector instruction","decode phase 1",int_ip);
		break;
	}
}
#endif /* end of #ifndef NOVECTOR */
