/* pipe-sim(+vector) ver.4.0 [94/05/30-06/15] */
/*		     ver.4.3a[95/02/04] */
/* printout decoded instruction */

#include <stdio.h>
#include <strings.h>

#include "defines.h"
#include "latches.h"
#include "cmnvar.h"
#include "opcode.h"
#include "opstr.h"

static char 	printout_inst[SSLEN];
static int 	int_ip;

char *printinst(inst,ip)
decoded_inst	inst;
int	ip;
{
	void print_ls(),print_ic(),print_fc(),print_jp(),print_cj(),print_v();

	int_ip = ip;
	switch (inst.optype) {
		case LOAD:
		case STORE:
			print_ls(inst);
			break;
		case INTCALC:
			print_ic(inst);
			break;
		case FLOATCALC:
			print_fc(inst);
			break;
		case JUMP:
			print_jp(inst);
			break;
		case COMPJUMP:
			print_cj(inst);
			break;
		case VECTOR:
			print_v(inst);
			break;
		default:
			error("undefined instruction","print instruction",ip);
			break;
	}

	return (printout_inst);
}

void print_ls(inst) /* print load/store instruction */
decoded_inst	inst;
{
	switch (inst.subtype1) {
	    case AD_IMMED:
	    case AD_DIRECT:
	    case AD_PCBASED:
		sprintf(printout_inst,
			"addr=0x%X:%s %s src1=%d dstn=Reg[%d]",
			inst.address,
			optype_s[inst.optype],
			addrmode_s[inst.subtype1],
			inst.src1,
			inst.dstn);
		if ((inst.optype==LOAD)&&(inst.subtype1==AD_IMMED)&&
		    (inst.src1 == 0)&&(inst.dstn == 0)) { 
			/* no operation (LDI 0,R0) */
			sprintf(printout_inst,
				"addr=0x%X:NOP (%s %s src1=%d dstn=Reg[%d])",
				inst.address,
				optype_s[inst.optype],
				addrmode_s[inst.subtype1],
				inst.src1,
				inst.dstn);
		}
		break;
	    case AD_BASED:
	    case AD_BASESFT:
		sprintf(printout_inst,
			"addr=0x%X:%s %s src1=%d src2=Reg[%d] dstn=Reg[%d]",
			inst.address,
			optype_s[inst.optype],
			addrmode_s[inst.subtype1],
			inst.src1,
			inst.src2,
			inst.dstn);
		break;
	    case AD_BASEINDX:
		sprintf(printout_inst,
			"addr=0x%X:%s %s src1=Reg[%d] src2=Reg[%d] dstn=Reg[%d]",
			inst.address,
			optype_s[inst.optype],
			addrmode_s[inst.subtype1],
			inst.src1,
			inst.src2,
			inst.dstn);
		break;
	    case AD_PCBINDX:
		sprintf(printout_inst,
			"addr=0x%X:%s %s src1=Reg[%d] dstn=Reg[%d]",
			inst.address,
			optype_s[inst.optype],
			addrmode_s[inst.subtype1],
			inst.src1,
			inst.dstn);
		break;
	default:
		error("undefined load/store instruction","print instruction",int_ip);
		break;
	}
}

void print_ic(inst) /* print intcalc */
decoded_inst	inst;
{
	char tmpstr[100];

	switch (inst.subtype1) {
	    case REGOPR:
		sprintf(printout_inst,
			"addr=0x%X:%s %s src1=Reg[%d] src2=Reg[%d] dstn=Reg[%d]",
			inst.address,
			optype_s[inst.optype],
			aluop_s[inst.subtype2],
			inst.src1,
			inst.src2,
			inst.dstn);
		break;
	    case IMMOPR:
		strcpy (tmpstr,aluop_s[inst.subtype2]);
		sprintf(printout_inst,
			"addr=0x%X:%s %s src1=%d src2=Reg[%d] dstn=Reg[%d]",
			inst.address,
			optype_s[inst.optype],
			strcat(tmpstr,"-IMMED"),
			inst.src1,
			inst.src2,
			inst.dstn);
		break;
	}
}

void print_fc(inst) /* print floatcalc */
decoded_inst	inst;
{
	sprintf(printout_inst,
		"addr=0x%X:%s %s src1=Reg[%d] src2=Reg[%d] dstn=Reg[%d]",
		inst.address,
		optype_s[inst.optype],
		fop_s[inst.subtype2],
		inst.src1,
		inst.src2,
		inst.dstn);
}

void print_jp(inst) /* print jump */
decoded_inst	inst;
{
	switch (inst.subtype1) {
	    case JAD_REL:
	    case JAD_REL_LK:
		sprintf(printout_inst,
			"addr=0x%X:%s %s %s src1=%d",
			inst.address,
			optype_s[inst.optype],
			jumpaddr_s[inst.subtype1],
			jumpcond_s[inst.subtype2],
			inst.src1);
	    case JAD_BASED:
	    case JAD_BASED_LK:
		sprintf(printout_inst,
			"addr=0x%X:%s %s %s src1=%d src2=Reg[%d]",
			inst.address,
			optype_s[inst.optype],
			jumpaddr_s[inst.subtype1],
			jumpcond_s[inst.subtype2],
			inst.src1,
			inst.src2);
		break;
	}
}

void print_cj(inst) /* print compare&jump */
decoded_inst	inst;
{
	char tmpstr[100];

	switch (inst.subtype2) {
	    case CJO_SUB:
	    case CJO_RSUB:
	    case CJO_AND:
	    case CJO_OR:
		sprintf(printout_inst,
			"addr=0x%X:%s %s %s src1=R[%d] src2=Reg[%d] dstn=%d",
			inst.address,
			optype_s[inst.optype],
			cjcond_s[inst.subtype1],
			cjop_s[inst.subtype2],
			inst.src1,
			inst.src2,
			inst.dstn);
		break;
	    case CJO_IMSUB:
	    case CJO_IMRSUB:
	    case CJO_IMAND:
	    case CJO_IMOR:
		strcpy(tmpstr,cjop_s[inst.subtype2]);
		sprintf(printout_inst,
			"addr=0x%X:%s %s %s src1=%d src2=Reg[%d] dstn=%d",
			inst.address,
			optype_s[inst.optype],
			cjcond_s[inst.subtype1],
			strcat(tmpstr,"-IMMED"),
			inst.src1,
			inst.src2,
			inst.dstn);
		break;
	}
}

void print_v(inst) /* print vector */
decoded_inst	inst;
{
	switch(inst.subtype1) {
	    case V_CALC:
		switch(inst.subtype2) {
		    case VOP_FADD:
		    case VOP_FSUB:
		    case VOP_FMULT:
		    case VOP_FDIV:
		    case VOP_IADD:
		    case VOP_ISUB:
		    case VOP_IMULT:
		    case VOP_IDIV:
		    case VOP_LAND:
		    case VOP_LOR:
		    case VOP_LXOR:
		    case VOP_LNOT:
			if (inst.subtype3 == CHAIN_OFF) {
			    /* chaining OFF */
			    sprintf(printout_inst,
				    "addr=0x%X:%s %s %s src1=VR[%d] src2=VR[%d] dstn=VR[%d] VMR=%d",
				    inst.address,
				    optype_s[inst.optype],
				    vop_s[inst.subtype1],
				    vcalcop_s[inst.subtype2],
				    inst.src1,
				    inst.src2,
				    inst.dstn,
				    inst.vmr);
			} else { 
			    /* chaining ON */
			    switch(inst.subtype3) {
				case CHAIN_RRC:
				    sprintf(printout_inst,
					    "addr=0x%X:%s %s %s src1=VR[%d] src2=VR[%d] dstn=CN[%d] VMR=%d",
					    inst.address,
					    optype_s[inst.optype],
					    vop_s[inst.subtype1],
					    vcalcop_s[inst.subtype2],
					    inst.src1,
					    inst.src2,
					    inst.dstn,
					    inst.vmr);
				    break;
				case CHAIN_RCR:
				    sprintf(printout_inst,
					    "addr=0x%X:%s %s %s src1=VR[%d] src2=CN[%d] dstn=VR[%d] VMR=%d",
					    inst.address,
					    optype_s[inst.optype],
					    vop_s[inst.subtype1],
					    vcalcop_s[inst.subtype2],
					    inst.src1,
					    inst.src2,
					    inst.dstn,
					    inst.vmr);
				    break;
				case CHAIN_RCC:
				    sprintf(printout_inst,
					    "addr=0x%X:%s %s %s src1=VR[%d] src2=CN[%d] dstn=CN[%d] VMR=%d",
					    inst.address,
					    optype_s[inst.optype],
					    vop_s[inst.subtype1],
					    vcalcop_s[inst.subtype2],
					    inst.src1,
					    inst.src2,
					    inst.dstn,
					    inst.vmr);
				    break;
				case CHAIN_CRR:
				    sprintf(printout_inst,
					    "addr=0x%X:%s %s %s src1=CN[%d] src2=VR[%d] dstn=VR[%d] VMR=%d",
					    inst.address,
					    optype_s[inst.optype],
					    vop_s[inst.subtype1],
					    vcalcop_s[inst.subtype2],
					    inst.src1,
					    inst.src2,
					    inst.dstn,
					    inst.vmr);
				    break;
				case CHAIN_CRC:
				    sprintf(printout_inst,
					    "addr=0x%X:%s %s %s src1=CN[%d] src2=VR[%d] dstn=CN[%d] VMR=%d",
					    inst.address,
					    optype_s[inst.optype],
					    vop_s[inst.subtype1],
					    vcalcop_s[inst.subtype2],
					    inst.src1,
					    inst.src2,
					    inst.dstn,
					    inst.vmr);
				    break;
				case CHAIN_CCR:
				    sprintf(printout_inst,
					    "addr=0x%X:%s %s %s src1=CN[%d] src2=CN[%d] dstn=VR[%d] VMR=%d",
					    inst.address,
					    optype_s[inst.optype],
					    vop_s[inst.subtype1],
					    vcalcop_s[inst.subtype2],
					    inst.src1,
					    inst.src2,
					    inst.dstn,
					    inst.vmr);
				    break;
				case CHAIN_CCC:
				    sprintf(printout_inst,
					    "addr=0x%X:%s %s %s src1=CN[%d] src2=CN[%d] dstn=CN[%d] VMR=%d",
					    inst.address,
					    optype_s[inst.optype],
					    vop_s[inst.subtype1],
					    vcalcop_s[inst.subtype2],
					    inst.src1,
					    inst.src2,
					    inst.dstn,
					    inst.vmr);
				    break;
				default:
				    error("undefined chaining instruction",
					  "print instruction",int_ip);
				    break;
			    }
			}
			break;
		    case VOP_BCAST:
		    case VOP_VSRL:
		    case VOP_VSLL:
			sprintf(printout_inst,
				"addr=0x%X:%s %s %s src1=Reg[%d] src2=VR[%d] dstn=VR[%d] VMR=%d",
				inst.address,
				optype_s[inst.optype],
				vop_s[inst.subtype1],
				vcalcop_s[inst.subtype2],
				inst.src1,
				inst.src2,
				inst.dstn,
				inst.vmr);
			break;
		    default:
			error("undefined vector calc instruction","print instruction",int_ip);
			break;
		}
		break; /* end of case V_CALC */
	    case V_LS: /* vector load/store */
		switch(inst.subtype2) {
		    case VOP_LDIR:
		    case VOP_STIR:
			sprintf(printout_inst,
				"addr=0x%X:%s %s %s src1(base)=Reg[%d] src2(incr)=Reg[%d] dstn=VR[%d] VMR=%d",
				inst.address,
				optype_s[inst.optype],
				vop_s[inst.subtype1],
				vlsop_s[inst.subtype2],
				inst.src1,
				inst.src2,
				inst.dstn,
				inst.vmr);
			break;
		    case VOP_LDII:
		    case VOP_STII:
			sprintf(printout_inst,
				"addr=0x%X:%s %s %s src1(base)=Reg[%d] src2(incr)=%d dstn=VR[%d] VMR=%d",
				inst.address,
				optype_s[inst.optype],
				vop_s[inst.subtype1],
				vlsop_s[inst.subtype2],
				inst.src1,
				inst.src2,
				inst.dstn,
				inst.vmr);
			break;
		}
		break;
	    case V_MISC: /* misc. vector instruction */
		switch (inst.subtype2) {
		    case VOP_VMROP: /* VMR operation */
			sprintf(printout_inst,
				"addr=0x%X:%s %s %s %s src1=VMR[%d] src2=VMR[%d] dstn=VMR[%d]",
				inst.address,
				optype_s[inst.optype],
				vop_s[inst.subtype1],
				vmisc_s[inst.subtype2],
				vmrop_s[inst.subtype3],
				inst.src1,
				inst.src2,
				inst.dstn);
			break;
		    case VOP_VMRSET: /* VMR set */
			sprintf(printout_inst,
				"addr=0x%X:%s %s %s %s src1=VR[%d] dstn=VMR[%d]",
				inst.address,
				optype_s[inst.optype],
				vop_s[inst.subtype1],
				vmisc_s[inst.subtype2],
				vmrsetcond_s[inst.subtype3],
				inst.src1,
				inst.dstn);
			break;
		    case VOP_LVLS: /* list vector load/store */
			sprintf(printout_inst,
				"addr=0x%X:%s %s %s %s src1(listvec)=VR[%d] dstn=VR[%d] VMR=%d",
				inst.address,
				optype_s[inst.optype],
				vop_s[inst.subtype1],
				vmisc_s[inst.subtype2],
				vlsop_s[inst.subtype3],
				inst.src1,
				inst.dstn,
				inst.vmr);
			break;
		    case VOP_LVBLS: /* based list vector load/store */
			sprintf(printout_inst,
				"addr=0x%X:%s %s %s %s src1(listvec)=VR[%d] src2(base)=Reg[%d] dstn=VR[%d]",
				inst.address,
				optype_s[inst.optype],
				vop_s[inst.subtype1],
				vmisc_s[inst.subtype2],
				vlsop_s[inst.subtype3],
				inst.src1,
				inst.src2,
				inst.dstn);
			break;
		    case VOP_CONV: /* int/float conversion */
			sprintf(printout_inst,
				"addr=0x%X:%s %s %s %s src1=VR[%d] dstn=VR[%d] VMR=%d",
				inst.address,
				optype_s[inst.optype],
				vop_s[inst.subtype1],
				vmisc_s[inst.subtype2],
				vconv_s[inst.subtype3],
				inst.src1,
				inst.dstn,
				inst.vmr);
			break;
		    case VOP_MOVES2V:	/* move scalar to vector */
		    case VOP_MOVEV2S:	/* move vector to scalar */
			sprintf(printout_inst,
				"addr=0x%X:%s %s %s src1=Reg[%d] src2(VR-pos)=Reg[%d] dstn=VR[%d]",
				inst.address,
				optype_s[inst.optype],
				vop_s[inst.subtype1],
				vmisc_s[inst.subtype2],
				inst.src1,
				inst.src2,
				inst.dstn);
			break;
		    default:
			error("undefined vector misc. instruction",
			      "print instruction",int_ip);
			break;
		}
		break; /* end of case V_MISC: */
	    default:
		error("undefined vector instruction","print instruction",int_ip);
		break;
	}
}
