/* pipe-sim(+vector) ver.4.0 [94/05/30-06/15] */
/* 		     ver.4.0a[94/06/16-06/] */
/* 		     ver.4.3 [94/12/26-95/01/20] */
/* pipeline MEM stage */

#include "defines.h"
#include "latches.h"
#include "vectorls.h"
#include "cmnvar.h"
#include "opcode.h"
#include "extfunc.h"
#include "regdef.h"
#include "register_set.h"

static decoded_inst i_latch[NIP];

static char mem_0_status[NIP][SSLEN];
static char mem_1_status[NIP][SSLEN];

static reglatch mem_internal_latch[NIP];

char *mem_0(ip)
int	ip;	/* ip = instruction pipe number */
{
	void vectorcalc_01(),vectorls_0(),vectormisc_0();

	i_latch[ip] = memlatch[ip];

	switch (i_latch[ip].optype) {
	    case LOAD:
		switch (i_latch[ip].subtype1) {
		    case AD_IMMED:
			mem_internal_latch[ip] = alu_out[ip];
			/* return status */
			sprintf(mem_0_status[ip],
				"alu_out %d set to mem_int_latch",
				alu_out[ip].val.i);
			break;
		    case AD_DIRECT:
		    case AD_BASED:
		    case AD_BASEINDX:
		    case AD_BASESFT:
		    case AD_PCBASED:
		    case AD_PCBINDX:
			loadaddr[ip] = alu_out[ip].val.i;
			loadflag[ip] = TRUE;
			/* return status */
			sprintf(mem_0_status[ip],
				"alu_out %d set to loadaddr latch",
				alu_out[ip].val.i);
			break;
		    default:
			error("undefined instruction(load irregal addr. mode)",
			      "mem phase 0",ip);
			break;
		}
		break;
	    case STORE:
		switch (i_latch[ip].subtype1) {
		    case AD_DIRECT:
		    case AD_BASED:
		    case AD_BASEINDX:
		    case AD_BASESFT:
		    case AD_PCBASED:
		    case AD_PCBINDX:
			storedata[ip] = reg_read(ip,i_latch[ip].dstn);
			storeaddr[ip] = alu_out[ip].val.i;
			storeflag[ip] = TRUE;
			/* return status */
			sprintf(mem_0_status[ip],
				"alu_out %d and SR[%d] set to store latch",
				alu_out[ip].val.i, i_latch[ip].dstn);
			break;
		    case AD_IMMED:
		    default:
			error("undefined instruction(store irregal addr. mode)",
			      "mem phase 0",ip);
			break;
		}
		break;
	    case INTCALC:
		switch (i_latch[ip].subtype2) {
		    case OP_ADD:
		    case OP_SUB:
		    case OP_MUL:
		    case OP_DIV:
		    case OP_MOD:
		    case OP_RSUB:
		    case OP_UMUL:
		    case OP_UDIV:
		    case OP_AND:
		    case OP_OR:
		    case OP_XOR:
		    case OP_NOT:
		    case OP_SRL:
		    case OP_SLL:
		    case OP_SRA:
		    case OP_SLA:
			mem_internal_latch[ip] = alu_out[ip];
			/* return status */
			sprintf(mem_0_status[ip],
				"alu_out %d set to mem_int_latch",
				alu_out[ip].val.i);
			break;
		    default:
			error("undefined intcalc instruction",
			      "exec1 phase 0",ip);
			break;
		}
		break;
	    case FLOATCALC: /* do nothing */
		/* return status */
		sprintf(mem_0_status[ip], "no operation ");
		break;
	    case JUMP:
		switch (i_latch[ip].subtype1) {
		    case JAD_REL:
		    case JAD_BASED: /* no operation */
			/* return status */
			sprintf(mem_0_status[ip], "no operation ");
			break;
		    case JAD_REL_LK:
		    case JAD_BASED_LK:
			if (i_latch[ip].jump_taken==TRUE) {
			    mem_internal_latch[ip] = alu_out[ip];
			    /* return status */
			    sprintf(mem_0_status[ip],
				    "alu_out %d set to mem_int_latch",
				    alu_out[ip].val.i);
			} else {
			    /* return status */
			    sprintf(mem_0_status[ip], "no operation ");
			}
			break;
		}
		break;
	    case COMPJUMP: /* do nothing */
		/* return status */
		sprintf(mem_0_status[ip], "no operation ");
		break;
	    case VECTOR:
                switch (i_latch[ip].subtype1) {
                    case V_CALC:
                    case V_LS:
                        /* return status */
                        sprintf(mem_0_status[ip], "no operation");
                        break;
                    case V_MISC:
                        switch (i_latch[ip].subtype2) {
                            case VOP_VMROP:
                            case VOP_VMRSET:
                            case VOP_CONV:
                            case VOP_MOVES2V:
                                /* return status */
                                sprintf(mem_0_status[ip], "no operation");
                                break;
                            case VOP_MOVEV2S:
				mem_internal_latch[ip] = alu_out[ip];
				/* return status */
				sprintf(mem_0_status[ip],
					"alu_out %d set to mem_int_latch",
					alu_out[ip].val.i);
                                break;
                            default:
                                error("undefined vector operation(subtype2)",
                                      "mem phase 0",ip);
                                break;
                        }
                        break;
                    default:
                        error("undefined vector operation(subtype1)",
                              "mem phase 0",ip);
                        break;
                }
		break;
	    case NO_OPR: /* do nothing */
		/* return status */
		sprintf(mem_0_status[ip], "no operation (by NO_OPR) ");
		break;
	} /* end of switch (i_latch[ip].optype) */

	/* return status */
	/* strcpy(mem_0_status[ip],printinst(i_latch[ip],ip)); */
	return(mem_0_status[ip]);
}

char *mem_1(ip)
int	ip;
{
	void vectorcalc_01(),vectormisc_1();

	wblatch[ip] = i_latch[ip];

        /* writeback latch destination register flag initialize */
	wb_latch[ip].regdst = FALSE;

	switch (i_latch[ip].optype) {
	    case LOAD:
		if (i_latch[ip].subtype1 == AD_IMMED) {
		    wb_latch[ip] = mem_internal_latch[ip];
		    wb_latch[ip].regdst = TRUE;
		    wb_latch[ip].dst = i_latch[ip].dstn;
		    /* return status */
		    sprintf(mem_1_status[ip],
			    "mem_int_latch %d set to wb_latch",
			    mem_internal_latch[ip].val.i);
		    break;
		} else { /* get load data from loadunit data latch */
		    wb_latch[ip].val = loaddata[ip];
		    wb_latch[ip].regdst = TRUE;
		    wb_latch[ip].dst = i_latch[ip].dstn;
		    /* return status */
		    sprintf(mem_1_status[ip],
			    "loaddata %d set to wb_latch",
			    loaddata[ip].i);
		}
		break;
	    case STORE:
		if (i_latch[ip].subtype1 == AD_IMMED) {
		    /* irregal instruction */
		    error("undefined instruction","mem phase 1",ip);
		} else { /* no operation */
		    /* return status */
		    sprintf(mem_1_status[ip],
			    "no operation");
		}
		break;
	    case INTCALC:
		wb_latch[ip] = mem_internal_latch[ip];
		/* return status */
		sprintf(mem_1_status[ip],
			"mem_int_latch %d set to wb_latch",
			mem_internal_latch[ip].val.i);
		break;
	    case FLOATCALC:
		if ((ip & 1) == 0) { /* ip = even number */
		    switch (i_latch[ip].subtype2) {
			case FOP_ADD:
			case FOP_SUB:
			case FOP_FLT:
			    wb_latch[ip] = faspout[ip>>1];
			    sprintf(mem_1_status[ip],
				    "faspout %f set to wb_latch",
				    wb_latch[ip].val.f);
			    break;
			case FOP_INT:
			    wb_latch[ip] = faspout[ip>>1];
			    sprintf(mem_1_status[ip],
				    "faspout %d set to wb_latch",
				    wb_latch[ip].val.i);
			    break;
			case FOP_MLT:
			case FOP_DIV:
			    /* return status */
			    sprintf(mem_1_status[ip], "no operation");
			    break;
			default:
			    error("undefined floatcalc operation",
				  "mem phase 1",ip);
			    break;
		    }
		} else { /* ip = odd number */
		    /* return status */
		    sprintf(mem_1_status[ip], "no operation");
		}
		break; /* end of case FLOATCALC */
	    case COMPJUMP: /* no operation */
		/* return status */
		sprintf(mem_1_status[ip], "no operation");
		break;
#ifdef NOVECTOR /* vector not implemented */
	    case VECTOR: /* no operation */
		/* return status */
		sprintf(mem_1_status[ip], "no operation");
		break;
#else /* vector implemented */
	    case VECTOR:
                switch (i_latch[ip].subtype1) {
                    case V_CALC:
                    case V_LS:
                        /* return status */
                        sprintf(mem_1_status[ip], "no operation");
                        break;
                    case V_MISC:
                        switch (i_latch[ip].subtype2) {
                            case VOP_VMROP:
                            case VOP_VMRSET:
                            case VOP_CONV:
                            case VOP_MOVES2V:
                                /* return status */
                                sprintf(mem_1_status[ip], "no operation");
                                break;
                            case VOP_MOVEV2S:
				wb_latch[ip] = mem_internal_latch[ip];
				/* return status */
				sprintf(mem_1_status[ip],
					"mem_int_latch %d set to wb_latch",
					mem_internal_latch[ip].val.i);
				break;
                            default:
                                error("undefined vector operation(subtype2)",
                                      "mem phase 1",ip);
                                break;
                        }
                        break;
                    default:
                        error("undefined vector operation(subtype1)",
                              "mem phase 1",ip);
                        break;
                }
		break;
#endif
	    case JUMP:
		switch (i_latch[ip].subtype1) {
		    case JAD_REL:
		    case JAD_BASED: /* no operation */
			/* return status */
			sprintf(mem_1_status[ip], "no operation ");
			break;
		    case JAD_REL_LK:
		    case JAD_BASED_LK:
			if (i_latch[ip].jump_taken==TRUE) {
			    wb_latch[ip] = mem_internal_latch[ip];
			    /* return status */
			    sprintf(mem_1_status[ip],
				    "mem_int_latch %d set to wb_latch",
				    mem_internal_latch[ip].val.i);
			} else {
			    /* return status */
			    sprintf(mem_1_status[ip], "no operation ");
			}
			break;
		}
		break;
	    case NO_OPR: /* do nothing */
		/* return status */
		sprintf(mem_1_status[ip], "no operation (by NO_OPR) ");
		break;
	}

	/* return status */
	/* strcpy(mem_1_status[ip],printinst(i_latch[ip],ip)); */
	return(mem_1_status[ip]);
}
