/* pipe-sim(+vector) ver.3.0 [92/05/28-7/29] */
/* pipe-sim(+vector) ver.4.0 [94/05/30-06/15] */
/* define mnemonics */

#include "global.h"

/* mnemonic defines */

/* instruction type */

#define LOAD	0		/* load from memory */
#define STORE	1		/* store to memory */

#define INTCALC	2		/* integer calculation */
#define FLOATCALC	3	/* float calculation */

#define JUMP	4		/* jump */
#define COMPJUMP	5	/* compare and jump */

#define VECTOR	6	/* vector instruction */

/* special instructions */
/* no operation (= load immediate 0,R0.0) */
#define NOP	(LOAD<<29)|(AD_IMMED<<26)|(0<<8)|0
/* halt CPU (= load immediate HALT,R0.2) */
#define HLT	(LOAD<<29)|(AD_IMMED<<26)|(HALT<<8)|2

/* LOAD/STORE instruction */

/* load immediate */
#define LDI(immed,dstn)	(LOAD<<29)|(AD_IMMED<<26)|(((immed)&0x1ffff)<<8)|((dstn)&0xff)
/* load direct */
#define LD(addr,dstn)	(LOAD<<29)|(AD_DIRECT<<26)|(((addr)&0x1ffff)<<8)|((dstn)&0xff)
/* load based */
#define LDBASE(disp,base,dstn)	(LOAD<<29)|(AD_BASED<<26)|(((disp)&0x1ff)<<16)|(((base)&0xff)<<8)|((dstn)&0xff)
/* load base-indexed */
#define LDBIDX(base,indx,dstn)	(LOAD<<29)|(AD_BASEINDX<<26)|(((base)&0xff)<<16)|(((indx)&0xff)<<8)|((dstn)&0xff)
/* load base-shifted */
#define LDBSFT(base,sft,dstn)	(LOAD<<29)|(AD_BASESFT<<26)|(((sft)&0x1ff)<<16)|(((base)&0xff)<<8)|((dstn)&0xff)
/* load PC-based */
#define LDPCB(disp,dstn)	(LOAD<<29)|(AD_PCBASED<<26)|(((disp)&0x1ffff)<<8)|((dstn)&0xff)
/* load PCbase-indexed */
#define LDPCBIDX(indx,dstn)	(LOAD<<29)|(AD_PCBINDX<<26)|(((indx)&0xff)<<16)|((dstn)&0xff)

/* store direct */
#define ST(addr,dstn)	(STORE<<29)|(AD_DIRECT<<26)|(((addr)&0x1ffff)<<8)|((dstn)&0xff)
/* store based */
#define STBASE(disp,base,dstn)	(STORE<<29)|(AD_BASED<<26)|(((disp)&0x1ff)<<16)|(((base)&0xff)<<8)|((dstn)&0xff)
/* store base-indexed */
#define STBIDX(base,indx,dstn)	(STORE<<29)|(AD_BASEINDX<<26)|(((base)&0xff)<<16)|(((indx)&0xff)<<8)|((dstn)&0xff)
/* store base-shifted */
#define STBSFT(base,sft,dstn)	(STORE<<29)|(AD_BASESFT<<26)|(((sft)&0x1ff)<<16)|(((base)&0xff)<<8)|((dstn)&0xff)
/* store PC-based */
#define STPCB(disp,dstn)	(STORE<<29)|(AD_PCBASED<<26)|(((disp)&0x1ffff)<<8)|((dstn)&0xff)
/* store PCbase-indexed */
#define STPCBIDX(indx,dstn)	(STORE<<29)|(AD_PCBINDX<<26)|(((indx)&0xff)<<16)|((dstn)&0xff)

/* ALU operation */
#define ADD(src1,src2,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_ADD<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define ADDI(immed,src2,dstn)	(INTCALC<<29)|(IMMOPR<<28)|(OP_ADD<<24)|(((immed)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define SUB(src1,src2,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_SUB<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define SUBI(immed,src2,dstn)	(INTCALC<<29)|(IMMOPR<<28)|(OP_SUB<<24)|(((immed)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define MUL(src1,src2,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_MUL<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define MULI(immed,src2,dstn)	(INTCALC<<29)|(IMMOPR<<28)|(OP_MUL<<24)|(((immed)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define DIV(src1,src2,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_DIV<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define DIVI(immed,src2,dstn)	(INTCALC<<29)|(IMMOPR<<28)|(OP_DIV<<24)|(((immed)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define MOD(src1,src2,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_MOD<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define MODI(immed,src2,dstn)	(INTCALC<<29)|(IMMOPR<<28)|(OP_MOD<<24)|(((immed)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define RSUB(src1,src2,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_RSUB<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define RSUBI(immed,src2,dstn)	(INTCALC<<29)|(IMMOPR<<28)|(OP_RSUB<<24)|(((immed)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define UMUL(src1,src2,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_UMUL<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define UMULI(immed,src2,dstn)	(INTCALC<<29)|(IMMOPR<<28)|(OP_UMUL<<24)|(((immed)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define UDIV(src1,src2,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_UDIV<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define UDIVI(immed,src2,dstn)	(INTCALC<<29)|(IMMOPR<<28)|(OP_UDIV<<24)|(((immed)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define AND(src1,src2,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_AND<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define ANDI(immed,src2,dstn)	(INTCALC<<29)|(IMMOPR<<28)|(OP_AND<<24)|(((immed)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define OR(src1,src2,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_OR<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define ORI(immed,src2,dstn)	(INTCALC<<29)|(IMMOPR<<28)|(OP_OR<<24)|(((immed)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define XOR(src1,src2,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_XOR<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define XORI(immed,src2,dstn)	(INTCALC<<29)|(IMMOPR<<28)|(OP_XOR<<24)|(((immed)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define NOT(src1,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_NOT<<24)|(((src1)&0xff)<<16)|((dstn)&0xff)
#define SRL(src1,src2,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_SRL<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define SRLI(immed,src2,dstn)	(INTCALC<<29)|(IMMOPR<<28)|(OP_SRL<<24)|(((immed)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define SLL(src1,src2,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_SLL<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define SLLI(immed,src2,dstn)	(INTCALC<<29)|(IMMOPR<<28)|(OP_SLL<<24)|(((immed)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define SRA(src1,src2,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_SRA<<24)|((src1&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define SRAI(immed,src2,dstn)	(INTCALC<<29)|(IMMOPR<<28)|(OP_SRA<<24)|(((immed)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define SLA(src1,src2,dstn)	(INTCALC<<29)|(REGOPR<<28)|(OP_SLA<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define SLAI(immed,src2,dstn)	(INTCALC<<29)|(IMMOPR<<28)|(OP_SLA<<24)|(((immed)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)

/* float operation */
#define FADD(src1,src2,dstn)	(FLOATCALC<<29)|(REGOPR<<28)|(FOP_ADD<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define FSUB(src1,src2,dstn)	(FLOATCALC<<29)|(REGOPR<<28)|(FOP_SUB<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define FMLT(src1,src2,dstn)	(FLOATCALC<<29)|(REGOPR<<28)|(FOP_MLT<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)
#define FDIV(src1,src2,dstn)	(FLOATCALC<<29)|(REGOPR<<28)|(FOP_DIV<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|((dstn)&0xff)

/* float to int conversion */
#define CONVINT(src1,dstn)	(FLOATCALC<<29)|(REGOPR<<28)|(FOP_INT<<24)|(((src1)&0xff)<<16)|((dstn)&0xff)
/* int to float conversion */
#define CONVFLOAT(src1,dstn)	(FLOATCALC<<29)|(REGOPR<<28)|(FOP_FLT<<24)|(((src1)&0xff)<<16)|((dstn)&0xff)

/* jump instructions */
/* jump direct */
#define JMP(addr)	(JUMP<<29)|(JAD_DIRECT<<27)|(JC_NONCND<<24)|(((addr)>>PADSFT)&0xffffff)
#define JC(cond,addr)	(JUMP<<29)|(JAD_DIRECT<<27)|(cond<<24)|(((addr)>>PADSFT)&0xffffff)
/* jump direct-based */
#define JMPBASE(base,disp)	(JUMP<<29)|(JAD_DIBASE<<27)|(JC_NONCND<<24)|(((base)&0xff)<<16)|(((disp)>>PADSFT)&0xffff)
#define JCBASE(cond,base,disp)	(JUMP<<29)|(JAD_DIBASE<<27)|(cond<<24)|(((base)&0xff)<<16)|(((disp)>>PADSFT)&0xffff)
/* relative jump (direct-based,base-reg=PC(R0.1)) */
#define JRBASE(disp)	(JUMP<<29)|(JAD_DIBASE<<27)|(JC_NONCND<<24)|((1&0xff)<<16)|(((disp)>>PADSFT)&0xffff)
#define JRC(cond,disp)	(JUMP<<29)|(JAD_DIBASE<<27)|(cond<<24)|((1&0xff)<<16)|(((disp)>>PADSFT)&0xffff)
/* jump indirect */
#define JMPIND(addr)	(JUMP<<29)|(JAD_INDIR<<27)|(JC_NONCND<<24)|((addr)&0xffffff)
#define JCIND(cond,addr)	(JUMP<<29)|(JAD_INDIR<<27)|(cond<<24)|((addr)&0xffffff)
/* jump indirect-based */
#define JMPINDBASE(base,disp)	(JUMP<<29)|(JAD_INDBASE<<27)|(JC_NONCND<<24)|(((base)&0xff)<<16)|((disp)&0xffff)
#define JCINDBASE(cond,base,disp)	(JUMP<<29)|(JAD_INDBASE<<27)|(cond<<24)|(((base)&0xff)<<16)|((disp)&0xffff)

/* compare&jump instruction */
#define CJPLUS(oper,src1,src2,disp)	(COMPJUMP<<29)|(CJC_PLUS<<27)|(oper<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|(((disp)>>PADSFT)&0xff)
#define CJMINUS(oper,src1,src2,disp)	(COMPJUMP<<29)|(CJC_MINUS<<27)|(oper<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|(((disp)>>PADSFT)&0xff)
#define CJZERO(oper,src1,src2,disp)	(COMPJUMP<<29)|(CJC_ZERO<<27)|(oper<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|(((disp)>>PADSFT)&0xff)
#define CJNONZERO(oper,src1,src2,disp)	(COMPJUMP<<29)|(CJC_NONZERO<<27)|(oper<<24)|(((src1)&0xff)<<16)|(((src2)&0xff)<<8)|(((disp)>>PADSFT)&0xff)
