/***
 ***	read.c:	read data from register file
 ***		for banked register set
 ***	from Oct.4 1993 to Oct.7 1993 by take.
 ***	
 ***	arranged for ps3v ver.3.5a-R1.0 [93/Oct/25] by kata.
 ***		  	  ver.3.5a-R1.0b[93/Oct/27] by kata.
 ***	
 ***	arranged for ps4v ver.4.0 [94/May/30-Jun/15] by kata.
 ***			  ver.4.0a[94/Jun16-Jun/] by kata.
 ***			  ver.4.1 [94/Dec03-] by kata.
 ***			  ver.4.2 [94/Dec19-] by kata.
 ***/

#include <stdio.h>
#include <stdlib.h>
#include "defines.h"
#include "cmnvar.h"
#include "regdef.h"
#include "register_set.h"
/* register number is from 0 to 127;
   if global registers are 4, their number is from 0 to 3.
   if global registers are 8, their number is from 0 to 8.
*/

static char errstr[100]; /* error message temp. */

#ifdef BANKED
word reg_read (int ip, int no)
{
	word r_value;	/* define the variable for return value */
	int  local_no;
	int  ipflag;

	ipflag = ip;	/* recognize the IP number */

	if (no < 0 || no >= REGSIZE)
	{
		sprintf (errstr,"invalid register No.%d",no);
		error(errstr,"reg_read",ip);
		return;
	}

	if (no >= SPECIALBASE && no < REGSIZE) {
		/* special register */
		r_value = specialregfile[no-SPECIALBASE];
	} else if (no < GLOBAL) {
	/* register number 0-7 */
		global_check(ip,no);
		r_value = global_register[no];
	}
	else
	{
		switch (ipflag)
		{
			case 0:
			/* register number 8-37 */
				local_no = no - (30 * ipflag + GLOBAL);
				if (local_no >= OVERLAP && local_no < OVERLAP+LOCAL) {
					local_no = local_no - OVERLAP;
					banked_check(ip,local_no);
					r_value = banked_registers_0[local_no];
				}
				else if (local_no < OVERLAP)
				{
					overlap_check(ip,local_no);
					r_value = overlap_registers_0[local_no];
				}
				else
				{
					local_no = local_no - (LOCAL + OVERLAP);
					overlap_check(ip,local_no);
					r_value = overlap_registers_1[local_no];
				}
				break;
			case 1:
			/* register number 38-67 */
				local_no = no - (30 * ipflag + GLOBAL);
				if (local_no >= OVERLAP && local_no < OVERLAP+LOCAL) {
					local_no = local_no - OVERLAP;
					banked_check(ip,local_no);
					r_value = banked_registers_1[local_no];
				}
				else if (local_no < OVERLAP)
				{
					overlap_check(ip,local_no);
					r_value = overlap_registers_1[local_no];
				}
				else
				{
					local_no = local_no - (LOCAL + OVERLAP);
					overlap_check(ip,local_no);
					r_value = overlap_registers_2[local_no];
				}
				break;
			case 2:
			/* register number 68-97 */
				local_no = no - (30 * ipflag + GLOBAL);
				if (local_no >= OVERLAP && local_no < OVERLAP+LOCAL) {
					local_no = local_no - OVERLAP;
					banked_check(ip,local_no);
					r_value = banked_registers_2[local_no];
				}
				else if (local_no < OVERLAP)
				{
					overlap_check(ip,local_no);
					r_value = overlap_registers_2[local_no];
				}
				else
				{
					local_no = local_no - (LOCAL + OVERLAP);
					overlap_check(ip,local_no);
					r_value = overlap_registers_3[local_no];
				}
				break;
			case 3:
			/* register number 98-127 */
				local_no = no - (30 * ipflag + GLOBAL);
				if (local_no >= OVERLAP && local_no < OVERLAP+LOCAL) {
					local_no = local_no - OVERLAP;
					banked_check(ip,local_no);
					r_value = banked_registers_3[local_no];
				}
				else if (local_no < OVERLAP)
				{
					overlap_check(ip,local_no);
					r_value = overlap_registers_3[local_no];
				}
				else
				{
					local_no = local_no - (LOCAL + OVERLAP);
					overlap_check(ip,local_no);
					r_value = overlap_registers_0[local_no];
				}
				break;
			default:
				/* printf ("Error: IP number "%d" is invalid.\n", ipflag); */
                                sprintf (errstr,"invalid IP No.%d", ipflag);
                                error(errstr,"reg_read",ip);
		}
	}
	return r_value;
}
#else	/* non-banked register version */
word reg_read (int ip, int no)
{
	word r_value;	/* define the variable for return value */
	int  local_no;
	int  ipflag;

	ipflag = ip;	/* recognize the IP number */

	if (no < 0 || no >= REGSIZE) /* reg-number check */
	{
		sprintf (errstr,"invalid register No.%d",no);
		error(errstr,"reg_read",ip);
		return;
	}
	if (ip < 0 || ip >= NIP) { /* ip-number check */
		sprintf (errstr,"invalid IP No.%d", ip);
		error(errstr,"reg_read",ip);
	}

	if (no >= SPECIALBASE && no < REGSIZE) {
		/* special register */
		r_value = specialregfile[no-SPECIALBASE];
	} else if (no < GLOBAL) {
		/* register number 0-127 */
		global_check(ip,no);
		r_value = global_register[no];
	}
	return r_value;
}
#endif	/* non-banked register version */
