/***
 ***	regmatch.c: check two register No. whether they are identical
 ***		    for fowarding with banked register set
 ***	from Oct.4 1993 to Oct.7 1993 by take.
 ***	
 ***	arranged for ps4v ver.4.0 [94/May/30-Jun/15] by kata.
 ***	arranged for ps4v ver.4.1 [94/Dec/03-] by kata.
 ***/

#include <stdio.h>
#include <stdlib.h>
#include "defines.h"
#include "cmnvar.h"
#include "regdef.h"
#include "register_set.h"

/* register type flags */
#define TYPE_GLOBAL	0
#define TYPE_OVERLAP	1
#define TYPE_LOCAL	2
#define TYPE_SPECIAL	3

/* register number is from 0 to 127;
   if global registers are 4, their number is from 0 to 3.
   if global registers are 8, their number is from 0 to 8.
*/

static char errstr[100]; /* error message temp. */

/*
   if register ip1-reg1 and register ip2-reg2 are physically 
   identical , then return true else returns false.
*/

#ifdef BANKED	/* banked register */
int regmatch (int ip1, int reg1, int ip2, int reg2)
{
	int  reg_type1,reg_type2;
	int  bank_no1,bank_no2;
	int  local_reg1,local_reg2;
	int  ipflag1,ipflag2;

	ipflag1 = ip1;	/* recognize the IP number 1 */
	ipflag2 = ip2;	/* recognize the IP number 2 */

	if (reg1 < 0 || reg1 >= REGSIZE)
	{
		sprintf (errstr,"invalid register No.%d",reg1);
		error(errstr,"regmatch",ip1);
		return;
	}

	if (reg2 < 0 || reg2 >= REGSIZE)
	{
		sprintf (errstr,"invalid register No.%d",reg2);
		error(errstr,"regmatch",ip2);
		return;
	}

	/* ip1-reg1 number conversion */

	if (reg1 >= SPECIALBASE && reg1 < REGSIZE) {
		/* special registers */
		reg_type1 = TYPE_SPECIAL;
		bank_no1 = 0;
		local_reg1 = reg1 - SPECIALBASE;
	} else if (reg1 < GLOBAL) {
		/* register number 0-7 = global */
		reg_type1 = TYPE_GLOBAL;
		bank_no1 = 0;
		local_reg1 = reg1;
	}
	else
	{
		switch (ipflag1)
		{
			case 0:
			/* register number 8-37 */
				local_reg1 = reg1 - (30 * ipflag1 + GLOBAL);
				if (local_reg1 >= OVERLAP && local_reg1 < OVERLAP+LOCAL) {
					local_reg1 = local_reg1 - OVERLAP;
					reg_type1 = TYPE_LOCAL;
					bank_no1 = 0;
				}
				else if (local_reg1 < OVERLAP)
				{
					reg_type1 = TYPE_OVERLAP;
					bank_no1 = 0;
				}
				else
				{
					local_reg1 = local_reg1 - (LOCAL + OVERLAP);
					reg_type1 = TYPE_OVERLAP;
					bank_no1 = 1;
				}
				break;
			case 1:
			/* register number 38-67 */
				local_reg1 = reg1 - (30 * ipflag1 + GLOBAL);
				if (local_reg1 >= OVERLAP && local_reg1 < OVERLAP+LOCAL) {
					local_reg1 = local_reg1 - OVERLAP;
					reg_type1 = TYPE_LOCAL;
					bank_no1 = 1;
				}
				else if (local_reg1 < OVERLAP)
				{
					reg_type1 = TYPE_OVERLAP;
					bank_no1 = 1;
				}
				else
				{
					local_reg1 = local_reg1 - (LOCAL + OVERLAP);
					reg_type1 = TYPE_OVERLAP;
					bank_no1 = 2;
				}
				break;
			case 2:
			/* register number 68-97 */
				local_reg1 = reg1 - (30 * ipflag1 + GLOBAL);
				if (local_reg1 >= OVERLAP && local_reg1 < OVERLAP+LOCAL) {
					local_reg1 = local_reg1 - OVERLAP;
					reg_type1 = TYPE_LOCAL;
					bank_no1 = 2;
				}
				else if (local_reg1 < OVERLAP)
				{
					reg_type1 = TYPE_OVERLAP;
					bank_no1 = 2;
				}
				else
				{
					local_reg1 = local_reg1 - (LOCAL + OVERLAP);
					reg_type1 = TYPE_OVERLAP;
					bank_no1 = 3;
				}
				break;
			case 3:
			/* register number 98-127 */
				local_reg1 = reg1 - (30 * ipflag1 + GLOBAL);
				if (local_reg1 >= OVERLAP && local_reg1 < OVERLAP+LOCAL) {
					local_reg1 = local_reg1 - OVERLAP;
					reg_type1 = TYPE_LOCAL;
					bank_no1 = 3;
				}
				else if (local_reg1 < OVERLAP)
				{
					reg_type1 = TYPE_OVERLAP;
					bank_no1 = 3;
				}
				else
				{
					local_reg1 = local_reg1 - (LOCAL + OVERLAP);
					reg_type1 = TYPE_OVERLAP;
					bank_no1 = 0;
				}
				break;
			default:
				/* printf ("Error: IP number "%d" is invalid.\n", ipflag1); */
                                sprintf (errstr,"invalid IP No.%d", ipflag1);
                                error(errstr,"regmatch",ip1);
		}
	}

	/* ip2-reg2 number conversion */

	if (reg2 >= SPECIALBASE && reg2 < REGSIZE) {
		/* special registers */
		reg_type2 = TYPE_SPECIAL;
		bank_no2 = 0;
		local_reg2 = reg2 - SPECIALBASE;
	} else if (reg2 < GLOBAL) {
		/* register number 0-7 = global */
		reg_type2 = TYPE_GLOBAL;
		bank_no2 = 0;
		local_reg2 = reg2;
	}
	else
	{
		switch (ipflag2)
		{
			case 0:
			/* register number 8-37 */
				local_reg2 = reg2 - (30 * ipflag1 + GLOBAL);
				if (local_reg2 >= OVERLAP && local_reg2 < OVERLAP+LOCAL) {
					local_reg2 = local_reg2 - OVERLAP;
					reg_type2 = TYPE_LOCAL;
					bank_no2 = 0;
				}
				else if (local_reg2 < OVERLAP)
				{
					reg_type2 = TYPE_OVERLAP;
					bank_no2 = 0;
				}
				else
				{
					local_reg2 = local_reg2 - (LOCAL + OVERLAP);
					reg_type2 = TYPE_OVERLAP;
					bank_no2 = 1;
				}
				break;
			case 1:
			/* register number 38-67 */
				local_reg2 = reg2 - (30 * ipflag1 + GLOBAL);
				if (local_reg2 >= OVERLAP && local_reg2 < OVERLAP+LOCAL) {
					local_reg2 = local_reg2 - OVERLAP;
					reg_type2 = TYPE_LOCAL;
					bank_no2 = 1;
				}
				else if (local_reg2 < OVERLAP)
				{
					reg_type2 = TYPE_OVERLAP;
					bank_no2 = 1;
				}
				else
				{
					local_reg2 = local_reg2 - (LOCAL + OVERLAP);
					reg_type2 = TYPE_OVERLAP;
					bank_no2 = 2;
				}
				break;
			case 2:
			/* register number 68-97 */
				local_reg2 = reg2 - (30 * ipflag1 + GLOBAL);
				if (local_reg2 >= OVERLAP && local_reg2 < OVERLAP+LOCAL) {
					local_reg2 = local_reg2 - OVERLAP;
					reg_type2 = TYPE_LOCAL;
					bank_no2 = 2;
				}
				else if (local_reg2 < OVERLAP)
				{
					reg_type2 = TYPE_OVERLAP;
					bank_no2 = 2;
				}
				else
				{
					local_reg2 = local_reg2 - (LOCAL + OVERLAP);
					reg_type2 = TYPE_OVERLAP;
					bank_no2 = 3;
				}
				break;
			case 3:
			/* register number 98-127 */
				local_reg2 = reg2 - (30 * ipflag1 + GLOBAL);
				if (local_reg2 >= OVERLAP && local_reg2 < OVERLAP+LOCAL) {
					local_reg2 = local_reg2 - OVERLAP;
					reg_type2 = TYPE_LOCAL;
					bank_no2 = 3;
				}
				else if (local_reg2 < OVERLAP)
				{
					reg_type2 = TYPE_OVERLAP;
					bank_no2 = 3;
				}
				else
				{
					local_reg2 = local_reg2 - (LOCAL + OVERLAP);
					reg_type2 = TYPE_OVERLAP;
					bank_no2 = 0;
				}
				break;
			default:
				/* printf ("Error: IP number "%d" is invalid.\n", ipflag2); */
                                sprintf (errstr,"invalid IP No.%d", ipflag2);
                                error(errstr,"regmatch",ip2);
		}
	}

	/* identical check */
	if ((reg_type1 == reg_type2) && (bank_no1 == bank_no2) &&
	    (local_reg1 == local_reg2)) {
		/* identical */
		return TRUE;
	} else {
		/* not identical */
		return FALSE;
	}
}
#else	/* non-banked register */
int regmatch (int ip1, int reg1, int ip2, int reg2)
{
	/* identical check */
	if (reg1 == reg2) {
		/* identical */
		return TRUE;
	} else {
		/* not identical */
		return FALSE;
	}
}
#endif
