/* pipe-sim(+vector) ver.3.0 [92/05/28-7/29] */
/*                   ver.3.5a-R1.0 [93/10/25] */
/* pipe-sim(+vector) ver.4.3 [94/12/26-95/01/20] */
/* send operands & opecodes to pipelines (for alu stage) */

#include "defines.h"
#include "latches.h"
#include "vectorls.h"
#include "cmnvar.h"
#include "vpipectl.h"
#include "opcode.h"
#include "extfunc.h"
#include "regdef.h"
#include "register_set.h"

/* start float add/sub pipe controller */
void startfasp(ip,pipeop,src1,src2,dstn,chain,vmr)
int ip;		/* No. of instruction pipe */
int pipeop;	/* opcode for pipeline */
int src1,src2;	/* source operands */
int dstn;	/* distination operands */
int chain;	/* chaining flag */
int vmr;	/* vector mask reg. */
{
	int fp;
	
	fp = ip >> 1;

	if (faspcflag[fp] != FASPC_NOP) {
	    error("float add/sub pipe controller busy.",
		  "alu phase 0",ip);
	} else {
	    faspop[fp] = pipeop;
	    faspchain[fp] = chain;
	    faspsrc1[fp] = src1;
	    faspsrc2[fp] = src2;
	    faspdstn[fp] = dstn;
	    faspvmr[fp] = vmr;
	    faspcflag[fp] = FASPC_INIT;
	    /* Vector Add/sub Counter Reg. initialize */
	    if (SPECIALREG(N_VACR+fp).i == 0) {
		SPECIALREG(N_VACR+fp).i = VREGLEN;
	    } else if (SPECIALREG(N_VACR+fp).i > VREGLEN) {
		error("vector length(VACR) is out of range",
		      "alu phase 0",ip);
	    }
	}
}

/* start float mult pipe controller */
void startfmp(ip,pipeop,src1,src2,dstn,chain,vmr)
int ip;		/* No. of instruction pipe */
int pipeop;	/* opcode for pipeline */
int src1,src2;	/* source operands */
int dstn;	/* distination operands */
int chain;	/* chaining flag */
int vmr;	/* vector mask reg. */
{
	int fp;
	
	fp = ip >> 1;

	if (fmpcflag[fp] != FMPC_NOP) {
	    error("float mult pipe controller busy.",
		  "alu phase 0",ip);
	} else {
	    fmpop[fp] = pipeop;
	    fmpchain[fp] = chain;
	    fmpsrc1[fp] = src1;
	    fmpsrc2[fp] = src2;
	    fmpdstn[fp] = dstn;
	    fmpvmr[fp] = vmr;
	    fmpcflag[fp] = FMPC_INIT;
	    /* Vector Mult Counter Reg. initialize */
	    if (SPECIALREG(N_VMCR+fp).i == 0) {
		SPECIALREG(N_VMCR+fp).i = VREGLEN;
	    } else if (SPECIALREG(N_VMCR+fp).i > VREGLEN) {
		error("vector length(VMCR) is out of range",
		      "alu phase 0",ip);
	    }
	}
}

/* start float div pipe controller */
void startfdp(ip,pipeop,src1,src2,dstn,chain,vmr)
int ip;		/* No. of instruction pipe */
int pipeop;	/* opcode for pipeline */
int src1,src2;	/* source operands */
int dstn;	/* distination operands */
int chain;	/* chaining flag */
int vmr;	/* vector mask reg. */
{
	int fp;
	
	fp = ip >> 1;

	if (fdpcflag[fp] != FDPC_NOP) {
	    error("float div pipe controller busy.",
		  "alu phase 0",ip);
	} else {
	    fdpop[fp] = pipeop;
	    fdpchain[fp] = chain;
	    fdpsrc1[fp] = src1;
	    fdpsrc2[fp] = src2;
	    fdpdstn[fp] = dstn;
	    fdpvmr[fp] = vmr;
	    fdpcflag[fp] = FDPC_INIT;
	    /* Vector Div Counter Reg. initialize */
	    if (SPECIALREG(N_VDCR+fp).i == 0) {
		SPECIALREG(N_VDCR+fp).i = VREGLEN;
	    } else if (SPECIALREG(N_VDCR+fp).i > VREGLEN) {
		error("vector length(VDCR) is out of range",
		      "alu phase 0",ip);
	    }
	}
}

/* start vector int/logic pipe controller */
void startvilp(ip,pipeop,src1,src2,scalarsrc,dstn,chain,vmr)
int ip;		/* No. of instruction pipe */
int pipeop;	/* opcode for pipeline */
int src1,src2;	/* source operands */
word scalarsrc;	/* scalar source operand (VBCAST,VSRL,VSLL) */
int dstn;	/* distination operands */
int chain;	/* chaining flag */
int vmr;	/* vector mask reg. */
{
	int vilp;
	
	vilp = (int)ip/2;

	if (vilpcflag[vilp] != VILPC_NOP) {
	    error("vector int/logic pipe controller busy.",
		  "alu phase 0",ip);
	} else {
	    vilpop[vilp] = pipeop;
	    vilpchain[vilp] = chain;
	    vilpvmr[vilp] = vmr;
	    vilpdstn[vilp] = dstn;
	    switch (pipeop) {
		case VILPOP_ADD:
		case VILPOP_SUB:
		case VILPOP_MUL:
		case VILPOP_DIV:
		case VILPOP_AND:
		case VILPOP_OR:
		case VILPOP_XOR:
		    vilpsrc1[vilp] = src1;
		    vilpsrc2[vilp] = src2;
		    break;
		case VILPOP_SLL:
		case VILPOP_SRL:
		    vilpsrc1[vilp] = src1;
		    vilpscalarsrc[vilp].i = scalarsrc.i;
		    break;
		case VILPOP_NOT:
		case VILPOP_I2F:
		case VILPOP_F2I:
		case VILPOP_ICZ:
		case VILPOP_FCZ:
		    vilpsrc1[vilp] = src1;
		    break;
		case VILPOP_BC:
		    vilpscalarsrc[vilp] = scalarsrc;
		    break;
	    }
	    vilpcflag[vilp] = VILPC_INIT;
	    /* Vector Int./logic Counter Reg. initialize */
	    if (SPECIALREG(N_VICR+vilp).i == 0) {
		SPECIALREG(N_VICR+vilp).i = VREGLEN;
	    } else if (SPECIALREG(N_VICR+vilp).i > VREGLEN) {
		error("vector length(VICR) is out of range",
		      "alu phase 0",ip);
	    }
	}
}
