/* pipe-sim(+vector) ver.3.0 [92/05/28-7/29] */
/*		     ver.3.2 [92/08/27-8/29] */
/*		     ver.3.4 [92/09/14] */
/*                   ver.3.5a-R1.0 [93/10/25] */
/*					*/
/*                   ver.4.0 [94/5/30-06/15] */
/*                   ver.4.3c[96/8/9] */
/*                   ver.4.3e[96/9/20] */
/* simulation main routine */

#include "defines.h"
#include "latches.h"
#include "cmnvar.h"
#include "opcode.h"
#include "extfunc.h"
#include "regdef.h"
#include "register_set.h"

#include "debugger.h"

simulate()
{
	void phase0(),phase1();

	display(cycle,0.0);
	while ((FLAGREG & HALT) != HALT) {
	    cycle++;
	    phase0();
	    display(cycle,0.0);
	    if ((FLAGREG & ERROR) == ERROR) break;
	    phase1();
	    display(cycle,1.0);
	    if ((FLAGREG & ERROR) == ERROR) break;

	    /* breakpoint check */
	    if (checkbreakpoint(PC) == TRUE) return BREAK_STOP;

	    /* breakcycle check */
	    if (checkbreakcycle(cycle) == TRUE) return BREAK_STOP;

	    /* keyboard interrupt check */
	    if (interruptflag == TRUE)  {
		interruptflag = FALSE;
		return INTERRUPT;
	    }
	}
	/* simulation end by HALT instruction */
        if ((FLAGREG & HALT) == HALT) return HALT_END;
	/* simulation end by ERROR */
        if ((FLAGREG & ERROR) == ERROR) return ERROR_END;
}

void phase0()
{
	int ip,fp,vilp,cn;

	for (fp = 0 ; fp < NFASP ; fp++) fasp_out(fp);
	for (fp = 0 ; fp < NFMP ; fp++) fmp_out(fp);
	for (fp = 0 ; fp < NFDP ; fp++) fdp_out(fp);
#ifndef NOVECTOR
	for (vilp = 0 ; vilp < NVILP ; vilp++) vilp_out(vilp);
	for (cn = 0 ; cn < NCN ; cn++) cn_out(cn);
#endif

	for (ip = 0 ; ip < NIP ; ip++) fwb_ss[ip] = fwb_0(ip);
	for (ip = 0 ; ip < NIP ; ip++) wb_ss[ip] = wb_0(ip);
	for (ip = 0 ; ip < NIP ; ip++) mem_ss[ip] = mem_0(ip);
	for (ip = 0 ; ip < NIP ; ip++) alu_ss[ip] = alu_0(ip);
	for (ip = 0 ; ip < NIP ; ip++) id_ss[ip] = decode0(ip);
	if_ss[0] = ifetch0();
	for (ip = 1 ; ip < NIP ; ip++) if_ss[ip] = if_ss[0] + SSLEN*ip;

	for (ip = 0 ; ip < NIP ; ip++) lu_ss[ip] = loadu0(ip);
	for (ip = 0 ; ip < NIP ; ip++) su_ss[ip] = storeu0(ip);
	for (ip = 0 ; ip < NIP ; ip++) pm0(ip);

#ifndef NOVECTOR
	for (ip = 0 ; ip < NIP ; ip++) vlu_ss[ip] = vlu0(ip);
	for (ip = 0 ; ip < NIP ; ip++) vsu_ss[ip] = vsu0(ip);

	for (fp = 0 ; fp < NFASP ; fp++) faspctl_ss[fp] = faspctl0(fp);
	for (fp = 0 ; fp < NFMP ; fp++) fmpctl_ss[fp] = fmpctl0(fp);
	for (fp = 0 ; fp < NFDP ; fp++) fdpctl_ss[fp] = fdpctl0(fp);
	for (vilp = 0 ; vilp < NVILP ; vilp++) vilpctl_ss[vilp] = vilpctl0(vilp);
#endif

	for (fp = 0 ; fp < NFASP ; fp++) fasp_in_ss[fp] = fasp_in(fp);
	for (fp = 0 ; fp < NFMP ; fp++) fmp_in_ss[fp] = fmp_in(fp);
	for (fp = 0 ; fp < NFDP ; fp++) fdp_in_ss[fp] = fdp_in(fp);
#ifndef NOVECTOR
	for (vilp = 0 ; vilp < NVILP ; vilp++) vilp_in_ss[vilp] = vilp_in(vilp);
	for (cn = 0 ; cn < NCN ; cn++) cn_in(cn);
#endif
	check_reg();
}

void phase1()
{
	int ip,fp,vilp,cn;

	for (fp = 0 ; fp < NFASP ; fp++) fasp_out(fp);
	for (fp = 0 ; fp < NFMP ; fp++) fmp_out(fp);
	for (fp = 0 ; fp < NFDP ; fp++) fdp_out(fp);
#ifndef NOVECTOR
	for (vilp = 0 ; vilp < NVILP ; vilp++) vilp_out(vilp);
	for (cn = 0 ; cn < NCN ; cn++) cn_out(cn);
#endif

	for (ip = 0 ; ip < NIP ; ip++) lu_ss[ip] = loadu1(ip);

	for (ip = 0 ; ip < NIP ; ip++) fwb_ss[ip] = fwb_1(ip);
	for (ip = 0 ; ip < NIP ; ip++) wb_ss[ip] = wb_1(ip);
	for (ip = 0 ; ip < NIP ; ip++) mem_ss[ip] = mem_1(ip);
	for (ip = 0 ; ip < NIP ; ip++) alu_ss[ip] = alu_1(ip);
	for (ip = 0 ; ip < NIP ; ip++) id_ss[ip] = decode1(ip);
	if_ss[0] = ifetch1();
	for (ip = 1 ; ip < NIP ; ip++) if_ss[ip] = if_ss[0] + SSLEN*ip;

	for (ip = 0 ; ip < NIP ; ip++) su_ss[ip] = storeu1(ip);
	for (ip = 0 ; ip < NIP ; ip++) pm1(ip);

#ifndef NOVECTOR
	for (ip = 0 ; ip < NIP ; ip++) vlu_ss[ip] = vlu1(ip);
	for (ip = 0 ; ip < NIP ; ip++) vsu_ss[ip] = vsu1(ip);

	for (fp = 0 ; fp < NFASP ; fp++) faspctl_ss[fp] = faspctl1(fp);
	for (fp = 0 ; fp < NFMP ; fp++) fmpctl_ss[fp] = fmpctl1(fp);
	for (fp = 0 ; fp < NFDP ; fp++) fdpctl_ss[fp] = fdpctl1(fp);
	for (vilp = 0 ; vilp < NVILP ; vilp++) vilpctl_ss[vilp] = vilpctl1(vilp);
#endif

	for (fp = 0 ; fp < NFASP ; fp++) fasp_in_ss[fp] = fasp_in(fp);
	for (fp = 0 ; fp < NFMP ; fp++) fmp_in_ss[fp] = fmp_in(fp);
	for (fp = 0 ; fp < NFDP ; fp++) fdp_in_ss[fp] = fdp_in(fp);
#ifndef NOVECTOR
	for (vilp = 0 ; vilp < NVILP ; vilp++) vilp_in_ss[vilp] = vilp_in(vilp);
	for (cn = 0 ; cn < NCN ; cn++) cn_in(cn);
#endif
	check_reg();
	memwait();
#ifndef NOVECTOR
	vectorcount();
#endif
}
