/* pipe-sim(+vector) ver.3.0 [92/05/28-7/29] */
/* pipe-sim(+vector) ver.4.3 [94/12/26-95/01/20] */
/* vector int/logic pipeline control unit */

#include "defines.h"
#include "latches.h"
#include "vectorls.h"
#include "vpipectl.h"
#include "cmnvar.h"
#include "opcode.h"

static int vilpc_op[NVILP],vilpc_chain[NVILP];
static int vilpc_src1[NVILP],vilpc_src2[NVILP],vilpc_dstn[NVILP];
static word vilpc_scalarsrc[NVILP];
static int vilpc_vmr[NVILP],vilpc_vmrcond[NVILP];
static int vilploop[NVILP];

static char vilpc0_status[NVILP][SSLEN];
static char vilpc1_status[NVILP][SSLEN];

char *vilpctl0(ilp)
int	ilp;
{
	switch(vilpcflag[ilp]) {
            case VILPC_NOP: /* no operation */
                strcpy(vilpc0_status[ilp],"no operation");
                break;
	    case VILPC_INIT: /* vector ilp controler initialize */
		vilpc_op[ilp] = vilpop[ilp];
		vilpc_chain[ilp] = vilpchain[ilp];
		vilpc_src1[ilp] = vilpsrc1[ilp];
		vilpc_src2[ilp] = vilpsrc2[ilp];
		vilpc_scalarsrc[ilp] = vilpscalarsrc[ilp];
		vilpc_dstn[ilp] = vilpdstn[ilp];
		vilpc_vmr[ilp] = vilpvmr[ilp];
		vilpc_vmrcond[ilp] = vilpvmrcond[ilp];
		vilploop[ilp] = 0;
		vilpc_vectorlength[ilp] = SPECIALREG(N_VICR+ilp).i;
		if (vil_op[ilp] != VILPOP_NOP) {
		    error("vector int/logic pipe busy.",
			  "vector ilp ctl. unit phase 0",ilp);
		    break;
		} else {
		    vil_op[ilp] = vilpc_op[ilp];
		}
		/* src1,src2 -> vector int/logic pipe */
		write_vilp(ilp);

		/* vector int/logic pipe -> dstn */
		/* read_vilp(ilp); */ /* result is not available yet. */

                /* make return status */
                strcpy(vilpc0_status[ilp],"initialize: ");
                strcat(vilpc0_status[ilp],
                       vilpc_state(ilp,
                                   vilpc_op[ilp],
                                   vilpc_src1[ilp],
                                   vilpc_src2[ilp],
                                   vilpc_scalarsrc[ilp],
                                   vilpc_dstn[ilp],
                                   vilpc_chain[ilp],
                                   vilpc_vmr[ilp],
                                   vilploop[ilp])
                      );

		/* decrement Vector Int/logic Count Reg. */
		SPECIALREG(N_VICR+ilp).i--;

		/* increment loop Count */
		vilploop[ilp]++;

		vilpcflag[ilp] = VILPC_RUN;
		break;	/* end of case VILPC_INIT */

	    case VILPC_RUN:  /* vector ilp controler is running */
		if (SPECIALREG(N_VICR+ilp).i > 0) {
		    if (vil_op[ilp] != VILPOP_NOP) {
			error("vector int/logic pipe busy.",
			      "vector ilp ctl. unit phase 0",ilp);
			break;
		    } else {
			vil_op[ilp] = vilpc_op[ilp];
		    }
		    /* src1,src2 -> vector int/logic pipe */
		    write_vilp(ilp);
		}

		/* vector int/logic pipe -> dstn */
		if (vilploop[ilp] >= VILPLEN) {
		    read_vilp(ilp);
		}

                /* make return status */
                strcpy(vilpc0_status[ilp],"running: ");
                strcat(vilpc0_status[ilp],
                       vilpc_state(ilp,
                                   vilpc_op[ilp],
                                   vilpc_src1[ilp],
                                   vilpc_src2[ilp],
                                   vilpc_scalarsrc[ilp],
                                   vilpc_dstn[ilp],
                                   vilpc_chain[ilp],
                                   vilpc_vmr[ilp],
                                   vilploop[ilp])
                      );

		/* decrement Vector Int/logic Count Reg. */
		if (SPECIALREG(N_VICR+ilp).i > 0) SPECIALREG(N_VICR+ilp).i--;

		if (++vilploop[ilp] == vilpc_vectorlength[ilp] + VILPLEN) {
		    vilpcflag[ilp] = VILPC_NOP;
		}

		break;	/* end of case VILPC_RUN */
	} /* end of switch(vilpcflag[ilp]) */

	/* return status */
	return(vilpc0_status[ilp]);
}

char *vilpctl1(ilp)
int	ilp;
{
	switch (vilpcflag[ilp]) {
            case VILPC_NOP: /* no operation */
                strcpy(vilpc1_status[ilp],"no operation");
                break;
	    case VILPC_RUN:  /* vector ilp controler is running */
		if (SPECIALREG(N_VICR+ilp).i > 0) {
		    if (vil_op[ilp] != VILPOP_NOP) {
			error("vector int/logic pipe busy.",
			      "vector ilp ctl. unit phase 1",ilp);
			break;
		    } else {
			vil_op[ilp] = vilpc_op[ilp];
		    }
		    /* src1,src2 -> vector int/logic pipe */
		    write_vilp(ilp);
		}

		/* vector int/logic pipe -> dstn */
		if (vilploop[ilp] >= VILPLEN) {
		    read_vilp(ilp);
		}

                /* make return status */
                strcpy(vilpc1_status[ilp],"running: ");
                strcat(vilpc1_status[ilp],
                       vilpc_state(ilp,
                                   vilpc_op[ilp],
                                   vilpc_src1[ilp],
                                   vilpc_src2[ilp],
                                   vilpc_scalarsrc[ilp],
                                   vilpc_dstn[ilp],
                                   vilpc_chain[ilp],
                                   vilpc_vmr[ilp],
                                   vilploop[ilp])
                      );

		/* decrement Vector Int/logic Count Reg. */
		if (SPECIALREG(N_VICR+ilp).i > 0) SPECIALREG(N_VICR+ilp).i--;

		if (++vilploop[ilp] == vilpc_vectorlength[ilp] + VILPLEN) {
		    vilpcflag[ilp] = VILPC_NOP;
		}

		break;	/* end of case VILPC_RUN */
	} /* end of switch(vilpcflag[ilp]) */

	/* return status */
	return(vilpc1_status[ilp]);
}

write_vilp(ilp)    /* src1,src2 -> vector int/logic pipe */
int ilp;
{
	switch (vilpc_op[ilp]) {
	    case VILPOP_ADD:
	    case VILPOP_SUB:
	    case VILPOP_MUL:
	    case VILPOP_DIV:
	    case VILPOP_AND:
	    case VILPOP_OR:
	    case VILPOP_XOR:
		write2_vilp(ilp); /* src1&src2=vector */
		break;
	    case VILPOP_NOT:
	    case VILPOP_I2F:
	    case VILPOP_F2I:
		write1_vilp(ilp,VREG); /* src1=vector */
		break;
	    case VILPOP_ICZ:
	    case VILPOP_FCZ:
		write1_vilp(ilp,SETVMR); /* src1=vector */
		break;
	    case VILPOP_SLL:
	    case VILPOP_SRL:
		write2s_vilp(ilp); /* src1=vector,src2=scalar */
		break;
	    case VILPOP_BC:
		write1s_vilp(ilp); /* src1=scalar */
		break;
	}
}

write2_vilp(ilp)    /* src1=vector,src2=vector */
int ilp;
{
	switch (vilpc_chain[ilp]) {
	    case CHAIN_OFF:
		vilpin1[ilp].val = vectorreg[vilpc_src1[ilp]][vilploop[ilp]];
		vilpin1[ilp].regsrc = VREG;
		vilpin1[ilp].src = VR_ENCODE(vilpc_src1[ilp],vilploop[ilp]);
		vilpin1[ilp].regdst = VREG;
		vilpin1[ilp].dst = VR_ENCODE(vilpc_dstn[ilp],vilploop[ilp]);
		vilpin2[ilp].val = vectorreg[vilpc_src2[ilp]][vilploop[ilp]];
		vilpin2[ilp].regsrc = VREG;
		vilpin2[ilp].src = VR_ENCODE(vilpc_src2[ilp],vilploop[ilp]);
		break;
	    case CHAIN_RRC:
		vilpin1[ilp].val = vectorreg[vilpc_src1[ilp]][vilploop[ilp]];
		vilpin1[ilp].regsrc = VREG;
		vilpin1[ilp].src = VR_ENCODE(vilpc_src1[ilp],vilploop[ilp]);
		vilpin1[ilp].regdst = CNET;
		vilpin1[ilp].dst = vilpc_dstn[ilp];
		vilpin2[ilp].val = vectorreg[vilpc_src2[ilp]][vilploop[ilp]];
		vilpin2[ilp].regsrc = VREG;
		vilpin2[ilp].src = VR_ENCODE(vilpc_src2[ilp],vilploop[ilp]);
		break;
	    case CHAIN_RCR:
		vilpin1[ilp].val = vectorreg[vilpc_src1[ilp]][vilploop[ilp]];
		vilpin1[ilp].regsrc = VREG;
		vilpin1[ilp].src = VR_ENCODE(vilpc_src1[ilp],vilploop[ilp]);
		vilpin1[ilp].regdst = VREG;
		vilpin1[ilp].dst = VR_ENCODE(vilpc_dstn[ilp],vilploop[ilp]);
		vilpin2[ilp].val = chainnet_out[vilpc_src2[ilp]];
		vilpin2[ilp].regsrc = CNET;
		vilpin2[ilp].src = vilpc_src2[ilp];
		break;
	    case CHAIN_RCC:
		vilpin1[ilp].val = vectorreg[vilpc_src1[ilp]][vilploop[ilp]];
		vilpin1[ilp].regsrc = VREG;
		vilpin1[ilp].src = VR_ENCODE(vilpc_src1[ilp],vilploop[ilp]);
		vilpin1[ilp].regdst = CNET;
		vilpin1[ilp].dst = vilpc_dstn[ilp];
		vilpin2[ilp].val = chainnet_out[vilpc_src2[ilp]];
		vilpin2[ilp].regsrc = CNET;
		vilpin2[ilp].src = vilpc_src2[ilp];
		break;
	    case CHAIN_CRR:
		vilpin1[ilp].val = chainnet_out[vilpc_src1[ilp]];
		vilpin1[ilp].regsrc = CNET;
		vilpin1[ilp].src = vilpc_src1[ilp];
		vilpin1[ilp].regdst = VREG;
		vilpin1[ilp].dst = VR_ENCODE(vilpc_dstn[ilp],vilploop[ilp]);
		vilpin2[ilp].val = vectorreg[vilpc_src2[ilp]][vilploop[ilp]];
		vilpin2[ilp].regsrc = VREG;
		vilpin2[ilp].src = VR_ENCODE(vilpc_src2[ilp],vilploop[ilp]);
		break;
	    case CHAIN_CRC:
		vilpin1[ilp].val = chainnet_out[vilpc_src1[ilp]];
		vilpin1[ilp].regsrc = CNET;
		vilpin1[ilp].src = vilpc_src1[ilp];
		vilpin1[ilp].regdst = CNET;
		vilpin1[ilp].dst = vilpc_dstn[ilp];
		vilpin2[ilp].val = vectorreg[vilpc_src2[ilp]][vilploop[ilp]];
		vilpin2[ilp].regsrc = VREG;
		vilpin2[ilp].src = VR_ENCODE(vilpc_src2[ilp],vilploop[ilp]);
		break;
	    case CHAIN_CCR:
		vilpin1[ilp].val = chainnet_out[vilpc_src1[ilp]];
		vilpin1[ilp].regsrc = CNET;
		vilpin1[ilp].src = vilpc_src1[ilp];
		vilpin1[ilp].regdst = VREG;
		vilpin1[ilp].dst = VR_ENCODE(vilpc_dstn[ilp],vilploop[ilp]);
		vilpin2[ilp].val = chainnet_out[vilpc_src2[ilp]];
		vilpin2[ilp].regsrc = CNET;
		vilpin2[ilp].src = vilpc_src2[ilp];
		break;
	    case CHAIN_CCC:
		vilpin1[ilp].val = chainnet_out[vilpc_src1[ilp]];
		vilpin1[ilp].regsrc = CNET;
		vilpin1[ilp].src = vilpc_src1[ilp];
		vilpin1[ilp].regdst = CNET;
		vilpin1[ilp].dst = vilpc_dstn[ilp];
		vilpin2[ilp].val = chainnet_out[vilpc_src2[ilp]];
		vilpin2[ilp].regsrc = CNET;
		vilpin2[ilp].src = vilpc_src2[ilp];
		break;
	}
}

write2s_vilp(ilp)    /* src1=vector,src2=scalar */
int ilp;
{
	switch (vilpc_chain[ilp]) {
	    case CHAIN_OFF:
	    case CHAIN_RCR:
		vilpin1[ilp].val = vectorreg[vilpc_src1[ilp]][vilploop[ilp]];
		vilpin1[ilp].regsrc = VREG;
		vilpin1[ilp].src = VR_ENCODE(vilpc_src1[ilp],vilploop[ilp]);
		vilpin1[ilp].regdst = VREG;
		vilpin1[ilp].dst = VR_ENCODE(vilpc_dstn[ilp],vilploop[ilp]);
		vilpin2[ilp].val = vilpc_scalarsrc[ilp];
		vilpin2[ilp].regsrc = FALSE;
		break;
	    case CHAIN_RRC:
	    case CHAIN_RCC:
		vilpin1[ilp].val = vectorreg[vilpc_src1[ilp]][vilploop[ilp]];
		vilpin1[ilp].regsrc = VREG;
		vilpin1[ilp].src = VR_ENCODE(vilpc_src1[ilp],vilploop[ilp]);
		vilpin1[ilp].regdst = CNET;
		vilpin1[ilp].dst = vilpc_dstn[ilp];
		vilpin2[ilp].val = vilpc_scalarsrc[ilp];
		vilpin2[ilp].regsrc = FALSE;
		break;
	    case CHAIN_CRR:
	    case CHAIN_CCR:
		vilpin1[ilp].val = chainnet_out[vilpc_src1[ilp]];
		vilpin1[ilp].regsrc = CNET;
		vilpin1[ilp].src = vilpc_src1[ilp];
		vilpin1[ilp].regdst = VREG;
		vilpin1[ilp].dst = VR_ENCODE(vilpc_dstn[ilp],vilploop[ilp]);
		vilpin2[ilp].val = vilpc_scalarsrc[ilp];
		vilpin2[ilp].regsrc = FALSE;
		break;
	    case CHAIN_CRC:
	    case CHAIN_CCC:
		vilpin1[ilp].val = chainnet_out[vilpc_src1[ilp]];
		vilpin1[ilp].regsrc = CNET;
		vilpin1[ilp].src = vilpc_src1[ilp];
		vilpin1[ilp].regdst = CNET;
		vilpin1[ilp].dst = vilpc_dstn[ilp];
		vilpin2[ilp].val = vilpc_scalarsrc[ilp];
		vilpin2[ilp].regsrc = FALSE;
		break;
	}
}

write1_vilp(ilp,dst)    /* src1=vector */
int ilp;
int dst;	/* regdst flag */
{
	switch (vilpc_chain[ilp]) {
	    case CHAIN_OFF:
	    case CHAIN_RCR:
		vilpin1[ilp].val = vectorreg[vilpc_src1[ilp]][vilploop[ilp]];
		vilpin1[ilp].regsrc = VREG;
		vilpin1[ilp].src = VR_ENCODE(vilpc_src1[ilp],vilploop[ilp]);
		vilpin1[ilp].regdst = dst;
		vilpin1[ilp].dst = VR_ENCODE(vilpc_dstn[ilp],vilploop[ilp]);
		break;
	    case CHAIN_RRC:
	    case CHAIN_RCC:
		vilpin1[ilp].val = vectorreg[vilpc_src1[ilp]][vilploop[ilp]];
		vilpin1[ilp].regsrc = VREG;
		vilpin1[ilp].src = VR_ENCODE(vilpc_src1[ilp],vilploop[ilp]);
		vilpin1[ilp].regdst = CNET;
		vilpin1[ilp].dst = vilpc_dstn[ilp];
		break;
	    case CHAIN_CRR:
	    case CHAIN_CCR:
		vilpin1[ilp].val = chainnet_out[vilpc_src1[ilp]];
		vilpin1[ilp].regsrc = CNET;
		vilpin1[ilp].src = vilpc_src1[ilp];
		vilpin1[ilp].regdst = dst;
		vilpin1[ilp].dst = VR_ENCODE(vilpc_dstn[ilp],vilploop[ilp]);
		break;
	    case CHAIN_CRC:
	    case CHAIN_CCC:
		vilpin1[ilp].val = chainnet_out[vilpc_src1[ilp]];
		vilpin1[ilp].regsrc = CNET;
		vilpin1[ilp].src = vilpc_src1[ilp];
		vilpin1[ilp].regdst = CNET;
		vilpin1[ilp].dst = vilpc_dstn[ilp];
		break;
	}
}

write1s_vilp(ilp)    /* src1=scalar */
int ilp;
{
	switch (vilpc_chain[ilp]) {
	    case CHAIN_OFF:
	    case CHAIN_RCR:
	    case CHAIN_CRR:
	    case CHAIN_CCR:
		vilpin1[ilp].val = vilpc_scalarsrc[ilp];
		vilpin1[ilp].regsrc = FALSE;
		vilpin1[ilp].regdst = VREG;
		vilpin1[ilp].dst = VR_ENCODE(vilpc_dstn[ilp],vilploop[ilp]);
		break;
	    case CHAIN_RRC:
	    case CHAIN_RCC:
	    case CHAIN_CRC:
	    case CHAIN_CCC:
		vilpin1[ilp].val = vilpc_scalarsrc[ilp];
		vilpin1[ilp].regsrc = FALSE;
		vilpin1[ilp].regdst = CNET;
		vilpin1[ilp].dst = vilpc_dstn[ilp];
		break;
	}
}

read_vilp(ilp)    /* vector int/logic pipe -> dstn */
int ilp;
{
	switch (vilpc_op[ilp]) {
	    case VILPOP_ADD:
	    case VILPOP_SUB:
	    case VILPOP_MUL:
	    case VILPOP_DIV:
	    case VILPOP_AND:
	    case VILPOP_OR:
	    case VILPOP_XOR:
	    case VILPOP_NOT:
	    case VILPOP_SLL:
	    case VILPOP_SRL:
	    case VILPOP_BC:
	    case VILPOP_I2F:
	    case VILPOP_F2I:
		read_vilp_to_dstn(ilp); /* pipe -> dstn.reg */
		break;
	    case VILPOP_ICZ:
	    case VILPOP_FCZ:
		read_vilp_set_vmr(ilp); /* pipe -> vector mask reg. */
		break;
	}
}

read_vilp_to_dstn(ilp)	/* vector int/logic pipe -> dstn.reg */
int ilp;
{
	switch (vilpc_chain[ilp]) {
	    case CHAIN_OFF:
	    case CHAIN_RCR:
	    case CHAIN_CRR:
	    case CHAIN_CCR:
		/* write to register */
		if (VMR(vilpc_vmr[ilp],vilploop[ilp]-VILPLEN)==TRUE) {
		    vectorreg[vilpc_dstn[ilp]][vilploop[ilp]-VILPLEN] = vilpout[ilp].val;
		}
		break;
	    case CHAIN_RRC:
	    case CHAIN_RCC:
	    case CHAIN_CRC:
	    case CHAIN_CCC:
		/* write to chaining network */
		chainnet_in[vilpc_dstn[ilp]] = vilpout[ilp].val;
		break;
	}
}

read_vilp_set_vmr(ilp)	/* vector int/logic pipe -> set vmr. */
int ilp;
{
	switch (vilpc_vmrcond[ilp]) {
	    case MC_F_ZERO:
	    case MC_I_ZERO:
		if (vilpout[ilp].val.i==ZERO) {
		    setVMR(vilpc_vmr[ilp],vilploop[ilp]-VILPLEN,TRUE);
		} else {
		    setVMR(vilpc_vmr[ilp],vilploop[ilp]-VILPLEN,FALSE);
		}
	    case MC_F_NOTZERO:
	    case MC_I_NOTZERO:
		if (vilpout[ilp].val.i!=ZERO) {
		    setVMR(vilpc_vmr[ilp],vilploop[ilp]-VILPLEN,TRUE);
		} else {
		    setVMR(vilpc_vmr[ilp],vilploop[ilp]-VILPLEN,FALSE);
		}
	    case MC_F_PLUS:
	    case MC_I_PLUS:
		if (vilpout[ilp].val.i==PLUS) {
		    setVMR(vilpc_vmr[ilp],vilploop[ilp]-VILPLEN,TRUE);
		} else {
		    setVMR(vilpc_vmr[ilp],vilploop[ilp]-VILPLEN,FALSE);
		}
	    case MC_F_NOTPLUS:
	    case MC_I_NOTPLUS:
		if (vilpout[ilp].val.i!=PLUS) {
		    setVMR(vilpc_vmr[ilp],vilploop[ilp]-VILPLEN,TRUE);
		} else {
		    setVMR(vilpc_vmr[ilp],vilploop[ilp]-VILPLEN,FALSE);
		}
	    case MC_F_MINUS:
	    case MC_I_MINUS:
		if (vilpout[ilp].val.i==MINUS) {
		    setVMR(vilpc_vmr[ilp],vilploop[ilp]-VILPLEN,TRUE);
		} else {
		    setVMR(vilpc_vmr[ilp],vilploop[ilp]-VILPLEN,FALSE);
		}
	    case MC_F_NOTMINUS:
	    case MC_I_NOTMINUS:
		if (vilpout[ilp].val.i!=MINUS) {
		    setVMR(vilpc_vmr[ilp],vilploop[ilp]-VILPLEN,TRUE);
		} else {
		    setVMR(vilpc_vmr[ilp],vilploop[ilp]-VILPLEN,FALSE);
		}
	}
}
