/* pipe-sim(+vector) ver.3.0 [92/05/28-7/29] */
/* pipe-sim(+vector) ver.4.3 [94/12/26-95/01/20] */
/* vector integer/logical pipeline */

#include "defines.h"
#include "latches.h"
#include "cmnvar.h"
#include "opcode.h"

#define	INT	0
#define FLOAT	1
#define COMP	2

static reglatch vilp[NVILP][VILPLEN],vilp_clear;
static int vilp_optype[NVILP][VILPLEN];
static int vilptop[NVILP];

static char     vilpstate[NVILP][SSLEN];
static char     vilp_outstate[NVILP][40];

void vilpinit()	/* vector integer/logic pipe initialize */
{
	int	ilp,i;

        /* initialize for fasp clear structure */
        vilp_clear.val.i = 0.0;
        vilp_clear.regsrc = FALSE;
        vilp_clear.src = 0;
        vilp_clear.regdst = FALSE;
        vilp_clear.dst = 0;

	for ( ilp = 0 ; ilp < NVILP ; ilp++) {
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilpout[ilp] = vilp_clear;
		vilptop[ilp] = 0;
		for (i=0 ; i < VILPLEN ; i++) {
			vilp[ilp][i] = vilp_clear;
			vilp_optype[ilp][i] = INT;
		}
	}
}

void vilp_out(ilp) /* ilpipe -> ilpipe_out_latch */
int ilp;
{
	vilpout[ilp] = vilp[ilp][vilptop[ilp]];
	switch (vilp_optype[ilp][vilptop[ilp]]) {
	    case INT:
		sprintf(vilp_outstate[ilp]," ->%6d",vilpout[ilp].val.i);
		break;
	    case FLOAT:
		sprintf(vilp_outstate[ilp]," ->%6.2f",vilpout[ilp].val.f);
		break;
	    case COMP:
		switch (vilp[ilp][vilptop[ilp]].val.i) {
		    case PLUS:
			sprintf(vilp_outstate[ilp]," ->%-6s","PLUS");
			break;
		    case ZERO:
			sprintf(vilp_outstate[ilp]," ->%-6s","ZERO");
			break;
		    case MINUS:
			sprintf(vilp_outstate[ilp]," ->%-6s","MINUS");
			break;
		}
		break;
	}
}
	
char *vilp_in(ilp)/*  ilpipe_in_latches -> ilpipe */
int	ilp;
{
	int	lc,t;
	char	vilp_instate[40],vilp_tmpstr[40];

	switch (vil_op[ilp]) {
	    case VILPOP_NOP:	/* do nothing */
		vilp[ilp][vilptop[ilp]] = vilp_clear;
		vilp_optype[ilp][vilptop[ilp]] = INT;
		sprintf(vilp_instate,"(no operation)=");
		break;
	    case VILPOP_ADD:	/* integer add */
		vilp[ilp][vilptop[ilp]].val.i = vilpin2[ilp].val.i + vilpin1[ilp].val.i ;
		vilp_optype[ilp][vilptop[ilp]] = INT;
		vilp[ilp][vilptop[ilp]].regdst = vilpin1[ilp].regdst;
		vilp[ilp][vilptop[ilp]].dst = vilpin1[ilp].dst ;
		sprintf(vilp_instate,"%6d+%6d =",
			vilpin2[ilp].val.i,vilpin1[ilp].val.i);
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilp_clear;
		break;
	    case VILPOP_SUB:	/* integer subtract */
		vilp[ilp][vilptop[ilp]].val.i = vilpin2[ilp].val.i - vilpin1[ilp].val.i ;
		vilp_optype[ilp][vilptop[ilp]] = INT;
		vilp[ilp][vilptop[ilp]].regdst = vilpin1[ilp].regdst;
		vilp[ilp][vilptop[ilp]].dst = vilpin1[ilp].dst ;
		sprintf(vilp_instate,"%6d-%6d =",
			vilpin2[ilp].val.i,vilpin1[ilp].val.i);
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilp_clear;
		break;
	    case VILPOP_MUL:	/* integer mult */
		vilp[ilp][vilptop[ilp]].val.i = vilpin2[ilp].val.i * vilpin1[ilp].val.i ;
		vilp_optype[ilp][vilptop[ilp]] = INT;
		vilp[ilp][vilptop[ilp]].regdst = vilpin1[ilp].regdst;
		vilp[ilp][vilptop[ilp]].dst = vilpin1[ilp].dst ;
		sprintf(vilp_instate,"%6d*%6d =",
			vilpin2[ilp].val.i,vilpin1[ilp].val.i);
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilp_clear;
		break;
	    case VILPOP_DIV:	/* integer div */
		if (vilpin1[ilp].val.i != 0) {
		    vilp[ilp][vilptop[ilp]].val.i = vilpin2[ilp].val.i / vilpin1[ilp].val.i ;
		} else { /* divide by 0 */
		    vilp[ilp][vilptop[ilp]].val.i = 0;
		}
		vilp_optype[ilp][vilptop[ilp]] = INT;
		vilp[ilp][vilptop[ilp]].regdst = vilpin1[ilp].regdst;
		vilp[ilp][vilptop[ilp]].dst = vilpin1[ilp].dst ;
		sprintf(vilp_instate,"%6d/%6d =",
			vilpin2[ilp].val.i,vilpin1[ilp].val.i);
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilp_clear;
		break;
	    case VILPOP_AND:	/* logical and */
		vilp[ilp][vilptop[ilp]].val.i = vilpin2[ilp].val.i & vilpin1[ilp].val.i ;
		vilp_optype[ilp][vilptop[ilp]] = INT;
		vilp[ilp][vilptop[ilp]].regdst = vilpin1[ilp].regdst;
		vilp[ilp][vilptop[ilp]].dst = vilpin1[ilp].dst ;
		sprintf(vilp_instate,"%6d&%6d =",
			vilpin2[ilp].val.i,vilpin1[ilp].val.i);
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilp_clear;
		break;
	    case VILPOP_OR:	/* logical or */
		vilp[ilp][vilptop[ilp]].val.i = vilpin2[ilp].val.i | vilpin1[ilp].val.i ;
		vilp_optype[ilp][vilptop[ilp]] = INT;
		vilp[ilp][vilptop[ilp]].regdst = vilpin1[ilp].regdst;
		vilp[ilp][vilptop[ilp]].dst = vilpin1[ilp].dst ;
		sprintf(vilp_instate,"%6d|%6d =",
			vilpin2[ilp].val.i,vilpin1[ilp].val.i);
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilp_clear;
		break;
	    case VILPOP_XOR:	/* logical xor */
		vilp[ilp][vilptop[ilp]].val.i = vilpin2[ilp].val.i ^ vilpin1[ilp].val.i ;
		vilp_optype[ilp][vilptop[ilp]] = INT;
		vilp[ilp][vilptop[ilp]].regdst = vilpin1[ilp].regdst;
		vilp[ilp][vilptop[ilp]].dst = vilpin1[ilp].dst ;
		sprintf(vilp_instate,"%6d^%6d =",
			vilpin2[ilp].val.i,vilpin1[ilp].val.i);
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilp_clear;
		break;
	    case VILPOP_NOT:	/* logical not */
		vilp[ilp][vilptop[ilp]].val.i = ~vilpin1[ilp].val.i ;
		vilp_optype[ilp][vilptop[ilp]] = INT;
		vilp[ilp][vilptop[ilp]].regdst = vilpin1[ilp].regdst;
		vilp[ilp][vilptop[ilp]].dst = vilpin1[ilp].dst ;
		sprintf(vilp_instate,"~%6d       =",vilpin1[ilp].val.i);
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilp_clear;
		break;
	    case VILPOP_SLL:	/* shift left logical */
		vilp[ilp][vilptop[ilp]].val.i = vilpin2[ilp].val.i << vilpin1[ilp].val.i ;
		vilp_optype[ilp][vilptop[ilp]] = INT;
		vilp[ilp][vilptop[ilp]].regdst = vilpin1[ilp].regdst;
		vilp[ilp][vilptop[ilp]].dst = vilpin1[ilp].dst ;
		sprintf(vilp_instate,"%6d<<%6d =",
			vilpin2[ilp].val.i,vilpin1[ilp].val.i);
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilp_clear;
		break;
	    case VILPOP_SRL:	/* shift right logical */
		vilp[ilp][vilptop[ilp]].val.i = vilpin2[ilp].val.i >> vilpin1[ilp].val.i ;
		vilp_optype[ilp][vilptop[ilp]] = INT;
		vilp[ilp][vilptop[ilp]].regdst = vilpin1[ilp].regdst;
		vilp[ilp][vilptop[ilp]].dst = vilpin1[ilp].dst ;
		sprintf(vilp_instate,"%6d>>%6d =",
			vilpin2[ilp].val.i,vilpin1[ilp].val.i);
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilp_clear;
		break;
	    case VILPOP_BC:	/* scalar value broadcast */
		vilp[ilp][vilptop[ilp]].val = vilpin1[ilp].val ;
		vilp_optype[ilp][vilptop[ilp]] = INT;
		vilp[ilp][vilptop[ilp]].regdst = vilpin1[ilp].regdst;
		vilp[ilp][vilptop[ilp]].dst = vilpin1[ilp].dst ;
		sprintf(vilp_instate,"b-cast %6d =",vilpin1[ilp].val.i);
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilp_clear;
		break;
	    case VILPOP_I2F:	/* integer to float conversion */
		vilp[ilp][vilptop[ilp]].val.f = (float)vilpin1[ilp].val.i ;
		vilp_optype[ilp][vilptop[ilp]] = FLOAT;
		vilp[ilp][vilptop[ilp]].regdst = vilpin1[ilp].regdst;
		vilp[ilp][vilptop[ilp]].dst = vilpin1[ilp].dst ;
		sprintf(vilp_instate,"(float)%6d =",vilpin1[ilp].val.i);
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilp_clear;
		break;
	    case VILPOP_F2I:	/* float to integer conversion */
		vilp[ilp][vilptop[ilp]].val.i = (int)vilpin1[ilp].val.f ;
		vilp_optype[ilp][vilptop[ilp]] = INT;
		vilp[ilp][vilptop[ilp]].regdst = vilpin1[ilp].regdst;
		vilp[ilp][vilptop[ilp]].dst = vilpin1[ilp].dst ;
		sprintf(vilp_instate,"( int )%7.2f=",vilpin1[ilp].val.i);
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilp_clear;
		break;
	    case VILPOP_ICZ:	/* integer compare to zero */
		if ( vilpin1[ilp].val.i > 0) {
		    vilp[ilp][vilptop[ilp]].val.i = PLUS;
		} else if ( vilpin1[ilp].val.i == 0) {
		    vilp[ilp][vilptop[ilp]].val.i = ZERO;
		} else {
		    vilp[ilp][vilptop[ilp]].val.i = MINUS;
		}
		vilp_optype[ilp][vilptop[ilp]] = COMP;
		vilp[ilp][vilptop[ilp]].regdst = vilpin1[ilp].regdst;
		vilp[ilp][vilptop[ilp]].dst = vilpin1[ilp].dst ;
		sprintf(vilp_instate,"compare%7d=",vilpin1[ilp].val.i);
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilp_clear;
		break;
	    case VILPOP_FCZ:	/* float compare to zero */
		if ( vilpin1[ilp].val.f > 0.0) {
		    vilp[ilp][vilptop[ilp]].val.i = PLUS;
		} else if ( vilpin1[ilp].val.f == 0.0) {
		    vilp[ilp][vilptop[ilp]].val.i = ZERO;
		} else {
		    vilp[ilp][vilptop[ilp]].val.i = MINUS;
		}
		vilp_optype[ilp][vilptop[ilp]] = COMP;
		vilp[ilp][vilptop[ilp]].regdst = vilpin1[ilp].regdst;
		vilp[ilp][vilptop[ilp]].dst = vilpin1[ilp].dst ;
		sprintf(vilp_instate,"compare%7.2f=",vilpin1[ilp].val.i);
		vil_op[ilp] = VILPOP_NOP;
		vilpin1[ilp] = vilpin2[ilp] = vilp_clear;
		break;
	}

        /* make vilp internal status */
        strcpy(vilpstate[ilp],vilp_instate);
        for (lc = vilptop[ilp] ; lc < vilptop[ilp]+VILPLEN ; lc++) {
	    t = lc % VILPLEN;
	    switch (vilp_optype[ilp][t]) {
		case INT:
		    sprintf(vilp_tmpstr," %6d",vilp[ilp][t].val.i);
		    break;
		case FLOAT:
		    sprintf(vilp_tmpstr," %6.2f",vilp[ilp][t].val.f);
		    break;
		case COMP:
		    switch (vilp[ilp][t].val.i) {
			case PLUS:
			    sprintf(vilp_tmpstr," %-6s","PLUS");
			    break;
			case ZERO:
			    sprintf(vilp_tmpstr," %-6s","ZERO");
			    break;
			case MINUS:
			    sprintf(vilp_tmpstr," %-6s","MINUS");
			    break;
		    }
		    break;
	    }
	    strcat(vilpstate[ilp],vilp_tmpstr);
        }
        strcat (vilpstate[ilp],vilp_outstate[ilp]);

        /* refresh vilptop pointer */
	if (--vilptop[ilp] < 0) vilptop[ilp]=VILPLEN-1;

        /* return status */
        return (vilpstate[ilp]);
}
