/* pipe-sim(+vector) ver.4.0 [94/05/30-06/15] */
/* 		     ver.4.3 [94/12/26-95/01/20] */
/* pipeline WB and FWB stage */

#include "defines.h"
#include "latches.h"
#include "vectorls.h"
#include "cmnvar.h"
#include "opcode.h"
#include "extfunc.h"
#include "regdef.h"
#include "register_set.h"

static decoded_inst wb_i_latch[NIP];
static decoded_inst fwb_i_latch[NIP];

static char wb_0_status[NIP][SSLEN];
static char wb_1_status[NIP][SSLEN];

static char fwb_0_status[NIP][SSLEN];
static char fwb_1_status[NIP][SSLEN];

static reglatch wb_internal_latch[NIP];
static reglatch fwb_internal_latch[NIP];

/* register write in WB stage with forwarding */
void reg_write_fwd_wb(int ip, int no, word value)
{
        int     chkip;

        if (no == 0) {  /* zero register */
            return;	/* write to zero register is no effect */
        } else {
            for (chkip = 0; chkip < NIP ; chkip++) {
                /* alu input latch 1 forwarding check */
                if (alu_in1[chkip].regsrc == TRUE) {
                    if (regmatch(ip,no,chkip,alu_in1[chkip].src) == TRUE) {
                        /* register matched!! fowarding... */
                        alu_in1[chkip].val = value;
                    }
                } 
                /* alu input latch 2 forwarding check */
                if (alu_in2[chkip].regsrc == TRUE) {
                    if (regmatch(ip,no,chkip,alu_in2[chkip].src) == TRUE) {
                        /* register matched!! fowarding... */
                        alu_in2[chkip].val = value;
                    }
                } 
            }
            /* write to register... */
            reg_write_wb(ip,no,value);
        }
}

/* register write in FWB stage with forwarding */
void reg_write_fwd_fwb(int ip, int no, word value)
{
        int     chkip;

        if (no == 0) {  /* zero register */
            return;	/* write to zero register is no effect */
        } else {
            for (chkip = 0; chkip < NIP ; chkip++) {
                /* alu input latch 1 forwarding check */
                if (alu_in1[chkip].regsrc == TRUE) {
                    if (regmatch(ip,no,chkip,alu_in1[chkip].src) == TRUE) {
                        /* register matched!! fowarding... */
                        alu_in1[chkip].val = value;
                    }
                } 
                /* alu input latch 2 forwarding check */
                if (alu_in2[chkip].regsrc == TRUE) {
                    if (regmatch(ip,no,chkip,alu_in2[chkip].src) == TRUE) {
                        /* register matched!! fowarding... */
                        alu_in2[chkip].val = value;
                    }
                } 
            }
            /* write to register... */
            reg_write_fwb(ip,no,value);
        }
}

char *wb_0(ip)
int	ip;	/* ip = instruction pipe number */
{
	wb_i_latch[ip] = wblatch[ip];

	switch (wb_i_latch[ip].optype) {
	    case LOAD:
		switch (wb_i_latch[ip].subtype1) {
		    case AD_IMMED:
		    case AD_DIRECT:
		    case AD_BASED:
		    case AD_BASEINDX:
		    case AD_BASESFT:
		    case AD_PCBASED:
		    case AD_PCBINDX:
			reg_write_wb(ip,wb_i_latch[ip].dstn,wb_latch[ip].val);
			/* return status */
			sprintf(wb_0_status[ip],
				"wb_latch %d write to SR%d",
				wb_latch[ip].val.i,wb_i_latch[ip].dstn);
			break;
		    default:
			error("undefined instruction(load irregal addr. mode)",
			      "wb phase 0",ip);
			break;
		}
		break;
	    case STORE: /* do nothing */
		/* return status */
		sprintf(wb_0_status[ip], "no operation ");
		break;
	    case INTCALC:
		switch (wb_i_latch[ip].subtype2) {
		    case OP_ADD:
		    case OP_SUB:
		    case OP_MUL:
		    case OP_DIV:
		    case OP_MOD:
		    case OP_RSUB:
		    case OP_UMUL:
		    case OP_UDIV:
		    case OP_AND:
		    case OP_OR:
		    case OP_XOR:
		    case OP_NOT:
		    case OP_SRL:
		    case OP_SLL:
		    case OP_SRA:
		    case OP_SLA:
			reg_write_wb(ip,wb_i_latch[ip].dstn,wb_latch[ip].val);
			/* return status */
			sprintf(wb_0_status[ip],
				"wb_latch %d write to SR%d",
				wb_latch[ip].val.i,wb_i_latch[ip].dstn);
			break;
		    default:
			error("undefined intcalc instruction",
			      "wb phase 0",ip);
			break;
		}
		break;
	    case FLOATCALC:
                if ((ip & 1) == 0) { /* ip = even number */
                    switch (wb_i_latch[ip].subtype2) {
                        case FOP_ADD:
                        case FOP_SUB:
                        case FOP_FLT:
			    reg_write_wb(ip,wb_i_latch[ip].dstn,wb_latch[ip].val);
                            sprintf(wb_0_status[ip],
                                    "wb_latch %f write to SR%d",
                                    wb_latch[ip].val.f,wb_i_latch[ip].dstn);
                            break;
                        case FOP_INT:
			    reg_write_wb(ip,wb_i_latch[ip].dstn,wb_latch[ip].val);
                            sprintf(wb_0_status[ip],
                                    "wb_latch %d write to SR%d",
                                    wb_latch[ip].val.i,wb_i_latch[ip].dstn);
                            break;
                        case FOP_MLT:
                        case FOP_DIV:
			    /* return status */
			    sprintf(wb_0_status[ip], "no operation");
			    break;
                        default:
                            error("undefined floatcalc operation",
                                  "wb phase 0",ip);
                            break;
                    }
                } else { /* ip = odd number */
                    switch (wb_i_latch[ip].subtype2) {
                        case FOP_ADD:
                        case FOP_SUB:
                        case FOP_FLT:
			    reg_write_fwd_wb(ip,wb_i_latch[ip].dstn,faspout[ip>>1].val);
                            sprintf(wb_0_status[ip],
                                    "faspout %f write to SR%d",
                                    faspout[ip>>1].val.f,wb_i_latch[ip].dstn);
                            break;
                        case FOP_INT:
			    reg_write_fwd_wb(ip,wb_i_latch[ip].dstn,faspout[ip>>1].val);
                            sprintf(wb_0_status[ip],
                                    "faspout %d write to SR%d",
                                    faspout[ip>>1].val.i,wb_i_latch[ip].dstn);
                            break;
                        case FOP_MLT:
                        case FOP_DIV:
			    /* return status */
			    sprintf(wb_0_status[ip], "no operation");
			    break;
                        default:
                            error("undefined floatcalc operation",
                                  "wb phase 0",ip);
                            break;
                    }
		}
		break; /* end of case FLOATCALC */
	    case JUMP:
		switch (wb_i_latch[ip].subtype1) {
		    case JAD_REL:
		    case JAD_BASED: /* no operation */
			/* return status */
			sprintf(wb_0_status[ip], "no operation ");
			break;
		    case JAD_REL_LK:
		    case JAD_BASED_LK:
			if (wb_i_latch[ip].jump_taken==TRUE) {
			    reg_write_wb(ip,wb_i_latch[ip].dstn,wb_latch[ip].val);
			    /* return status */
			    sprintf(wb_0_status[ip],
				    "wb_latch %d write to SR%d",
				    wb_latch[ip].val.i,wb_i_latch[ip].dstn);
			} else {
			    /* return status */
			    sprintf(wb_0_status[ip], "no operation ");
			}
			break;
		}
		break;
	    case COMPJUMP: /* do nothing */
		/* return status */
		sprintf(wb_0_status[ip], "no operation ");
		break;
#ifdef NOVECTOR /* vector not implemented */
	    case VECTOR:
		/* return status */
		sprintf(wb_0_status[ip], "no operation ");
		break;
#else /* vector implemented */
	    case VECTOR:
		switch (wb_i_latch[ip].subtype1) {
		    case V_CALC:
		    case V_LS:
			/* return status */
			sprintf(wb_0_status[ip], "no operation");
			break;
		    case V_MISC:
			switch (wb_i_latch[ip].subtype2) {
			    case VOP_VMROP:
			    case VOP_VMRSET:
			    case VOP_CONV:
			    case VOP_MOVES2V:
				/* return status */
				sprintf(wb_0_status[ip], "no operation");
				break;
			    case VOP_MOVEV2S:
				reg_write_wb(ip,wb_i_latch[ip].src1,wb_latch[ip].val);
				/* return status */
				sprintf(wb_0_status[ip],
				        "wb_latch %d write to SR%d",
					wb_latch[ip].val.i,wb_i_latch[ip].src1);
					break;
				break;
			    default:
				error("undefined vector operation(subtype2)",
				      "wb phase 0",ip);
				break;
			}
			break;
		    default:
			error("undefined vector operation(subtype1)",
			      "wb phase 0",ip);
			break;
		}
                break; /* end of case VECTOR */
#endif
	    case NO_OPR: /* do nothing */
		/* return status */
		sprintf(wb_0_status[ip], "no operation (by NO_OPR)");
		break;
	} /* end of switch (wb_i_latch[ip].optype) */

	/* return status */
	/* strcpy(wb_0_status[ip],printinst(wb_i_latch[ip],ip)); */
	return(wb_0_status[ip]);
}

char *wb_1(ip)
int	ip;
{
	fwblatch[ip] = wb_i_latch[ip];

	/* float WB latch destination register flag initialize */
	fwb_latch[ip].regdst = FALSE;

	switch (wb_i_latch[ip].optype) {
	    case LOAD: /* no operation */
	    case STORE:
	    case INTCALC:
	    case JUMP:
	    case COMPJUMP:
	    case VECTOR:
		/* return status */
		sprintf(wb_1_status[ip], "no operation");
		break;
	    case FLOATCALC:
                if ((ip & 1) == 0) { /* ip = even number */
                    switch (wb_i_latch[ip].subtype2) {
                        case FOP_MLT:
                            fwb_latch[ip] = fmpout[ip>>1];
                            sprintf(wb_1_status[ip],
                                    "fmpout %f set to fwb_latch",
                                    fwb_latch[ip].val.f);
                            break;
                        case FOP_DIV:
                            fwb_latch[ip] = fdpout[ip>>1];
                            sprintf(wb_1_status[ip],
                                    "fdpout %f set to fwb_latch",
                                    fwb_latch[ip].val.f);
                            break;
                        case FOP_ADD:
                        case FOP_SUB:
                        case FOP_FLT:
                        case FOP_INT:
			    /* return status */
			    sprintf(wb_1_status[ip], "no operation");
                            break;
                        default:
                            error("undefined floatcalc operation",
                                  "wb phase 1",ip);
                            break;
                    }
                } else { /* ip = odd number */
                    /* return status */
                    sprintf(wb_1_status[ip], "no operation");
                }
                break; /* end of case FLOATCALC */
	    case NO_OPR: /* do nothing */
		/* return status */
		sprintf(wb_1_status[ip], "no operation (by NO_OPR)");
		break;
	}

	/* return status */
	/* strcpy(wb_1_status[ip],printinst(wb_i_latch[ip],ip)); */
	return(wb_1_status[ip]);
}

char *fwb_0(ip)
int	ip;	/* ip = instruction pipe number */
{
	fwb_i_latch[ip] = fwblatch[ip];

	switch (fwb_i_latch[ip].optype) {
	    case LOAD: 
	    case STORE:
	    case INTCALC:
	    case JUMP:
	    case COMPJUMP:
	    case VECTOR: /* do nothing */
		/* return status */
		sprintf(fwb_0_status[ip], "no operation ");
		break;
	    case FLOATCALC:
                if ((ip & 1) == 0) { /* ip = even number */
                    switch (fwb_i_latch[ip].subtype2) {
                        case FOP_MLT:
                        case FOP_DIV:
			    reg_write_fwb(ip,fwb_i_latch[ip].dstn,fwb_latch[ip].val);
                            sprintf(fwb_0_status[ip],
                                    "fwb_latch %f write to SR%d",
                                    fwb_latch[ip].val.f,fwb_i_latch[ip].dstn);
			    break;
                        case FOP_ADD:
                        case FOP_SUB:
                        case FOP_FLT:
                        case FOP_INT:
			    /* return status */
			    sprintf(fwb_0_status[ip], "no operation");
                            break;
                        default:
                            error("undefined floatcalc operation",
                                  "fwb phase 0",ip);
                            break;
                    }
                } else { /* ip = odd number */
                    switch (fwb_i_latch[ip].subtype2) {
                        case FOP_MLT:
			    reg_write_fwd_fwb(ip,fwb_i_latch[ip].dstn,fmpout[ip>>1].val);
                            sprintf(fwb_0_status[ip],
                                    "fmpout %f write to SR%d",
                                    fmpout[ip>>1].val.f,fwb_i_latch[ip].dstn);
                            break;
                        case FOP_DIV:
			    reg_write_fwd_fwb(ip,fwb_i_latch[ip].dstn,fdpout[ip>>1].val);
                            sprintf(fwb_0_status[ip],
                                    "fdpout %f write to SR%d",
                                    fdpout[ip>>1].val.f,fwb_i_latch[ip].dstn);
			    break;
                        case FOP_ADD:
                        case FOP_SUB:
                        case FOP_FLT:
                        case FOP_INT:
			    /* return status */
			    sprintf(fwb_0_status[ip], "no operation");
                            break;
                        default:
                            error("undefined floatcalc operation",
                                  "fwb phase 0",ip);
                            break;
                    }
		}
		break; /* end of case FLOATCALC */
	    case NO_OPR: /* do nothing */
		/* return status */
		sprintf(fwb_0_status[ip], "no operation (by NO_OPR)");
		break;
	} /* end of switch (fwb_i_latch[ip].optype) */

	/* return status */
	/* strcpy(fwb_0_status[ip],printinst(fwb_i_latch[ip],ip)); */
	return(fwb_0_status[ip]);
}

char *fwb_1(ip)
int	ip;
{
	switch (fwb_i_latch[ip].optype) {
	    case LOAD: /* no operation */
	    case STORE:
	    case INTCALC:
	    case FLOATCALC:
	    case JUMP:
	    case COMPJUMP:
	    case VECTOR:
		/* return status */
		sprintf(fwb_1_status[ip], "no operation");
		break;
	    case NO_OPR: /* do nothing */
		/* return status */
		sprintf(fwb_1_status[ip], "no operation (by NO_OPR)");
		break;
	}

	/* return status */
	/* strcpy(fwb_1_status[ip],printinst(fwb_i_latch[ip],ip)); */
	return(fwb_1_status[ip]);
}
