/***
 ***	write_nochk.c:	write data to register file
 ***			(without collision check)
 ***		for banked register set
 ***	from Oct.7 1993 to Oct.7 1993 by take.
 ***    
 ***    arranged for ps3v ver.3.5a-R1.0 [93/Oct/25] by kata.
 ***    		  ver.3.5a-R1.0b[93/Oct/27] by kata.
 ***    
 ***    arranged for ps4v ver.4.0 [94/May/30-Jun/15] by kata.
 ***    		  ver.4.1 [94/Dec/03-] by kata.
 ***    		  ver.4.2 [94/Dec/19-] by kata.
 ***/

#include <stdio.h>
#include <stdlib.h>
#include "defines.h"
#include "cmnvar.h"
#include "regdef.h"
#include "register_set.h"
/* register number is from 0 to 127;
   if global registers are 4, their number is from 0 to 3.
   if global registers are 8, their number is from 0 to 8.
*/

static char errstr[100]; /* error message temp. */

int reg_write_nochk_i (int ip, int no, int intvalue) 
{
	word writevalue;

	writevalue.i = intvalue;
	return reg_write_nochk(ip,no,writevalue);
}

int reg_write_nochk_f (int ip, int no, float floatvalue) 
{
	word writevalue;

	writevalue.f = floatvalue;
	return reg_write_nochk(ip,no,writevalue);
}

int reg_write_nochk_u (int ip, int no, unsigned int uintvalue) 
{
	word writevalue;

	writevalue.u = uintvalue;
	return reg_write_nochk(ip,no,writevalue);
}

#ifdef BANKED /* banked register */
int reg_write_nochk (int ip, int no, word value)
{
	word r_value;	/* define the variable for return value */
	int  local_no;
	int  ipflag;

	ipflag = ip;	/* recognize the IP number */

	if (no < 0 || no >= REGSIZE)
	{
	/*	printf ("Error: register No.%d is invalid.\n",no); */
		sprintf (errstr,"invalid register No.%d",no);
		error(errstr,"reg_write",ip);
		return FALSE;
	}
	if (no >= SPECIALBASE && no < REGSIZE) {
                /* special register */
                if (no != SPECIALBASE) { /* special[0] is another Zero reg. */
                        specialregfile[no-SPECIALBASE] = value;
                }
        } 
	else if (no < GLOBAL) {
		global_check(ip,no);
		/* register number 0-7 ( 0 = zero-reg ) */
		if (no != 0) global_register[no] = value;
	}
	else
	{
		switch (ipflag)
		{
			case 0:
			/* register number 8-37 */
				local_no = no - (30 * ipflag + GLOBAL);
				if (local_no >= OVERLAP && local_no < OVERLAP+LOCAL) {
					local_no = local_no - OVERLAP;
					banked_check(ip,local_no);
					banked_registers_0[local_no] = value;
				} else if (local_no < OVERLAP)
				{
					overlap_check(ip,local_no);
					overlap_registers_0[local_no] = value;
				}
				else
				{
					local_no = local_no - (LOCAL + OVERLAP);
					overlap_check(ip,local_no);
					overlap_registers_1[local_no] = value;
				}
				break;
			case 1:
			/* register number 38-67 */
				local_no = no - (30 * ipflag + GLOBAL);
				if (local_no >= OVERLAP && local_no < OVERLAP+LOCAL) {
					local_no = local_no - OVERLAP;
					banked_check(ip,local_no);
					banked_registers_1[local_no] = value;
				} else if (local_no < OVERLAP)
				{
					overlap_check(ip,local_no);
					overlap_registers_1[local_no] = value;
				}
				else
				{
					local_no = local_no - (LOCAL + OVERLAP);
					overlap_check(ip,local_no);
					overlap_registers_2[local_no] = value;
				}
				break;
			case 2:
			/* register number 68-97 */
				local_no = no - (30 * ipflag + GLOBAL);
				if (local_no >= OVERLAP && local_no < OVERLAP+LOCAL) {
					local_no = local_no - OVERLAP;
					banked_check(ip,local_no);
					banked_registers_2[local_no] = value;
				} else if (local_no < OVERLAP)
				{
					overlap_check(ip,local_no);
					overlap_registers_2[local_no] = value;
				}
				else
				{
					local_no = local_no - (LOCAL + OVERLAP);
					overlap_check(ip,local_no);
					overlap_registers_3[local_no] = value;
				}
				break;
			case 3:
			/* register number 98-127 */
				local_no = no - (30 * ipflag + GLOBAL);
				if (local_no >= OVERLAP && local_no < OVERLAP+LOCAL) {
					local_no = local_no - OVERLAP;
					banked_check(ip,local_no);
					banked_registers_3[local_no] = value;
				} else if (local_no < OVERLAP)
				{
					overlap_check(ip,local_no);
					overlap_registers_3[local_no] = value;
				}
				else
				{
					local_no = local_no - (LOCAL + OVERLAP);
					overlap_check(ip,local_no);
					overlap_registers_0[local_no] = value;
				}
				break;
			default:
				/* printf ("Error: IP No.%d is invalid.\n", ipflag); */
				sprintf (errstr,"invalid IP No.%d", ipflag);
				error(errstr,"reg_write",ip);
				return FALSE;
		}
	}
	return SUCCESS;
}
#else 	/* non-bank register */
int reg_write_nochk (int ip, int no, word value)
{
	if (no < 0 || no >= REGSIZE)
	{
	/*	printf ("Error: register No.%d is invalid.\n",no); */
		sprintf (errstr,"invalid register No.%d",no);
		error(errstr,"reg_write",ip);
		return FALSE;
	}
	if (no >= SPECIALBASE && no < REGSIZE) {
                /* special register */
                if (no != SPECIALBASE) { /* special[0] is another Zero reg. */
                        specialregfile[no-SPECIALBASE] = value;
                }
        } else if (ip < 0 || ip >= NIP) {
		sprintf (errstr,"invalid IP No.%d", ip);
		error(errstr,"reg_write",ip);
		return FALSE;
	}
	else if (no < GLOBAL) {
		global_check(ip,no);
		/* register number 0-7 ( 0 = zero-reg ) */
		if (no != 0) global_register[no] = value;
	}
	return SUCCESS;
}
#endif
